pro readstandard, name, inwav, flux, pos=pos, unit=unit, path=path, check=check, result=result, stdname=stdname
   ; Reads in a standard star spectrum from absolute calibration database.
   ; If wav is specified, interpolate flux onto wav.
   ; Returns in microJy.

   common stdtable, stdpath, standardtable

    ; should probably enable loading just by RA/DEC after all.

   ; check just makes sure the standard is real; it doesn't actually load it.

   ;common specstdpath, path

   if n_elements(unit) eq 0 then unit = 'flambda/ang'
   if n_elements(path) gt 0 then stdpath = path
   if n_elements(stdpath) eq 0 then stdpath = '~/progs/idl/lris/refdata/' ; slash included this time...

   if n_elements(pos) ge 2 then begin
      if strpos(pos[0],':') gt 0 then radeg = 15*ten(pos[0]) else radeg = strnum(pos[0])
      if strpos(pos[1],':') gt 0 then decdeg = ten(pos[1]) else decdeg = strnum(pos[1])
   endif

   if n_elements(standardtable) eq 0 then begin
      testfile = file_test(stdpath+'specstandards.dat')
      ; could find the location of this file, see galbuilder bc03
      if testfile eq 0 then begin
         print, ' ERROR - File '+stdpath+'specstandards.dat'+' not found.'
         return
      endif
  
      standardtable = readstandardtable(stdpath+'specstandards.dat')
   endif

   if size(name,/tn) ne 'STRING' then print, 'Warning: Standard name is not a string.'
   name = clip(name) ; non-string input should be used with caution
   name = repstr(name,' ','')

   ; output parameters on failure
   flux = -1
   result = 0

   ; Find matching standard based on name, RA, or dec - loop until found
   nstd = n_elements(standardtable)
   for n = 0, nstd-1 do begin
      nmatchname = -1
      match = 0
      matchpos = -1

      if name ne '' then begin
         allnames = [standardtable[n].name]
         if strlen(standardtable[n].altnames) gt 1 then $
            allnames = [allnames, strsplit(standardtable[n].altnames,',',/extract)]
         m = where(name eq allnames, nmatchname)  ; m is 0 if matches the primary name, 1+ if it matches an alternate name
      endif

      if n_elements(radeg) gt 0 and n_elements(decdeg) gt 0 then begin
         matchpos = 0
         cosdec = cos((decdeg < standardtable[n].dec) * !pi/180.)
         if abs(radeg-standardtable[n].ra)*cosdec lt 0.05 and abs(decdeg-standardtable[n].dec) lt 0.05 then begin
            match = 1
            matchpos = 1
            break
         endif
      endif else begin
         ; Name match alone is fine if no coordinates specified
         if nmatchname ge 1 then begin
           match = 1
           break
         endif
      endelse

   endfor

   if match eq 0 then begin
      if keyword_set(check) eq 0 then begin
         if nmatchname eq 0 then print, 'Standard name '+name+' not recognized.'  ; name specified (and no coords), no match
         if matchpos eq 0  then print, 'Coordinates do not match any standard in the database.' ; coords specified, no match
      endif
      return
   endif
   
   result = 1
   stdname = standardtable[n].name
   if keyword_set(check) then return

   file = standardtable[n].file
   indir = (strsplit(file, '/', /extract)) [0] ; the survey subdirectory 

   result = file_test(stdpath+file) ; just check if the file exists.
   if result eq 0 then begin
      print, 'Standard reference spectrum ', stdpath+file, ' not found.'  ; result kind of does double duty, but this shouldn't happen...
      return
   endif

 
   ; read in the standard
      if indir eq 'ctiostan' or indir eq 'hststan' or indir eq 'okestan' or indir eq 'wdstan' then begin
         readcol, stdpath+file, wav, flam_16, flux_mJy, format='f,f,f', /silent
         flux = flux_mJy * 1000. ; we want uJy
      endif
      if indir eq 'calspec' then begin
         readcol, stdpath+file, wav, flam, format='f,f', /silent
         flux = flam / 1d-8 / 1d-29 / 2.998e10 * (wav*1d-8)^2  ; we want uJy
      endif
      if indir eq 'ing' then begin
         nl = countlines(stdpath+file)
         openr, 10, stdpath+file
         iline = ''
         inunit = ''
         wav = fltarr(nl)
         f = fltarr(nl)
         i = 0L
         while not eof(10) do begin
            readf, 10, iline
            if strlen(iline) lt 3 then continue
            if strmid(iline,0,1) eq '*' then continue
            if strmid(iline,0,1) eq '#' then continue
            if strmid(iline,0,3) eq 'SET' then begin
               if strpos(iline,'UNIT') gt 0 then begin
                 p = strpos(iline,'"')
                 inunit = strmid(iline,p+1)
                 p = strpos(inunit,'"')
                 inunit = strlowcase(strmid(inunit,0,p))
               endif
               continue
            endif
            if inunit eq '' then print, 'Unit not recognized?'
            idata = strsplit(iline,/extract)
            wav[i] = idata[0]
            f[i] = idata[1]
            i += 1
         endwhile
         close, 10
         wav = wav[0:i-1] 
         f = f[0:i-1]
         if inunit eq 'ujy' or inunit eq 'micro-janskys' then flux = f
         if inunit eq 'mjy' or inunit eq 'milli-janskys' then flux = f*1000.   ; we want uJy
         if inunit eq 'ab magnitudes' then flux = 10.0^(-(f-23.9)/2.5)
      endif
      if indir eq 'others' then begin ; actually NAOJ, in mags
         readcol, stdpath+file, wav, mag
         flux = mag2ujy(mag,'AB')
         unit = 'uJy'
      endif


   ; units don't get changed if not recognized which could confuse user
   if unit ne 'uJy' and unit ne 'fnu' and unit ne 'flambda/cgs' and unit ne 'flambda/ang' then begin
     print, '  WARNING:  Unit not recognized; assuming uJy'
     return
   endif
   if unit eq 'uJy'         then flux *= 1
   if unit eq 'fnu'         then flux *= 1d-29
   if unit eq 'flambda/cgs' then flux *= 1d-29 * 2.998e10 / (wav*1d-8)^2
   if unit eq 'flambda/ang' then flux *= 1d-29 * 2.998e10 / (wav*1d-8)^2 * 1d-8

   ;readcol, 'fltt7987.oke', wav, flux, fluxunc, fluxbin skip=13

   ;readcol, path+'vma2.tab', wav, flux, skip=15

   if n_elements(inwav) gt 0 then begin
     oldflux = flux
     flux = interpol(flux, wav, inwav)   ; interpolate it onto wav

     ; Extrapolate values above the maximum assuming RJ law
     wabovemax = where(inwav gt max(wav), ct)
     weqmax = (where(wav eq max(wav))) [0]
     if ct gt 0 then flux[wabovemax] = oldflux[weqmax] * (inwav[wabovemax]/max(wav))^(-4.)

     ; the extrapolation for below max is less clear.  implement this when you actually need to.

   endif else begin
     inwav = wav
   endelse

end

