;+
; NAME:
;   specextract.pro
;
; PURPOSE:
;   Extraction of 2D spectra, including some capabilities to automatically trace and wavelength-calibrate
;
; CALLING SEQUENCE:
;   specextract, filename, ycent, $
;   ysearch=ysearch, xinc=xinc, tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace, $
;   arcfile=arcfile, arclist=arclist, brightarclist=brightarclist, skyfile=skyfile, skylist=skylist, $
;   aprad=aprad, apsharp=apsharp, bakdist=bakdist, bakwidth=bakwidth, yoff=yoff, readdsky=readdsky, $
;   cwav=cwav, dispersion=dispersion, uncdisp=uncdisp, uncrefwav=uncrefwav, initsol=initsol, fixsol=fixsol, maxnudge=maxnudge, $
;   outfile=outfile, keywords=keywords, minorder=minorder, maxorder=maxorder, success=success, setkey=setkey, setval=setval
;
;
; INPUTS:
;     I/O parameters
;   filename -   The 2D spectrum.
;   outfile  -   Output file for the extracted 1D spectrum.
;
;     Tracing parameters
;   ycent      - The y coordinate (+-few pix, in middle of image) of the object to extract
;   ysearch    - Region (possibly large) to search for the object to trace
;   xinc       - Step to use in tracing.
;   tracefile  - External file to use for trace instead of current file (use yoff for offset)
;                Can be a text file (list of centroid positions)
;   traceycent - The y coordinate (+-few pix) of object to trace (if different from object to extract)
;   fixtrace   - Flag, set to use exact ycent (instead of searching nearest few pixels)
;
;     Extraction parameters
;   aprad    -   Radius from the central trace to extract (half width)
;   apsharp  -   Sharpness parameter for extraction kernel:  2=gaussian, 6=default, infinity=square
;   bakdist  -   Optional background extraction separation.  default=7/4xaprad if not set.
;   bakwidth -   Background separation width (full width).  Set to 0 to suppress background subtraction.
;   yoff     -   Additional extraction center offset in pixels from best-fit trace center
;
;     Wavelength solution parameters
;   arcfile  -   The 2D spectrum of an arc.  
;   arclist  -   List of arc lines to use in the wavelength solution
;   skyfile  -   2D spectrum of the sky for wavelength refinement.  
;                If an integer, use the extension of the 2D spectrum
;   skylist  -   List of skylines to use in the refined wavelength solution
;   dispersion - Guess at the dispersion of the spectrum (Ang/pix)
;   initsol   -  Guess at the full wav(x) polynomial, optional
;   maxnudge  -  Maximum change in central wavelength possible from initsol
;   fixsol    -  Use initsol with no refinement/improvement
;  
;     Other parameters
;   keywords -   List of FITS header keywords to preserve in output file
;   success  -   Returns a negative flag if procedure failed, or 1 on success
;   setkey   -   Set additional keywords (keyword names)
;   setval   -   Set additional keywords (values)
;
;-
;------------------------------------------------------------------------------



function findobject, img, xsearch=xsearch, ysearch=ysearch, apmin=apmin, apmax=apmax

      s = size(img)
      nx = s[1]
      ny = s[2]

      ; blind search algorithm not rigorously tested for faint objects

      if n_elements(ysearch) ge 2 then begin
         ysearchmin = round(ysearch[0])
         ysearchmax = round(ysearch[1])
      endif
      if n_elements(ysearch) eq 0 then begin
         ysearchmin = 0  + 2
         ysearchmax = ny - 2
      endif

      if n_elements(xsearch) ge 2 then begin
         xmin = round(xsearch[0])
         xmax = round(xsearch[1])
      endif else begin
         xmin = (nx*4)/10
         xmax = (nx*6)/10
      endelse

      logprint, '  Searching for trace between y=',clip(ysearchmin),'-',clip(ysearchmax), '  (x=', clip(xmin),'-',clip(xmax),')'

      wavsum = fltarr(ysearchmax-ysearchmin)
      for y = ysearchmin, ysearchmax-1 do begin
         wavsum[y-ysearchmin] = median(img[xmin:xmax,y]) 
      endfor
      kernel = [0.2,1.0,1.5,1.0,0.2]
      if n_elements(kernel) gt n_elements(wavsum) then kernel = [0.3,1.0,0.3]
      if n_elements(kernel) gt n_elements(wavsum) then kernel = [1.0]
      kernel = kernel / total(kernel)
      wavsumconv = convol(wavsum, kernel, /edge_truncate)

      centersigma = max(wavsum)/stdev(wavsum)   ; stdev is a huge overestimate for a bright star being present.
      if centersigma lt 2 then begin
         logprint, '  WARNING:  Cannot find the source!   Best center is only ', clip(fpr(centersigma,5.2)), '-sigma above background.'
         logprint, '  Continuing anyway, but you will likely want to rerun after specifying ysearch.'
      endif

      avcenty = ysearchmin + (where(wavsumconv eq max(wavsumconv))) [0] ; max val after convolution
      if avcenty lt ysearchmin then avcenty = ysearchmin
      if avcenty gt ysearchmax then avcenty = ysearchmax

      profmax = max(wavsum)
      baseline = median(wavsum, /even)
      for y = avcenty, ysearchmax-1 do begin
         if wavsum[y-ysearchmin] lt baseline + (profmax-baseline)*0.2 then begin
            apmax = y
            break
         endif
      endfor
      for y = avcenty, ysearchmin, -1 do begin  ; formerly ysearchmin-1, which caused a rare error...
         if wavsum[y-ysearchmin] lt baseline + (profmax-baseline)*0.2 then begin
            apmin = y
            break
         endif
      endfor
     
      if n_elements(apmax) eq 0 then apmax = ysearchmax
      if n_elements(apmin) eq 0 then apmin = ysearchmin

      if (apmax-avcenty) gt 6 then apmax = avcenty + 5
      if (avcenty-apmin) gt 6 then apmin = avcenty - 5

      logprint, '  Found object at y=', strtrim(avcenty,2), ' (+', clip(apmax-avcenty), ' -', clip(avcenty-apmin), ')'

   return, avcenty

end

function traceobject, sci, avcenty=avcenty, xinc=xinc, startx=startx
      ; add avcentx to allow starting further from middle?

      ; trace the object across the chip
   
      ; march across the image, left to right, in blocks of 100, summing up horizontally for every y to get 
      ; the variable spatial profile

      debug = 0
      order = 3  ; default order (might change depending on significance)
      tracecheckplot = 0

      imsize = size(sci)
      nx = imsize[1]
      ny = imsize[2]

      xmax = nx-1

      ni = fix(xmax/xinc)
      centy = intarr(ni)
      centx = intarr(ni)
      centsig = fltarr(ni)
      isedge = fltarr(ni)
      isgood = intarr(ni)
      initscansize = 5;3
      scansize = 4;2  ; how far away from the most recently-found center to look for trace variation
      wavsum = fltarr(ny)  ; fltarr(scansize*2+1)
      lastgoodtracei = 0
      ntraceloss = 0
      ntracefound = 0
      lastgoodtracex = -1
      lastgoodtracey = avcenty
      ;kernel = [0.1,0.2,0.4,0.2,0.1]  ; this might be the better kernel for point sources, keep an eye on this
      kernel = [0.15,0.2,0.3,0.2,0.15]  ; should use the input aperture as the size?
      kernel = kernel / total(kernel)

      if debug gt 0 then begin
         window, 0, xsize=700, ysize=700
         !p.multi = [0, 4, 4]
      endif

      if n_elements(startx) eq 0 then starti = fix(i/2) else starti = fix(startx/xinc)
      direction = 1
      i = starti
      for j = 0, ni-1 do begin        ; now : starts in the middle
         if j eq 0 then begin
            lasty = avcenty
            rady = initscansize
         endif else begin
            i += direction
            rady = scansize
         endelse
         if i eq ni then begin
            direction = -direction
            i = starti-1
            lasty = firsty
         endif
         
         ymin = (lasty-rady) > 0
         ymax = (lasty+rady) < (ny-1)
         
         x1 = fix(i*xinc)
         x2 = fix(x1+xinc-1) < xmax

         ; get the maximum (block-summed) pixel and significance
         for y = 0, ny-1 do begin
            wavsum[y] = total(sci[x1:x2,y])  ; sum across xinc pix within the block
         endfor
         bad = where(finite(wavsum) eq 0, ct)
         if ct gt 0 then wavsum[bad] = 0.
         djs_iterstat, wavsum, mean=fullymean, median=fullymedian, sigma=noiserms
         wavsum -= fullymedian
         wavsumsearch = wavsum[ymin:ymax]
         wavsumconv = convol(wavsum, kernel, /edge_truncate) ; not worried about extra proc time
         wavsumsearchconv = wavsumconv[ymin:ymax]
         maxval = max(wavsumsearchconv)
         significance = maxval/noiserms

         ; get the y-coordinate of maximum pixel
         ymaxpix = ymin + (where(wavsumsearchconv eq maxval)) [0]
         if ymaxpix eq ymin or ymaxpix eq ymax then isedge[i] = 1

         ; print, wavsumsearch, ymaxpix

         ; get the refined centroid
         rymin = (ymaxpix-2) > 0
         rymax = (ymaxpix+2) < (ny-1)
         rwavsum = wavsum[rymin:rymax]
         ycentroid = total((findgen(rymax-rymin+1)+rymin) * rwavsum) / total(rwavsum)
         centxi = i*xinc + xinc/2
  
         if debug gt 0 then begin
            plot,  indgen(ymax-ymin+5*2+1)+ymin-5, wavsum[ymin-5:ymax+5], /nodata, /xsty, /ysty
            oplot, indgen(ymax-ymin+5*2+1)+ymin-5, wavsum[ymin-5:ymax+5], psym=10, color=rgb(128,128,128)
            oplot, indgen(ymax-ymin+5*2+1)+ymin-5, wavsumconv[ymin-5:ymax+5], color=rgb(128,128,0)
            oplot, indgen(ymax-ymin+1)+ymin, wavsumsearch, psym=10
            oplot, indgen(ymax-ymin+1)+ymin, wavsumsearchconv, color=rgb(255,255,0), thick=2
            oplot, [ymaxpix, ymaxpix], [0, 60000], color=rgb(0,0,255)
            oplot, [ycentroid, ycentroid], [0, 60000], color=rgb(0,255,0)

            print, '  ',clip(i,3),' ', rclip(x1,4)+'-'+clip(x2,4),rclip(ymin,4),'-',clip(ymax,4),  ymaxpix, ycentroid, significance
         endif

         if isedge[i] eq 0 then lasty = ymaxpix $
                           else lasty = fix((lasty+ymaxpix)/2) ; if on the edge, only move the aperture halfway
         if j eq 0 then firsty = lasty
         
         centx[i] = centxi
         if ycentroid gt rymin and ycentroid lt rymax then $
            centy[i] = ycentroid $
         else $
            centy[i] = ymaxpix
         centsig[i] = significance
      endfor

      sigthresh = 2.5
      if max(centsig) lt 5  then sigthresh = 2.0
      if max(centsig) gt 30 then sigthresh = 3.5
      if max(centsig) gt 50 then sigthresh = 5.0
      if max(centsig) gt 200 then sigthresh = 9.0
      if max(centsig) gt 800 then sigthresh = 30.0  ; this is mainly to avoid centroiding on ghosts or second order light.

      goodtrace = where(centsig gt sigthresh, ct)
      if ct lt 3 then begin
          logprint, '  Failed to trace - insufficient signal.'
          if tracecheckplot then psclose
          return, -1
      endif
      if ct lt n_elements(centsig)/2 then order = 1 
      if ct ge 6 and max(centsig) gt 500 then order += 1    ; fit a higher-order solution to very strong sources (standards)
      ; if ct ge 6 and median(centsig) gt 50 then order += 1     ; could consider turning these on with more testing...
      ; if ct ge 7 and median(centsig) gt 100 then order += 1  

      ; need to make this plot more useful - specific file output, write pixels, etc.

      !p.multi = 0
      if tracecheckplot then begin
         psopen, 'traceplot.ps', xsize=7, ysize=5, /inch
         device, /color
         colors = transpose([[0,0,0],$
                          [180,180,210],$ 
                          [64, 64, 192],$ 
                          [64, 136,128],$ 
                          [64, 158, 64],$  
                          [112,112, 64],$ 
                          [192, 64, 64],$
                          [128, 64, 128]$
                       ])
          tvlct, colors

          plot, centx, centy, /ynozero
          ;oplot, centx, centy2, linestyle=1   centy2 is not reliable 
       endif

      ; fit a polynomial to the profile, excluding possible bad traces



      tracepar = polyiterfit(centx[goodtrace], centy[goodtrace],order=order,niter=2)   ; this should be updated
      logprint, '  Trace function: '      , format='($,A)'
      for o = 0, n_elements(tracepar)-1 do begin
         logprint, clip(tracepar[o]), format='($,A)'
         if o eq 1 then logprint, 'x', format='($,A)'
         if o ge 2 then logprint, 'x^'+clip(o), format='($,A)'
         if o ne n_elements(tracepar)-1 then logprint, ' + ' , format='($,A)'
      endfor
      logprint
      if debug gt 0 then begin
         for x = 0, xmax, 250 do begin
            print, x, poly(x, tracepar)
         endfor
      endif

      tracey = poly(findgen(nx), tracepar)

      ; should consider doing the linear extrapolation method for regions beyond the end of the "goodtrace"

      if tracecheckplot then begin
         oplot, centx[goodtrace], centy[goodtrace], psym=7
         oplot, findgen(nx), tracey, linestyle=2

         psclose
      endif
      return, tracey
 
end



function extract, img, tracey, aprad, apsharp, bakdistin, bakwidthin, fbak=fbak, regfile=regfile, saturate=saturate, nsaturated=nsaturated, medbak=medbak

    if n_elements(medbak) eq 0 then medbak = 1 ; median of the background now on by default.  Set to 0 for older-style behavior.

    if n_elements(bakdistin) eq 1 then bakdist = [bakdistin, bakdistin] 
    if n_elements(bakdistin) ge 2 then bakdist = bakdistin
    if n_elements(bakwidthin) eq 1 then bakwidth = [bakwidthin, bakwidthin] 
    if n_elements(bakwidthin) ge 2 then bakwidth = bakwidthin

    imsize = size(img)
    nx = imsize[1]
    ny = imsize[2]

    baksharp = 12.d

    regstep = 40

    nsaturated = 0

    if n_elements(regfile) gt 0 then begin
       openw, 2, regfile
       printf, 2, '# Region file format: DS9 version 4.1'
       printf, 2, 'global color=green'
       printf, 2, 'image'
    endif

    nans = where(finite(img) eq 0, ct)
    if ct gt 0 then img[nans] = 0.
    for x = 0, nx-1 do if max(img[x,*]) eq 0 and min(img[x,*]) then img[x,*] = !values.f_nan ; blank column

    ; extract the object (and the background, if desired)
    fobj = fltarr(nx)
    fbak = fltarr(nx)

    offmsg = 0
    for x = 0, nx-1 do begin
       cy = tracey[x] 
       if cy lt 0 then begin
           if offmsg eq 0 then logprint, '   WARNING - trace extends off the chip?'
           offmsg = 1
           prevx = x
           prevy = cy
           continue
       endif

       if n_elements(regfile) gt 0 then begin
       if x mod regstep eq 0 then begin
           if x gt 0 then begin
              ; the +1 is to offset IDL/data frame (first cell 0) to DS9 frame (first cell 1)
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy)+','+clip(1+x)+','+clip(1+cy)+') # line=0 0 dash=1'
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+aprad)+','+clip(1+x)+','+clip(1+cy+aprad)+') # line=0 0'
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-aprad)+','+clip(1+x)+','+clip(1+cy-aprad)+') # line=0 0'
              if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. then begin
               if bakdist[1] gt 0 and bakwidth[1] gt 0 then begin
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+bakdist[1])+','+clip(1+x)+','+clip(1+cy+bakdist[1])+') # line=0 0 color=yellow'
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+bakdist[1]+bakwidth[1])+','+clip(1+x)+','+clip(1+cy+bakdist[1]+bakwidth[1])+') # line=0 0 color=yellow'
               endif
               if bakdist[0] gt 0 and bakwidth[0] gt 0 then begin
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-bakdist[0])+','+clip(1+x)+','+clip(1+cy-bakdist[0])+') # line=0 0 color=yellow'
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-bakdist[0]-bakwidth[0])+','+clip(1+x)+','+clip(1+cy-bakdist[0]-bakwidth[0])+') # line=0 0 color=yellow'
               endif
             endif
             prevx = x
             prevy = cy
          endif else begin
             prevx = x
             prevy = cy
          endelse
       endif
       endif


       eky = indgen(ny)
       extractionkernel = exp(-abs(((eky-cy)*1.0d/aprad)^apsharp)/2.)
       if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. and max(bakdist) gt 0. then begin
          if bakdist[0] gt 0 or bakwidth[0] gt 0 then $
             bakkernelm = exp(-abs(((eky-(cy-(bakdist[0]+bakwidth[0]/2.)))*1.0d/(bakwidth[0]/2.))^baksharp)/2.) $
             else bakkernelm = 0.
          if bakdist[1] gt 0 or bakwidth[1] gt 0 then $
             bakkernelp = exp(-abs(((eky-(cy+(bakdist[1]+bakwidth[1]/2.)))*1.0d/(bakwidth[1]/2.))^baksharp)/2.) $
             else bakkernelp = 0.
          bakkernel = bakkernelm + bakkernelp
       endif
            ; weight the two halves

        ; if x eq 1232 then begin
        ;    window, 2
        ;    plot, extractionkernel, xrange=[cy-bakdist-3*bakwidth/2,cy+bakdist+3*bakwidth/2], thick=3, /xstyle
        ;    if bakwidth gt 0. then oplot, bakkernel, linestyle=1, thick=2
        ;    oplot, sci[x,*]/(median(sci[x,cy-20:cy+20])+4*stdev(sci[x,cy-20:cy+20])), psym=10
        ; endif

       if n_elements(saturate) gt 0 then begin
          saturated = where(img[x,*] gt saturate and extractionkernel gt 0.01*max(extractionkernel), ctsaturated)
          nsaturated += ctsaturated
       endif

       fobj[x] = total(img[x,*]*extractionkernel)
       if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. and max(bakdist) gt 0. then begin
          if medbak eq 0 then begin
             fbak[x] = total(img[x,*]*bakkernel) * total(extractionkernel)/total(bakkernel)
          endif else begin
             w = where(bakkernel gt 0.5, ct)
             if ct eq 0 then fbak[x] = !values.f_nan else $
             fbak[x] = median(img[x,w], /even) * total(extractionkernel)
          endelse
       endif
    endfor

    if n_elements(regfile) then close, 2

    return, fobj
end


pro specextract, filename, ycent, $
   xsearch=xsearch, ysearch=ysearch, xinc=xinc, tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace, $
   arcfile=arcfile, arclist=arclist, brightarclist=brightarclist, skyfile=skyfile, skylist=skylist, $
   aprad=aprad, apsharp=apsharp, bakdist=bakdist, bakwidth=bakwidth, yoff=yoff, readdsky=readdsky, $
   cwav=cwav, dispersion=dispersion, unccwav=unccwav, uncdisp=uncdisp, uncrefwav=uncrefwav, initsol=initsol, fixsol=fixsol, $
   outfile=outfile, keywords=keywords, minorder=minorder, maxorder=maxorder, maxnudge=maxnudge, success=success, setkey=setkey, setval=setval, traceonly=traceonly, savetrace=savetrace

   ;if n_elements(arclist) eq 0 then arclist = '~/research/redux/arc/HgArNegood.list'
   ;if n_elements(brightarclist) eq 0 then brightarclist = '~/research/redux/arc/HgNeArbright.list'
   ;if n_elements(skylist) eq 0 then skylist = '~/research/Useful/skyspec/gident_most.dat'
   if n_elements(skylist) eq 0 then skylist = ''

   if n_elements(outfile) eq 0 then outfile = repstr(filename,'.fits','.spec')

   filefail = 0
   checkfiles = [filename]
   if n_elements(arcfile) then checkfiles = [checkfiles, arcfile]
   if n_elements(arclist) eq 1 then checkfiles = [checkfiles, arclist]
   if n_elements(brightarclist) eq 1 then checkfiles = [checkfiles, brightarclist]
   if n_elements(tracefile) gt 0 then checkfiles = [checkfiles, tracefile]
   if n_elements(skyfile) gt 0 then begin
      skyfile = clip(skyfile)
      if skylist ne '' then checkfiles = [checkfiles, skylist]
      if strlen(skyfile) gt 2 then begin
         checkfiles = [checkfiles, skyfile]
      endif else begin
         ; need to check that the extension exists, not clear how to do this efficiently
      endelse
   endif
   for c = 0, n_elements(checkfiles)-1 do begin
      if file_test(checkfiles[c]) eq 0 then begin
         logprint, '  Cannot find ', checkfiles[c]
         filefail += 1
      endif
   endfor
   if filefail gt 0 then begin 
      ; print
      success = -1 ; couldn't find input file
      return
   endif


   ; --- Read in file ---

   sci = mrdfits(filename,0,h, /silent)

   imsize = size(sci)
   nx = imsize[1]
   ny = imsize[2]
   if nx eq 0 or ny eq 0 then begin
     logprint, '  Not a valid 2D FITS spectrum.'
     ; print
     success = -2 ; bad file
     return
   endif

   if n_elements(skyfile) gt 0 then begin
       if strlen(clip(skyfile)) gt 1 then begin
          sky = mrdfits(skyfile, 0, skyh, /silent)
       endif else begin
          sky = mrdfits(filename, fix(skyfile), skyh, /silent) ; use extension
       endelse
   endif

   if n_elements(xsearch) eq 0 then xsearch = [nx*0.4, nx*0.6]
   if xsearch[0] lt 0 then          xsearch = [nx*0.4, nx*0.6] 
   xcent = fix(round(median(xsearch,/even)))

   sat = sxpar(h,'SATURATE', count=ct)
   if ct gt 0 then saturate = sat
   

   ; --- Read in trace file or trace function ---

   if n_elements(xinc) eq 0 then xinc = nx/32.  ; formerly nx/16.

   if n_elements(tracefile) gt 0 then begin
      if strpos(tracefile,'.fit') gt 0 then begin
         if file_test(tracefile) eq 0 then begin
            logprint, 'ERROR: trace file ', tracefile, ' does not exist.'
         endif
         traceimg = mrdfits(tracefile,0,traceh, /silent)
      endif else begin
         ntl = countlines(tracefile)
         if ntl le 2 then begin
            openr, 3, tracefile
            inline = ''
            readf, 3, inline
            while eof(3) eq 0 and (strlen(inline) eq 0 or strmid(clip(inline),0,1) eq '#') do readf, 3, inline
            close, 3
            tracepar = float(strsplit(inline,/extract))
            tracey = poly(findgen(nx), tracepar)
         endif else begin
            rdfloat, tracefile, tracey
         endelse
      endelse
   endif else begin
      traceimg = sci  ; (science image *is* the trace image unless specified)
   endelse


   ; --- Find and trace tracing object, if necessary ---

   if n_elements(traceimg) gt 0 then begin

      if n_elements(traceycent) eq 0 and n_elements(ycent) gt 0 and n_elements(tracefile) eq 0 then traceycent = ycent
      if n_elements(traceycent) eq 0 and n_elements(ysearch) eq 1 then traceycent = ysearch

      ;;if n_elements(traceycent) gt 0 then stop
      if n_elements(traceycent) eq 0 then begin
         traceycent = findobject(traceimg,xsearch=xsearch,ysearch=ysearch, apmin=yfindmin, apmax=yfindmax)
      endif ;else begin
         ;dum = findobject(traceimg,xsearch=xsearch,ysearch=ysearch, apmin=yfindmin, apmax=yfindmax)
      ;endelse
      if n_elements(ycent) eq 0 then ycent = traceycent

      if n_elements(aprad) eq 0 then begin
          if n_elements(yfindmax) gt 0 then aprad = min([yfindmax-traceycent,traceycent-yfindmin]) $
                                       else aprad = 3.
          if aprad lt 2. then aprad = 2. ; need to work on improving the aperture determiner in findobject.
          logprint, '  Using aperture radius ', clip(aprad)
      endif else begin
          logprint, '  Radius was set to ', clip(aprad)
      endelse

      tracey = traceobject(traceimg, avcenty=traceycent, xinc=xinc, startx=median(xsearch,/even))
      if n_elements(tracey) eq 1 then begin
        if tracey eq -1 then begin
           ; print
           success = -3 ; trace failure
           return
        endif
      endif
   endif 

   if n_elements(aprad) eq 0 then aprad = 12
   if n_elements(apsharp) eq 0 then apsharp = 6
   if n_elements(bakdist) eq 0 then bakdist = aprad*7./4
   if n_elements(bakwidth) eq 0 then bakwidth = aprad*4.
     ; formerly didn't set these if a skyfile was given, but it's good to subtract background even with sky by default
     ; since there might be galaxy, twilight gradients, etc.


   ; --- Get optimum trace y-offset, if necessary ----


;   if keyword_set(fixtrace) eq 0 and (n_elements(tracefile) gt 0 or traceycent ne ycent)  then begin
;      ; get the optimum yoff by repeatedly extracting and measuring total flux
;      ; seems to not work
;      print, 'Using multi-extraction method which is currently buggy.....'
;      yoffsearch = ycent-tracey[nx/2] + [-3,3]
;
;      yoffarr = min(yoffsearch) + findgen(max(yoffsearch)-min(yoffsearch))
;      totfluxy = fltarr(n_elements(yoffarr))
;      for ei = 0, n_elements(yoffarr)-1 do begin
;         ftry = extract(sci, tracey+yoffarr[ei], aprad/2. > 1.5, apsharp) ; note no background
;         totfluxy[ei] = total(ftry)
;         ;print, yoffarr[ei], tracey[nx/2]+yoffarr[ei], totfluxy[ei]
;      endfor
;      yoffmax = yoffarr[(where(totfluxy eq max(totfluxy))) [0]] ; maybe fit a gaussian in the future?
;      yofffit = getlinecenter(yoffarr, totfluxy, yoffmax, fitrad=3)
;      ;print, 'Optimum yoff = ', yoff
;      ;print, yoffmax, yofffit
;      print, '  Optimum trace center at x=',clip(nx/2), ': y=', clip(yofffit+tracey[nx/2])
;      yoff = yofffit + yoff
;   endif

   ; if we used a separate tracing object, calculate the offset to the extracted object
   if keyword_set(fixtrace) and n_elements(ycent) gt 0 and n_elements(yoff) eq 0 then begin
      if n_elements(tracey) ne nx then logprint, 'WARNING - trace file has different dimensions from science file.'
      yoff = ycent-tracey[xcent]
   endif else begin
      if n_elements(yoff) eq 0 then yoff = 0
   endelse

   ypixcen = tracey+yoff

   if n_elements(savetrace) gt 0 then begin
       openw, 11, savetrace
         ;printf, 11, '# APRAD = '+aprad
         for i = 0, n_elements(ypixcen)-1 do printf, 11, clip(ypixcen[i])
       close, 11
   endif

   if keyword_set(traceonly) then return


   ; --- Extract 1D spectrum ---


   traceregfile = repstr(outfile,'.spec','.trace.reg')

   if keyword_set(readdsky) and n_elements(skyfile) gt 0 then esci = sci+sky else esci = sci

   fobj = extract(esci, ypixcen, aprad, apsharp, bakdist, bakwidth, fbak=fbak, regfile=traceregfile, saturate=saturate, nsaturated=nsaturated)

   if nsaturated gt 1 then begin
     logprint, '  WARNING: '+clip(nsaturated)+' saturated pixels detected on the trace!!!'
     logprint, '     Spectrum may not be usable.'
   endif




   ; ------------------ Calibrate Wavelength -------------------

    checkplotfile = repstr(outfile,'.spec','.wavsol.ps')

    pix = findgen(nx)

    if n_elements(arcfile) gt 0 then begin

      maxsep = nx*dispersion/2.

      arc = mrdfits(arcfile, 0, arch, /silent) ; use the arcfile

      logprint, '  Extracting arc spectrum.'

      farc = extract(arc, ypixcen, aprad, apsharp)

      if n_elements(brightarclist) eq 1 then $
        brightreflinewav = grabcolumn(brightarclist, 0) $
      else $
        brightreflinewav = brightarclist    

      if n_elements(arclist) eq 1 then $
        morereflinewav = grabcolumn(arclist, 0) $
      else $
        morereflinewav = arclist    

      arcsol = solvearcspec(farc, wav, refwav=morereflinewav, brightrefwav=brightreflinewav, $
       cwav=cwav, unccwav=unccwav, dispersion=dispersion, uncdisp=uncdisp, $
       maxsep=maxsep, minorder=minorder, maxorder=maxorder, uncrefwav=uncrefwav, initsol=initsol, nbrightline=20, $
       nmatch=nmatch, ngoodmatch=ngoodmatch, medabsresidual=medabsresidual, $
       checkplotfile=checkplotfile, maxnudge=maxnudge, /leaveplotopen)
      storecheckplotfile = checkplotfile
      checkplotfile = '' ; don't reopen next time

      sol = arcsol
      initsol = arcsol
    endif
    if n_elements(skyfile) gt 0 then begin

       logprint, '  Extracting sky spectrum.'

       fsky = extract(sky, ypixcen, aprad, apsharp)  

       if skylist ne '' then begin
          readcol, skylist, idn, allreflinewav, allreflinestr, format='i,f,f', /silent
          thresh = 100.
          if n_elements(cwav) eq 0 then cwav = poly(nx/2,initsol)
          if n_elements(dispersion) eq 0 then dispersion = poly(nx/2,initsol)-poly(nx/2-1,initsol)
          if n_elements(uncdisp) eq 0 then uncdisp = dispersion*0.1
          minwav = cwav-unccwav[0] - (dispersion+max(uncdisp))*nx/2. ; estimation only
          maxwav = cwav+unccwav[0] + (dispersion+max(uncdisp))*nx/2. ; estimation only
          w = where(allreflinestr gt thresh and allreflinewav gt minwav and allreflinewav lt maxwav)
          skylinewav = allreflinewav[w] ; ultimately should edit the file?
          skylinestr = allreflinestr[w]

          nline = 100
          if maxwav lt 7000. then nline = 20
          if maxwav lt 6000. then nline = 5

          brightskywav = findbrightest(skylinewav, skylinestr, (nline/5)>3)   
          moreskywav = findbrightest(skylinewav, skylinestr, nline)

          maxsep = nx*dispersion/2.
          logprint, '  Solving for wavelength using sky lines...'
          ;skysol = solvearc(linex, brightlinex, moreskywav, brightskywav, $
          ;                  cwavrange=cwavrange, disprange=disrange, maxsep=maxsep, initsol=initsol, nx=nx, uncrefwav=1.0)

          nofitsky = 1
          skysol = solvearcspec(fsky, wav, refwav=moreskywav, brightrefwav=brightskywav, $
           cwav=cwav, unccwav=unccwav, dispersion=dispersion, uncdisp=uncdisp, $
           maxsep=maxsep, uncrefwav=uncrefwav, initsol=initsol, nofit=nofitsky, nbrightline=20, $
           nmatch=nmatch, ngoodmatch=ngoodmatch, medabsresidual=medabsresidual, $
           checkplotfile=checkplotfile, maxnudge=maxnudge, /leaveplotopen)
          nudge = skysol[0]-initsol[0]  ; the others should be the same
          sol = skysol

       endif

    endif else begin  ; alternatively, could use fbak
       fsky = 0
    endelse 

    if n_elements(arcfile) gt 0 or n_elements(skyfile) gt 0 and skylist ne '' then psclose ; close it here

    if checkplotfile eq '' then checkplotfile = storecheckplotfile
    if file_test(checkplotfile) then $
       retitleplot, checkplotfile, removepath(checkplotfile)+' ('+clip(sxpar(h,'TARGNAME'))+')'
    !p.multi = 0

    if n_elements(sol) eq 0 and n_elements(initsol) gt 1 then begin
       logprint, '  Performing no refinements to the input wavelength calibration.'
       sol = initsol
       wav = poly(pix-nx/2., initsol)
    endif 

    success = 1
    if n_elements(wav) eq 0 then begin
       if n_elements(arcfile) gt 0 and n_elements(arclist) gt 0 or $
          n_elements(skyfile) gt 0 and skylist ne '' then begin
         logprint, '  Unable to calculate wavelength solution.'
         success = -4  ; failed wavelength solution
         return
       endif else begin
         logprint, '  No wavelength calibration information provided; expressing in pixels.'
       endelse
    endif
    if n_elements(nmatch) gt 0 then begin
;      if ngoodmatch lt fix(0.3*nmatch)-2 then begin
    if ngoodmatch lt fix(0.3*nmatch)-2 then begin
         success = 2   ; possible bad wavelength solution
      endif
    endif




   ; Save to disk
   objspec = fobj-fbak
   if n_elements(fsky) gt 0 then $
      skyspec = fsky $
   else $
      skyspec = fbak   
   ; formerly fsky+fbak: this included wings of bright objects which added confusing signal and extra noise.

   ; note crude treatment of uncertainties...
   readnoise = 4.6*sqrt(2.0*aprad) ; formerly, total of extractionkernel
   gain = 1.2
   uncspec = sqrt(readnoise^2 + ((objspec + skyspec) > 0)*gain)/gain

   x = indgen(nx)

   ; now uses writespec
   outh = ['']
   for k = 0, n_elements(keywords)-1 do begin
      if strlen(keywords[k]) le 8 then kval = sxpar(h, keywords[k], comment=comment, count=ct) $
                                  else kval = sxparh(h, keywords[k], comment=comment, count=ct)
      if ct ge 1 then begin
         if strlen(keywords[k]) le 8 then sxaddpar, outh, keywords[k], kval, comment $
                                     else h = [h, keywords[k] + ' = ' + kval + ' / ' + comment]
      endif
   endfor
   if n_elements(arcfile) gt 0 then sxaddpar, outh, 'ARCFILE', arcfile
   if n_elements(cwav) gt 0 then sxaddpar, outh, 'CWAV', cwav
   if n_elements(dispersion) gt 0 then sxaddpar, outh, 'DISPAV', dispersion
   sxaddpar, outh, 'YEXTRACT', ypixcen[nx/2]
   sxaddpar, outh, 'APRAD', aprad
   sxaddpar, outh, 'APSHARP', apsharp
   sxaddpar, outh, 'NSATURAT', nsaturated
   sxaddpar, outh, 'NWAV', n_elements(wav)
   if n_elements(wav) gt 0 then begin
     sxaddpar, outh, 'MINWAV', min(wav)
     sxaddpar, outh, 'MAXWAV', max(wav)
     sxaddpar, outh, 'WAVORDER', n_elements(sol)
     if n_elements(nmatch) gt 0 then sxaddpar, outh, 'WAVMATCH', nmatch
     if n_elements(nmatchgood) gt 0 then sxaddpar, outh, 'WAVMATCG', nmatchgood
     if n_elements(medabsresidual) gt 0 then sxaddpar, outh, 'WAVRESID', medabsresidual
     if n_elements(nudge) gt 0 then sxaddpar, outh, ' WAVNUDGE', nudge
   endif
   if n_elements(bakdist) eq 1 then sxaddpar, outh, 'BAKDIST', bakdist
   if n_elements(bakdist) ge 2 then sxaddpar, outh, 'BAKDISTM', bakdist[0]
   if n_elements(bakdist) ge 2 then sxaddpar, outh, 'BAKDISTP', bakdist[1]
   if n_elements(bakwidth) eq 1 then sxaddpar, outh, 'BAKWIDTH', bakwidth
   if n_elements(bakwidth) ge 2 then sxaddpar, outh, 'BAKWIDTM', bakwidth[0]
   if n_elements(bakwidth) ge 2 then sxaddpar, outh, 'BAKWIDTP', bakwidth[1]
   if n_elements(tracefile) gt 0 then sxaddpar, outh, 'TRACFILE', tracefile
   for k = 0, n_elements(setkey)-1 do begin
      if setkey[k] ne '' then sxaddpar, outh, setkey[k], setval[k]
   endfor
   ;get_date, now
   now = tstamp(/ut)
   sxaddpar, outh, 'EXTDATE', now, 'UT time 1D spectrum extracted by specextract'
   hist = sxpar(h,'HISTORY')
   for hi = 0, n_elements(hist)-1 do begin
     if clip(hist[hi]) ne '' then sxaddhist, hist[hi], outh
   endfor
   hist = 'Processed by specextract '+now+' UT'
   sxaddhist, hist, outh

   nl = n_elements(objspec)
   spec = replicate({wav:0.0d, flux:0.0, sky:0.0, unc:0.0, pix:0L, ypix:0., resp:0.},nl)
   if n_elements(wav)     ge nl then spec.wav = wav
   spec.flux = objspec
   if n_elements(skyspec) ge nl then spec.sky = skyspec
   if n_elements(uncspec) ge nl then spec.unc = uncspec
   if n_elements(pix)     ge nl then spec.pix = pix
   if n_elements(ypixcen) ge nl then spec.ypix = ypixcen

   writespec, outfile, spec, outh

   logprint, '  Spectrum saved to ', outfile


end

