; NAME:
;   lpipe.pro                                (release version 2017.05)
;
; PURPOSE:
;   Fully-automated reduction of LRIS imaging and spectroscopy.
;
; CALLING SEQUENCE:
;   lpipe, [settings]
;
; INPUTS (all optional):
;   datadir      - Location of raw data (current directory if unspecified)
;   mode         - Mode to process (imaging or spectroscopy)
;   camera       - Camera to process (red or blue)
;   chip         - Chip to process (left or right)
;   start        - Start with this step, skipping previous ones
;   stop         - End with this step, skipping subsequent ones
;   step         - Do only this step
;   only         -    (Identical to "step")
;  (files        - Specific file numbers to process)
;  (filters      - Specific filters to process)
;  (gratings     - Specific gratings to process)
;  (targets      - Specific targets to process)
;   redo         - Repeat step(s), overwriting any existing files
;   nocrclean    - Do not zap cosmic rays
;   notelluric   - Skip telluric correction
;  (live         - Don't stack anything we are currently integrating on)
;   continuous   - Keep running, assimilating new data as it appears
;   photometric  - Use zeropoint solution to calibrate non-SDSS fields
;  (background   - Run in background; don't pop up windows)
;   help         - Print some basic help information
;   date         - Process a specific night or month (e.g. '150423')
;   reducer      - Specify name of user to add to the file headers
;
;
; ADDITIONAL OPTIONS:
;   If any of the following files are found in the directory where lpipe
;   is run, it will change the default behavior.
;
;   lpipe.par - Contains various defaults (mainly directory paths)
;   catalog.txt - Source positional catalog.  Use to auto-correct the object
;        name in the header based on position (independent of TARGNAME.)
;   brokenlamps.txt - List of arc lamps that are not functioning.
;   userobjectpos.txt - List of files and extraction positions (y-coordinates).
;        Enables the user to specify which object(s) to extract.
;        Now handled visually by setapertures.pro, or see below to set manually
;   seeing.txt - A file with only two numbers, the minimum and maximum seeing
;        experienced in arcseconds.  (obsolete?)
;   extendedsources.txt - List of imaging fields (TARGNAMEs) containing large 
;        galaxies or other sources for which median-filtered sky subtraction is
;        not appropriate. Scalar background subtraction is used instead. 
;        These will also be excluded from constructing supersky flats.  
;   badflattargets.txt - List of additional targets not to use in making a
;        supersky flat (e.g., fields with lots of scattered light.)
;   badfiles.txt - List of files not to process at all.
;
;
;  (Options in parenthesis above are in development and may not work properly.)
;
;
; OUTPUTS:
;   The following files are written to disk (from most to least processed):
;
;   Imaging -
;   * lris[object]_[filter].fits         - Reduced, stacked image
;     stars[object].cat                  - Photometric catalog for a target
;     coadd[object]_[filter].[l|r].fits  - Reduced, stacked image, one chip only
;     eazfp[b|r]yymmdd_nnnn[l|r].fits    - Individual aligned image
;                                          (one exposure)
;     eazfp[b|r]yymmdd_nnnn[l|r].tca[t|l|z]-Calibrated list of stars in image
;                                          using another exposure of this field
;     eazfp[b|r]yymmdd_nnnn[l|r].cal     - Calibrated list of stars in image
;                                           using SDSS imaging of this field
;     eazfp[b|r]yymmdd_nnnn[l|r].caz     - Calibrated list of stars in image
;                                          using a zeropoint/airmass solution
;     eazfp[b|r]yymmdd_nnnn[l|r].cat     - Uncalibrated list of stars in image
;     azfp[b|r]yymmdd_nnnn[l|r].fits     - Image solved directly against an
;                                          astrometric catalog catalog

;
;   Spectroscopy -
;   * lris[object].spec                 - Flux-calibrated spectrum of an object, 
;                                          combining blue/red and all exposures.
;                                          ASCII table (read with readspec.pro)
;     cels[b|r]yymmdd_nnnn[l|r].[obj].spec - Flux/wave-calibrated spectrum
;     els[b|r]yymmdd_nnnn[l|r].[obj].spec  - Wave-calibrated 1D spectrum
;                                              (not flux-calibrated)  
;     ls[b|r]yymmdd_nnnn[l|r].[obj].fits - Summed 2D spectrum
;     szfp[b|r]yymmdd_nnnn[l|r].fits    - Sky-background subtracted 2D spectrum
;
;
;   Both -
;     zfp[b|r]yymmdd_nnnn[l|r].spec     - Cosmic ray-cleaned image
;     fp[b|r]yymmdd_nnnn[l|r].spec      - Flat-fielded image, split between the
;                                          disconnected left/right chips
;     fp[b|r]yymmdd_nnnn.spec           - Flat-fielded image
;     p[b|r]yymmdd_nnnn.spec            - Overscan-subtracted, bad-pixel 
;                                          marked, coverted to simple format,
;                                          updated header (otherwise raw) image
;
;   Only the starred (*) files are written to the current directory.  Other 
;   files are written to the imredux/ or spredux*/ directories.
;
;         
;
; COMMENTS:
;   This code is meant to be fully automated, producing science-quality
;   output with a single command line with zero intervention.  Reduces
;   both LRIS-B and LRIS-R 2/3 data (LRIS-R1 will be enabled soon) in
;   either imaging or spectroscopy mode.
; 
;   The production of images is quite high-level and includes photometric 
;   calibration of the field (although the accuracy of this has not been
;   robustly tested).  Astrometric distortion corrections are currently
;   in development and may not always work correctly: please report any
;   difficulties with the astrometry.
;
;   Full spectroscopic extraction and reduction (including flux calibration)
;   is also available.
;
;   The program works in a series of steps following standard CCD reduction
;   techniques, automatically recognizing calibration files, matching
;   calibrations together and science frames with appropriate calibrations,
;   etc.  Users looking for more control over the reductions can run each
;   step individually with the "step" command.
;
;   If the reduction is interrupted, the program will resume without
;   incident (although any failed steps may be repeated); each task is 
;   run independently of all others.
;   
;   The program tries very hard to correct for observer mistakes, such
;   as inconsistent windowing of the data.  But it's not perfect.
;   If problems occur, generally the easiest fix is to delete any 
;   offending files (images that are difficult to distinguish whether they 
;   are twilight flats can cause various problems, for example) and rerun.
;   More significant problems generally require direct modification of the
;   code, which is still in production.  See 'Troubleshooting' below for
;   more information.
;
;   Filenames for the input raw images are expected to be in 
;   [b|r]YYMMDD_NNNN.fits format for LRIS-B and LRIS-R, respectively.
;   (You can rename the files by running "lrisrename" if this is not the case.)
;   They can be either in the working directory or in a different directory
;   specified by datadir.
; 
;   The code runs in a series of steps in the following order:
;
;   1. "Prepare" the data by converting the multi-header extension FITS files
;      to a standard single frame, correcting/flagging known pixel defects,
;      cropping unused area of the chip, bias-subtracting using the overscan,
;      and adding extra information to the header.  The conversion/bias 
;      correction employs a modified version of readmhdu.pro.  The X and Y axes
;      for spectroscopic frames are transposed for easier viewing of traces on 
;      standard monitors and to match 1D spectrum plots (wavelength=x-axis).
;      Output: p[b|r]*.fits (written to ./imredux/ and ./spredux/ by default.)
;
;   2. Create flat-fields.  The program searches through all science exposures
;      and determines what flat fields are needed, then attempts to make the
;      best possible flat for each case (generally, the order of preference is
;      twilight > dome > internal halogen, but which type of flat
;      is actually made depends on the number of frames available and their
;      quality.)  In the case of imaging flats, stars are identified and
;      masked out.  For spectroscopic flats, a response correction is applied
;      (these are pixel flats only).  Regions of a spectroscopic flat with low
;      signal (common for LRIS-B in the UV) are not flat-fielded in the 
;      wavelength direction (slit correction only).
;      Output:  lris[b|r]flat*.fits
;
;   3. Flat-correct data.  Divide each image by the flatfield.  A more
;      refined cropping is also done at this stage, depending on the placement
;      of the image area during the run (which is variable.)
;      Output:  fp[b|r]*.fits
;   
;   4. "Re"-flatten imaging data by constructing a super-sky flat using a 
;      median of the science images in a given configuration.   This is 
;      customized for each exposure (the exposure itself is excluded from the
;      flat-field to avoid self-division, which can create peculiar image noise
;      statistics). For LRIS-R1 (pre-2009) I- or RG850-band data this process 
;      will be subtractive rather than divisive (fringe correction), but this 
;      is not implemented yet.  (This can be slow and isn't helpful unless a
;      great deal of imaging is acquired during the night, and is off by 
;      default.)
;      Output:  ifp[b|r]*.fits
;
;   5. Split the data into separate right and left chips.  (Because of the
;      physical gap and rotation between the LRIS chips, they must be treated
;      independently after this point.)  A standard astrometric distortion is
;      added at this point for imaging exposures.
;      Output:  [i]fp[b|r]*[l|r].fits
;
;   6. Removes cosmic rays, using the independent routines pzap.pro and
;      pzapspec.pro.  See those programs for more information.  
;      This can be a time-consuming process.
;      Output: zfp[b|r]*[l|r].fits     (the "i" is removed from imaging.)
;
;   At this point different procedures are applied for imaging and spectroscopy.
;   For imaging:
;  
;   7. Solve astrometry of the field against the best available online catalog
;      (SDSS or USNO-B1.0), using the independent autoastrometry.py code.
;      (Also requires sextractor to be installed.)
;      The functionality of this was changed in May 2015: all images are solved
;      against a catlog, then each image is solved in a relative sense
;      (optionally including distortion correction) against the image nearest 
;      to it, starting from the central image.
;      Output:  azfp[b|r]*[l|r].fits   (catalog alignment)
;              eazfp[b|r]*[l|r].fits   (relative alignment)
;   
;   8. Produce photometric solution of the field.  The GSFC IDLastro aper.pro
;      program is used to measure photometry of point sources in each field
;      for comparison of different exposures (on the same target) to establish 
;      the relative zeropoints.  
;      Output: eazfp[b|r]*[l|r].[t|][cat,cal,caz] and other text files
;  
;   9. Stack exposures.  A weighted, masked median is performed for each field.
;      Requires swarp to be installed and properly linked.
;      Output: coadd[object].[filter].[l|r|]fits
;              lris[object]_[filter].fits
;
;   For spectroscopy:
;
;   7. Subtract night-sky emission lines using subtractsky2, which fits
;      low-order polynomials in blocks passing across the data.   Can 
;      introduce artifacts in some situations (e.g. 0.7" slit)
;      Output: szfp[b|r]*[l|r].fits
;
;   8. Sum exposures.  This uses only the header information: in the event of
;      guiding loss this will not be desired!
;      Output: ls[b|r]*.fits
;
;   9. Determines the approximate trace position of all objects, starting with
;      standard-star observations.
;      Output:  objectpos.dat   [formerly tracepos.dat, stdtracepos.dat]
;
;   10. Wavelength-calibrate arc frames.
;      Output:  fp[b|r]*[l_r].sol
;
;   11. Extract trace to produce a 1D spectrum of the object of interest.
;      Also applies and adjusts wavelength calibration.
;      Output:  els[b|r]*.spec
;
;   12. Flux-calibrate the 1D spectrum using observations of standard stars.
;      Output:  cels[b|r]*.spec
;
;   13. Combine all spectra of a single object in a single grating.
;      (This is an optional step and is now skipped by default.)
;
;   14. Connect red and blue sides of an object to produce final spectrum.
;      Output: lris[object].spec
;
;
; 
; EXAMPLES:
;   1.  In a directory containing a full night worth of LRIS data, enter
;
;       IDL> lpipe
;
;       This will automatically execute all of the steps above.
;       Depending on the quantity and binning of data from your run and the 
;       speed of your computer, the reduction of data from an entire night
;       takes 20 minutes up to about three hours.
;       This defaults to reducing only the right chip, ignoring the left.
;
;
;   2.  Fully reduce only spectroscopy, only the red camera:
;
;       IDL> lpipe, mode='s', camera='red'
;
;
;   3.  Run only the "prepare" step (bias subtraction/formatting), on data 
;       stored in a separate directory:
;
;       IDL> lpipe, step='prepare', datadir='/scr3/user/lris/20120501/'
;
;
;   4.  Run all the 2D spectroscopy reduction steps, but don't do any later
;       (1D) reductions.
;
;       IDL> lpipe, mode='s', stop='skysubtract'
;
;
;   5.  Reduce all the imaging, include both left and chips (full field):
;
;       IDL> lpipe, mode='i', chips='rl'
;
;
;   6.  Convert KOA-style filenames, re-populate missing header information,
;       move raw data to another directory, then run lpipe, and plot all
;       the spectra for quick viewing:
;
;       IDL> lrisrename, 'L?.*.fits'
;       IDL> lriscorrectheaders
;       IDL> $mv [b|r]*_????.fits raw/
;       IDL> lpipe, datadir='raw/'
;       IDL> lrisplotall
;
;
;   7.  Display some information at the command line:
;      
;       IDL> lpipe, /help
;
;
;
;   INSTALLATION:
;
;   Create the subdirectory 'lpipe' somewhere on your hard drive (probably
;   in your IDL directory), and unpack the contents of this installation file
;   there (e.g.,  tar -xvf lpipe.tar.gz).  You will need to tell IDL about
;   the existence of this new directory by editing your idl_startup file
;   (which in turn is specified by .idlenv or in your .bashrc or .cshrc file)
;   to add the string ":+/path/to/lpipe:+/path/to/lpipe/dependencies/"
;   to the IDL_PATH argument.  The GSFC IDLastro routines must also be
;   installed (and visible within IDL_PATH); see http://idlastro.gsfc.nasa.gov/
;   for programs and instructions.
;
;   In order to process images, you will also need to have autoastrometry,
;   swarp and sextractor installed (this requirement will eventually be removed
;   via a simplification of the astrometric solver method).  If the latter two
;   cannot simply be called via "swarp" and "sex", you will need to edit the 
;   file lpipe.par AND also edit the dependencies/autoastrometry.py file
;   to indicate the actual commands to call these routines in the global
;   variables at the top of the code.  The standard UNIX routine wget is also 
;   used to download star catalogs.
;
;   For a Caltech astronomy cluster matchine, it is recommended to simply point
;   your IDL path variable to the latest release version in Daniel Perley's 
;   home directory (see citsetup.txt for more info.)  This will ensure you 
;   always have the most up-to-date version.
;
;
;
;
;   TROUBLESHOOTING:
;
;   While this pipeline is designed to deal with all possible observing
;   circumstances (including many common mistakes), much more testing and 
;   development will be required before this ideal is fully reached.
;   Despite best efforts, the program may crash if it encounters an 
;   unanticipated situation or has problems accomplishing its goals and is
;   unable to proceed.  If you encounter problems, try e-mailing Daniel Perley
;   (dperley@astro.caltech.edu) for assistance, after checking the below.
;
;   If the pipeline does not crash, but does not process any files:
;
;   * Are you using old LRIS data or data downloaded from the KOA archive?
;       The pipeline assumes files are named according to the current on-site
;       filename convention, [b|r]yymmdd_nnnn.fits.   If your files are named
;       using a different convention you can rename them by executing the
;       following command (assuming only raw data is in the current directory):
;         IDL> lrisrename, 'l*.fits'
;  
;   * Is the data in a different directory?
;       If this is the first run of the pipeline on a night, make sure that
;       your raw data is in the current working directory or that you have 
;       correctly pointed to the directory containing raw data with the 
;       "datadir" keyword in lpipe.par.
;
;   * Did you already run the pipeline once on these data, but want to re-run?
;       If you want to re-do a step, it will not overwrite existing
;       files by default (the pipeline is designed to be able to automatically
;       resume from wherever it left off if interrupted or if new data is
;       added.)  To override this behavior, set the /redo keyword, which will
;       overwrite files from any previously-attempted steps.  Be sure to set 
;       the "start" or "step" keywords to the step you want to start at unless
;       you want to restart the pipeline from the beginning for all files.
;
;   * Did it execute the "prepare" step, but will not proceed further?
;       The pipeline will not construct flats it doesn't need, so if no science 
;       exposures exist yet it will not undertake steps beyond 'prepare'.
;       Wait until dark.  (In the future an option to create flats in advance
;       will probably be added.)
;
;
;   If the pipeline crashes during operation:
;
;   * Did it report a 'file not found' or 'variable is undefined' error?  
;       If so, a subroutine, configuration file, or catalog it needed was not 
;       found at the expected location.  Generally this is an installation 
;       problem, but you can inspect the paths in lpipe.par to make sure
;       that the directories and files specified there actually exist.  
;  
;   * Did an external call (autoastrometry, swarp, sex) fail?
;       Check that you have specified the paths correctly and actually have the
;       required software installed for astrometry/coadding (see "installation", 
;       above.)  Alternatively if you don't care about reducing images, specify
;       mode='s' at runtime to process only the spectroscopy, which uses only
;       IDL routines and requires no external packages (except IDLastro).
;   
;   * Did it encounter some other error while processing a file?  
;       Check which file the program was working on when it crashed.  If the
;       file is not essential, try deleting it and re-running the pipeline
;       starting with the current step (or delete ALL of the file's precursors
;       with the same file number and rerun the pipeline.)  Please report the
;       error (and perhaps include a copy of the deleted file) so the problem
;       can be identified and fixed in future releases.
;   
;
;   If processing completed, but the results are problematic:
;
;   * Do the reduced images look bad?
;       The most common source of this is bad flat-fielding.  Especially if
;       twilight flats were not taken, the routine may have attempted to
;       construct a supersky flat, which is difficult to do with Keck due
;       to bright stellar halos and (especially) galaxies.  There is a way 
;       to specify fields not to use in supersky flats (badflattargets.txt).
;       Fields with bright galaxies are also susceptible to sky over-
;       subtraction during stacking (within swarp).  Sky subtraction can be
;       turned off for these fields (currently manually by modifying the swarp
;       file, but a semi-automated method will be coming soon.)  Other times, 
;       a bad image can result from unavoidable instrument issues such as 
;       scattered light from a bright star outside the field.
;
;   * Is the cosmic ray removal malfunctioning?
;       In some circumstances bright nebular emission lines can be zapped, 
;       especially in conditions of good seeing and binned data.  The cosmic
;       ray cleaner settings can be modified internally if this is a problem.
;       Check both the 1D reduction and the 2D subtracted frames to be sure
;       cosmic ray cleaning is not causing problems if nebular emission lines
;       are important to your science.  By default, the clean threshold is set
;       relatively high to avoid zapping real features at the expense of some
;       cosmic rays getting through (especially in imaging mode) but very
;       bright emission lines or images in exceptional seeing can still cause
;       problems.
;
;   * Is the zeropoint calibration of the image unreasonable?
;       The imaging zeropoint-calibration has not been well-tested yet, so 
;       please report this problem so this part of the pipeline can be 
;       improved.  Note that .cat/.tcat catalogs have no direct calibration.
;
;   * Is the wavelength calibration of a spectrum off?
;       Wavelength calibration is probably the most fragile part of the
;       pipeline as there is not yet a careful check on the accuracy of the
;       wavelength solution, and gratings the pipeline hasn't been tested on
;       are not guaranteed to produce good results.  If the wavelength
;       calibration looks suspect, load the relevant *.wavsol.ps plot (in 
;       addition, the f*.wavsol.ps plots contain the master arc solutions)
;       and inspect.  If the wavelength solution is bad, check the input
;       arc file for problems (perhaps it wasn't warmed up?).  If one of the
;       arc lamps was broken (typically the Ne lamp), you can inform the
;       pipeline of this by placing a file containing the abbreviation of 
;       the broken arc element in brokenarcs.txt (in your main reduction 
;       directory).  If all else fails, you can try to solve the arc yourself
;       and place the polynomial terms in the format "[arcname].sol".
;       Note that it is sometimes possible that the grating got 'stuck' and
;       misreported its position, causing the arc and a file to look like
;       they are the same setup (in the header) but actually have a very
;       different wavelength solution.  If the actual central wavelength is
;       more than 200 Angstroms from the keyword value, this will cause a
;       failure in the pipeline unless the WAVELEN keyword in the affected
;       files is edited to the correct value.
;
;   * Were some files skipped?
;       If the pipeline encountered a non-fatal problem processing an
;       individual image or spectrum (such as a failed wavelength calibration
;       or inability to flatfield) then it will simply not process that file 
;       any further.  For most such cases a summary of the problems will be
;       printed out at the end of processing.  If a file is not being
;       processed and you do not see it in the final summary, you can simply
;       rerun the pipeline (without deleting any files and without setting the
;       redo flag) and it will try to repeat any failed steps of this nature. 
;       Recently, the telescope FITS writer has been also randomly omitting 
;       some critical header keywords (including the target name) about 3% 
;       of the time.  Run 'lriscorrectheaders' before running the pipeline 
;       to fix these.
;
;   * Did it extract the wrong object, or no object at all?
;       NOTE: BEHAVIOR HAS CHANGED IN 2016-08 RELEASE AND AGAIN in 2017-03!
;       You can tell if this is the case in several different ways: you can
;       open up a sky-subtracted 2D spectrum and load the corresponding
;       [filename].tracepos.reg file to see the traced aperture, or you can
;       see every file at once by opening "lristraces.html" in a browser
;       window (such as firefox).  Even better, you can view all the extraction
;       apertures at once in the profiles[b|r].ps plots.
;       NEW: Even better than that, setapertures.pro provides a visual way to
;       both inspect and modify your object apertures.
;       The latest version of the pipeline should be much better about guessing
;       the "correct" object to extract and its parameters.  However, if you
;       want to extract a different one, run:
;          IDL> setapertures
;       This offers an intuitive visual interface to change, add, or subtract
;       extraction apertures.  (However, the "old" way of re-specifying
;       user apertures via userobjectpos.txt is still available.)
;       You can update the plots as follows:
;          IDL> lrisplotprofiles, camera='b'  ; (or camera='r')
;       If you're happy with this, delete the higher-level files for this
;       source and re-run the pipeline starting with start='extract'.
;
;
;   * Does the flux calibration / output spectrum look bad?
;       This can have several origins.
;          Trace object mismatch:
;       The program grabbed one object on the red frames and different object
;       in the blue, creating a chimera-spectrum.  Fix this by setting the
;       aperture center manually using userobjectpos.txt, see above.
;          Flux/wavelength calibration:
;       The flux calibration itself is usually robust except when the 
;       wavelength calibration fails first and is undetected.  Try deleting
;       problematic arc files and rerunning.  Imperfect flux calibration near
;       the edges of the chip (especially the dichroic edge) can cause problems
;       with the red+blue connect step.  This is hard to solve except by
;       manually fixing the spectrum.
;          Bad tracing:
;       The tracing routine is overhauled and should normally work, but when it
;       fails it will fail subtly; take a look at the tracing .html files.
;       You can use one frame as the trace standard for another frame by adding
;       the appropriate .trace file as the fifth column in userobjectpos.txt.
;          Contamination:
;       The problem may also relate to contamination of the 2D spectrum by a
;       nearby source: to fix this you can try adjusting the extraction center
;       and width (using userobjectpos.txt; see above) to exclude it.
;       If the contaminator is large and extended (e.g. a nearby host galaxy),
;       you can set a background aperture to the data by changing the
;       "0.0-0.0" column in objectpos.txt to the inner and outer radii of a
;       background aperture, e.g. "10-20", or e.g. "8-14|15-21" to set
;       different backgrounds in the -(down) and +(up) directions.
;
;
;   The risk of encountering pipeline issues can be significantly reduced
;       by following standard observing procedures, below.  
;
;
;
;   OBSERVING TIPS:
;
;   Spectroscopy -
;
;   Slit: The 0.7" slit has many defects, and because this slit moves around
;      due to instrument flexure this inevitably results in extremely poor
;      flat-fielding and sky subtraction.  Use the 1.0" or 1.5" slit instead.
;      (However, a new routine is now available to correct 0.7" slit data
;      using the sky lines.  For now it needs to be run manually.)
;
;   Binning:  Do not bin the red side by more than 2x1 for low-resolution 
;      spectroscopy.  Further binning causes severe saturation issues on bright
;      arc lines needed for automated wavelength calibration, and makes 
;      cosmic-ray identification more difficult.  Be sure to get calibrations 
;      in the same binning setting(s) as your science data for each setup.
;
;   Arcs:  Please turn the mercury lamp on during red-side arc exposures (in
;      addition to the standard neon and argon lamps), and during setup tune 
;      the central wavelength to be sure it appears on the red side.  Having 
;      this line helps the wavelength solution across the D560 dichroic.
;      Note that the new (Fe, Kr, Xe) lamps are NOT yet supported and
;      may even confuse the program, although future upgrades will try to
;      accommodate them.
;
;   Flats:  Halogen flats are very poor: there is signficant spatial banding
;      and a large overall gradient across the y-axis, and significant
;      scattered red light reaches the blue camera.  The intensity of the lamp
;      is also not constant between exposures.  Only in desparate straits
;      should the halogen lamps be relied upon.  Spectroscopic dome flats are
;      much superior:  turn on the spectral dome lamp and get flats as you 
;      would for imaging.  Note that the UV cannot be pixel-calibrated with 
;      either the halogen OR with dome flats - the lamp is not hot enough. 
;      (In principle twilight spectroscopic flats can accomplish this, and 
;      while not supported by the pipeline yet, this is recommended if precise
;      flat-fielding is essential to your UV science.  The new deuterium 
;      lamp is also able to provide UV flat calibration, but has not been
;      implemented in the pipeline.)
;
;   Standards:  The list of recognized spectrophotometric flux-calibration 
;      standards can be viewed in the 'refdata' subdirectory as the file
;      specstandards.dat, or online at:
;      http://astro.caltech.edu/~dperley/programs/lpipe/specstandards.txt
;      Standards not on this list will not be recognized as standards (if you
;      have a favorite standard not on this list, please suggest it.)  Ideally
;      the same standard should be observed twice at significantly different 
;      airmasses in each setup.
;      It is recommended to take arcs alongside observations of bright 
;      standards where a short integration time (<20 sec) is employed or for
;      standards observed in bright twilight.  This will ensure the standard 
;      has a good wavelength solution if refinement based on night-sky OH 
;      emission is impossible, which in turn will help ensure accurate 
;      flux-calibration.  This is important mainly for telluric calibration
;      of the red side (errors in the wavelength solution will produce
;      significant calibration artifacts in the telluric bands.)
;
;
;   Imaging -
;
;   Windowing:  You can window the red side to Image Area on the XLRIS CONFIG
;      menu to save readout and processing time with essentially no compromise.
;      One can window it even more to save more readout time at the expense
;      of FOV (use the ETC menu).  Windowing is fully integrated into the 
;      pipeline, and images of different windowing settings can be processed
;      without difficulty even if the windowing of the flats differs from that
;      of the target.
;
;   Flats:  In general, twilight flats produce the most reliable results, so
;      try to get them if possible.  Observing in the U-band REQUIRES twilight
;      flats since there are not enough dark-sky counts and the dome lamp is
;      not hot enough to emit in the UV.  (Twilight flats do seem to create
;      artifacts with the glue defects in some filters, however.)  Also, note
;      that the mirror does not return precisely to its original state after
;      changes in and out of spectroscopy mode, changing the vignetting and
;      dust pattern.  Try to avoid changing the configuration between flats
;      and science frames, or take sufficient science frames that super-sky
;      correction flats ('reflat' step) can be constructed reliably.  Of
;      course, this is not always possible.
;
;   Standards:  If you are imaging any uncalibrated fields and conditions are 
;      photometric, try to get at least two measurements of 
;      Landolt photometric standard fields (preferably two observations of 
;      the same standard field) per filter at significantly different 
;      airmasses.  This can be done during twilight.  While the program now
;      uses imaging of science fields in SDSS to help calculate the zerpoint
;      solution, these can be susceptible to crowding when calculating the
;      aperture correction and Landolt fields are useful to verify them.
;
;   Science fields:  Always specify that you want the source placed at the 
;      "LRIS-B imaging position" to avoid the chip gap.  A test exposure is
;      usually not necessary if the operator is trustworthy, but a short
;      exposure (< ~30s) will help significantly with calibration later if a 
;      moderately deep (e.g. SDSS) calibration of the field is not available,
;      especially in the redder filters (z- / RG850- band in particular).
;      When dithering, try to make sure that every image in the
;      dither sequence is at a different vertical position (as seen on the
;      live DS9 display) by e.g. dithering 1" vertically to accompany each
;      otherwise horizontal dither. LRIS-R3 has some bad columns (horizontal
;      lines in the viewer orientation) which one should avoid keeping over 
;      the same horizontal location (helpful, since these columns actually 
;      fluctuate in intensity and do not always flat-field out.)
;
;   Organization -
; 
;   Use save_state and restore_state (NOT the GUI) to store your LRIS 
;      configurations.  This includes the CCD parameters and will ensure that
;      you do not forget to change the binning and windowing which are NOT
;      included in the XLRIS save/load options.  
;
;   To check what calibrations (etc.) you have or need, a useful tool is the
;      lriscat.py code, available at: 
;         http://www.astro.caltech.edu/~dperley/programs/lriscat.py
;      Run this at the command-line (python lriscat.py *.fits) to produce an
;      immediate catalog table containing all relevant LRIS settings.
;      A slightly more basic version is also available within this package:
;         IDL> lriscat, '*.fits'
;      The IDL version also works on .spec files.
;
;   You can also plot all your spectra acquired during the night as follows:
;         IDL> lrisplotall
;
;   And you can check your traces / extraction positions as follows:
;         IDL> lrisplottraces
;         IDL> lrisplotprofiles
;   Or, to modify and view them at the same time:
;         IDL> setapertures
;
;
; -----------------------------------------------------------------------------

; ----
; Long list of desired features for future implementation:
; -----
; wavlength calibrate 2D files
; response summary plot
; replace the common blocks with structures
; may need to add a match step after sum to keep B and R files aware of each other.
;   e.g. lrisplotprofiles aligns differently from autolrisconnect.
; determine S/N when doing the tracing
; readd sky in specextract (and then resubtract out of background aperture)
; imaging zapper zaps the edges sometimes - 150715_0151r
; multiple data-dir
; check-related files into check directory
; split step actually causes blue to be overcropped!! - fixed?
; "pre-astrometry": check for regular offset shift
; matching of apertures for multiple reobservations and red/blue...?
; use supersky reflat constructed using multiple filters if aren't enough images in one?
; imaging thumbnails
; need to find a way to unify r/l chip photometric catalogs to use both in air solution / scaling - calling of autolrisphot is already setup
; implement new lamps
; do wavelength calibration in 2D (or at least write basic solution to 2D spectra)
; bias frames
; use X's pixel flats
; evaluate if photometric or not
; lots more check plots
; archival response curves and airmass adjustment
; gain correction
; check that USNO catalogs are loaded OK
; check for user-specified star catalogs for calibration
; need an option for skipping spectroscopic flats entirely
; use system variables to make life easier
; old red side re-enable - mostly done, needs some checks (e.g. r+l imaging reconnection)
; better text output; error/warning log written to disk
; background subtraction neighbor recognition and exclusion
; consistent aperture size for repeated observations?  Or, stack in 2D space?
; automated halo removal for bright stars?
; option to make flats that aren't needed if files exist (beginning of night, compare dome vs. sky, etc.)
; during flux calibration:  re-normalize consistently for proper airmass interpolation/extrapolation.
; interpolate standards in log space
; split r and l into multi-extension files in prepare and enable ccd/ files to recognize this (or not; has downsides)
; not sure if bpm is working, still a bad line on red side (when binned, anyway)
;  this appears to be due to movement of the weakened columns.  Make sure this isn't a crop problem and 
;  if not try to do something to detect this...
; need better sky subtraction for twilight blue spectra (try a poly model with lines removed)
; future options
;  (nofringe)    - Skip fringe production and correction for LRIS-R1
; specify files=100-102, or filter='B', or grating='400/8000', etc. - partially done but needs fixing
;  gratings=
;  filters=
;  files=
; simple median 'counts' from lrisprepare not appropriate for some gratings/dichroics
; need a new method for the slit-averaged flats - maybe linear extrapolation fit

; overwrite prompt in photometry - fixed
; standard function to convert targname to unix-friendly file string - done
; use good standards where possible (e.g., don't use BD+26 for blue) - done
; use standard stars to guess trace pos - done
; mask specific lines for specific flux standards - done?
; common LRIS header reader function - mostly done
; faster CR cleaner for blue (and old red): done?
; use region of best mutual response for red/blue connection - done
; replace photometric solver with generic one - done
; prevent zapping causing interpolation of edge regions. - done
; telluric correction - done
; probably want a global arc solution, rather than each exposure (more robust) - done
; need sky-line refined wavelength solution - done
; deal with flats taken in wrong binning mode - done
; recognize mira files - done
; need to recognize saturated standards - done
; cleanup of autoastrometry files - done
; standard quality - red vs. blue - done
; done - superskyflats looking at mirror settings to better remove dust rings and glue defect (on red)
; done - selectfile messes up the order on multi-night runs.
; deal with when object name isn't read in properly



; -------------------------
pro autolrissetup

   ; The first time the code is run in a directory, lpipe.par is copied to the local directory and
   ; the program immediately returns to the command line.  Since the default trace position has moved 
   ; around runs this allows the user to specify this easily.
   ; (However, this does lock the config file against pipeline upgrades, and automatic trace default
   ;  measurement is a desired future feature.)

   ; Now obsolete due to added trace-finding step.

   islocalfile = file_test('lpipe.par')

   if islocalfile eq 0 then begin
      print, '  LPIPE - LRIS automatic processing pipeline'
      print, '          version 2016.12'
      print, '          authored by Daniel Perley (dperley@astro.ljmu.ac.uk)'

      paths = strsplit(!path,':',/extract)
      for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
         if p ge 0 then path = paths[p] $
                        else path = ''
         sfile = path+'/lpipe.pro'
         n = countfiles(sfile)
         if n gt 0 then begin
           defaultparfile = path+'/lpipe.par'
           if file_test(defaultparfile) eq 0 then defaultparfile = path+'/defaults/lpipe.par'
           if file_test(defaultparfile) then begin
              spawn, 'cp '+defaultparfile+' ./' 
              break
           endif
         endif
      endfor

      if file_test('lpipe.par') then begin
         print, '  To edit configuration for this reduction, modify ./lpipe.par'
         print, '  To edit default configuration for future reductions, modify '+defaultparfile
         print, '  To run the pipeline, re-run the command.'
      endif else begin
         print, '  ERROR: Unable to find default parameter file.  Installation problem?'
      endelse
      retall
   endif

end



pro autolrisdefaults, modes=modes

   ; Load the configuration file that sets the external commands, directory locations, etc.
   ; If no local file is found it will look in the code directory.
   ; If no configuration file is found at all, internal defaults are available.
   ; (Neither of these two features are strictly necessary with the new setup implementation.
   ; Also creates the imredux/ and spredux/ subdirectories if necessary.

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common landoltpathinfo, landoltpath

   paths = strsplit(!path,':',/extract)
   for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
      if p ge 0 then path = paths[p] $
                     else path = ''
      sfile = path+'/lpipe.pro'
      n = countfiles(sfile)
      if n gt 0 then lrisautopath = path
      if n gt 0 then break
   endfor
   if n eq 0 then begin
      print, 'Problem finding lpipe directory.  This error should not occur.'
   endif
   for p = -1, 1 do begin
      if p eq -1 then path = '.'
      if p eq  0 then path = lrisautopath
      if p eq  1 then path = lrisautopath + '/defaults'
      sfile = path+'/lpipe.par'       ; get the configuration file
      n = countfiles(sfile)
      if n gt 0 then begin
         openr, 5, sfile
         iline = ''
         while not eof(5) do begin
            readf, 5, iline
            hashpos = strpos(iline,'#')
            if hashpos ge 0 then iline = strmid(iline,0,hashpos)
            colonpos = strpos(iline,':')
            if colonpos lt 1 or colonpos ge strlen(iline) then continue
            varname = clip(strmid(iline,0,colonpos))
            varset = clip(strmid(iline,colonpos+1))
            if varname eq 'refdatadir' then refdatapath = varset
            if varname eq 'defaultsdir' then defaultspath = varset
            if varname eq 'autoastrocommand' then autoastrocommand = varset
            if varname eq 'swarpcommand' then swarpcommand = varset
            if varname eq 'sexcommand' then sexcommand = varset
            if varname eq 'datadir' and n_elements(datadir) eq 0 then datadir = varset ; command line overrides file value
            if varname eq 'imworkingdir' then imworkingdir = varset
            if varname eq 'spworkingdir' then spworkingdir2d = varset
            if varname eq 'spworkingdir' then spworkingdir1d = varset
            if varname eq 'spworkingdir2d' then spworkingdir2d = varset
            if varname eq 'spworkingdir1d' then spworkingdir1d = varset
            if varname eq 'imoutputdir'  then imoutputdir = varset
            if varname eq 'spoutputdir'  then spoutputdir = varset
            if varname eq 'bluetracepos' then defaulttraceyblue = float(varset)
            if varname eq 'redtracepos' then defaulttraceyred = float(varset)
         endwhile
         close, 5
;         print 'Using parameters in ', sfile
         break ; found the file: stop looping
      endif
   endfor

   ; use internal defaults to set anything still unspecified
   if n_elements(refdatapath) eq 0 then refdatapath = lrisautopath+'/refdata'
   if n_elements(defaultspath) eq 0 then defaultspath = lrisautopath+'/defaults'
   if n_elements(autoastrocommand) eq 0 then autoastrocommand = 'autoastrometry'
   if n_elements(swarpcommand) eq 0 then swarpcommand = 'swarp'
   if n_elements(sexcommand) eq 0 then sexcommand = 'sex'
   if n_elements(datadir) eq 0 then datadir = '' 
   if n_elements(imworkingdir) eq 0 then imworkingdir = ''
   if n_elements(spworkingdir2d) eq 0 then spworkingdir2d = ''
   if n_elements(spworkingdir1d) eq 0 then spworkingdir1d = ''
   if n_elements(imfinaldir) eq 0 then imfinaldir = ''
   if n_elements(spfinaldir) eq 0 then spfinaldir = ''
   if n_elements(wildchar) eq 0 then wildchar = '??????_????'
   if n_elements(overwrite) eq 0 then overwrite = 0
   if n_elements(defaulttraceyblue) eq 0 then defaulttraceyblue = 0.     ; Now defaults to zero and the pipeline will   
   if n_elements(defaulttraceyred) eq 0 then defaulttraceyred = 0.       ; calculate it using the actual data later.
   ;if defaulttraceyblue le 0 then defaulttraceyblue = 140.     
   ;if defaulttraceyred le 0 then defaulttraceyred = 95.

   setswarppath, swarpcommand   ;extractpath(swarpcommand)
   setsexpath, sexcommand       ;extractpath(sexcommand)
   landoltpath = refdatapath+'/'

   ; these must always start blank
   prepfail = ['']
   flatfail = ['']
   catastrofail = ['']
   relastrofail = ['']
   fullastrofail = ['']
   extractfail = ['']
   wavsolfail = ['']
   wavsolwarn = ['']
   fluxcalfail = ['']

  if total(modes eq 'im') ge 1 and strlen(imworkingdir) gt 0 and file_test(imworkingdir) eq 0 then begin
     print, 'Creating imaging working directory ', imworkingdir
     spawn, 'mkdir '+imworkingdir
     check = file_test(imworkingdir)
     if check eq 0 then begin 
       print, 'WARNING - Failed to create image working directory!'
       print, '          Using current directory.  Check lpipe.par inputs.'
     endif
  endif
  if total(modes eq 'sp') ge 1 and strlen(spworkingdir2d) gt 0 and file_test(spworkingdir2d) eq 0 then begin
     print, 'Creating spectroscopy working directory ', spworkingdir2d
     spawn, 'mkdir '+spworkingdir2d
     check = file_test(spworkingdir2d)
     if check eq 0 then begin 
       print, 'WARNING - Failed to create spectrum working directory!'
       print, '          Using current directory.  Check lpipe.par inputs.'
     endif
  endif
  if total(modes eq 'sp') ge 1 and strlen(spworkingdir1d) gt 0 and file_test(spworkingdir1d) eq 0 then begin
     print, 'Creating spectroscopy working directory ', spworkingdir1d
     spawn, 'mkdir '+spworkingdir1d
     check = file_test(spworkingdir1d)
     if check eq 0 then begin 
       print, 'WARNING - Failed to create spectrum working directory!'
       print, '          Using current directory.  Check lpipe.par inputs.'
     endif
  endif

  displayoff = 0
end


function readseq, h
   
    ; Get some specific information out of the header regarding the
    ; configuration of the image, including when the grating/filter was
    ; last moved.   (Used by the reflat procedure.)

    seq = {dich:'', rfilt:'', bfilt:'', grat:'', grangle:0., $
           dichf:'', rfiltf:'', bfiltf:'', gratf:'', granglef:''}
    seq.dich = sxpar(h, 'DICHNAME')
    seq.rfilt = sxpar(h,'REDFILT')
    seq.bfilt = sxpar(h, 'BLUFILT')
    seq.grat = sxpar(h, 'GRANAME')
    seq.grangle = sxpar(h, 'GRANGLE')
    dum = sxpar(h, 'DICHSEQF', count=ct) 
    if ct gt 0 then begin
      seq.dichf = sxpar(h, 'DICHSEQF')
      seq.rfiltf = sxpar(h, 'RFLTSEQF')
      seq.bfiltf = sxpar(h, 'BFLTSEQF')
      seq.gratf = sxpar(h, 'GRATSEQF')
      seq.granglef = sxpar(h, 'GRANSEQF')
    endif

    return, seq
end

pro updateseq, seq, prevseq, seqfile

    ; Used by the reflat procedure to look for moves of the gratings and filters.

    if seq.dich eq prevseq.dich then seq.dichf = prevseq.dichf else seq.dichf = seqfile
    if seq.rfilt eq prevseq.rfilt then seq.rfiltf = prevseq.rfiltf else seq.rfiltf = seqfile
    if seq.bfilt eq prevseq.bfilt then seq.bfiltf = prevseq.bfiltf else seq.bfiltf = seqfile
    if seq.grat eq prevseq.grat then seq.gratf = prevseq.gratf else seq.gratf = seqfile
    if seq.grangle eq prevseq.grangle then seq.granglef = prevseq.granglef else seq.granglef = seqfile
end

; -------------------------
pro autolrisprepare, modestr=modestr, camstr=camstr
   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

  ; Prepare images for processing - debias, crop, fix some bad pixels, add header information,
  ; and merge the separate extensions into a simple fits file - see lrisprepare.pro for details.

  mosworkingdir = 'mosredux/'

  if n_elements(modestr) gt 0 then begin
     if modestr eq 'is' or modestr eq 'i,s' or modestr eq 'im,sp'  then mode = ''
     if modestr eq 'si' or modestr eq 's,i' or modestr eq 'sp,im'  then mode = ''
     if n_elements(mode) eq 0 then mode = strmid(modestr,0,1)
  endif else begin
     mode = ''
  endelse

  prefchar = strmid(camstr,0,1)
 
  files =        file_search(datadir+    prefchar+wildchar+'.fits')
  pfiles = [file_search(imworkingdir+'p'+prefchar+wildchar+'.fits'), $
            file_search(spworkingdir2d+'p'+prefchar+wildchar+'.fits'), $
            file_search(mosworkingdir+'p'+prefchar+wildchar+'.fits')]
  nfiles = n_elements(files)
  if nfiles eq 1 then if files[0] eq '' then nfiles = 0

  if datadir ne '' then  begin
     logprint, 'Looking for raw data at: ', datadir+    prefchar+wildchar+'.fits'
     if nfiles gt 0 then begin
        logprint, '   Found ', clip(nfiles), ' files'
     endif else begin
        logprint, '   Did not find any files!  Check your data directory path!'
     endelse
  endif

  badfiles = ['']
  badfilefile = 'badfiles.txt'  ; ignorefiles.txt would be a better name?
  if file_test(badfilefile) then readcol, badfilefile, badfiles, format='a'
  ;'~/research/redux/allgrbpos.txt' 
  ;'~/research/redux/standardpos.txt'

  namefixfiles = ['']
  catalogfile = 'catalog.txt' ; current directory
  if file_test(catalogfile) then namefixfiles = [namefixfiles, catalogfile]
  catalogfile = 'starlist.txt'
  if file_test(catalogfile) then namefixfiles = [namefixfiles, catalogfile]
  landoltfieldposfile = refdatapath+'/landoltpos.txt'
  if file_test(landoltfieldposfile) then namefixfiles = [namefixfiles, landoltfieldposfile]
  specstandardposfile = refdatapath+'/specstandards.dat'
  if file_test(specstandardposfile) then namefixfiles = [namefixfiles, specstandardposfile]
  if n_elements(namefixfiles) gt 1 then namefixfiles = namefixfiles[1:*] else delvarx, namefixfiles

   prevtarg = ''
   prevobj = ''
   prevradeg = 0.
   prevdecdeg = 0.

  for f = 0, n_elements(files)-1 do begin
     if files[f] eq '' then continue
     if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue

     dum = where(files[f] eq badfiles, ct)
     if ct gt 0 then continue
    
     fileroot = removepath(files[f])
     outnameim = imworkingdir + 'p' + fileroot
     outnamesp = spworkingdir2d + 'p' + fileroot
     outnamemos = mosworkingdir + 'p' + fileroot
     matchi = where(outnameim eq pfiles, ct1)
     matchs = where(outnamesp eq pfiles, ct2)
     matchm = where(outnamemos eq pfiles, ct3)
     ct = ct1+ct2+ct3
     if ct eq 0 or overwrite gt 0 then begin
        h = headfits(files[f], /silent)

        ; Store in the header the name of the first file in which the current configuration was set.
        ; This is useful for identifying a block of files in the same config 
        ; (since LRIS doesn't always return to the same position exactly) - currently the reflat
        ; procedure uses this to limit blocks of supersky flats, and the spectroscopic extractions
        ; could also use it to optimize the choice of standard frame.
        ; This should be robust to re-runs and unprepared files, etc.
        ; However, if that first file is deleted in between program runs, this will cause files 
        ; processed later to be assigned different identifiers.
        if n_elements(seq) gt 0 then begin
           prevseq = seq
           prevfilef = where(files eq prevfile)
        endif else begin
           prevfilef = -999
        endelse
        seq = readseq(h)
        if f eq prevfilef + 1 then begin
           ; No breaks since last time seq was populated; just check against last file.
           updateseq, seq, prevseq, files[f]
        endif else begin
           ; break in sequence.  Have to look back at the most recent processed file and then run forward from there.
           ;print, 'Break in sequence before ', files[f]
           oldreffile = ''
           for fn = f-1, 0, -1 do begin  ; Find the most recent processed file.
             prevoutnameim = imworkingdir + 'p' + removepath(files[fn])
             prevoutnamesp = imworkingdir + 'p' + removepath(files[fn])
             if file_test(prevoutnameim) then oldreffile = prevoutnameim
             if file_test(prevoutnamesp) then oldreffile = prevoutnamesp
             if oldreffile ne '' then break
           endfor
           if oldreffile ne '' then begin
             ;print, 'Found an earlier file:', oldreffile
             oldseq = readseq(headfits(oldreffile))  ; Read it in.
             prevoldseq = oldseq
             for fn = fn+1, f-1 do begin     ; Iterate forward, updating oldseq until the previous file is reached.
                ;;print, '   ', files[fn]
                oldh = headfits(files[fn])
                oldseq = readseq(oldh)
                updateseq, oldseq, prevoldseq, files[fn]
             endfor
             prevseq = oldseq
          endif else begin                 ; There are no earlier processed files - read in a blank seq.
             prevseq = {dich:'', rfilt:'', bfilt:'', grat:'', grangle:0., $
                        dichf:'', rfiltf:'', bfiltf:'', gratf:'', granglef:''}
          endelse
          updateseq, seq, prevseq, files[f]
        endelse
        prevfile = files[f]

        ; slitname = strtrim(sxpar(h, 'SLITNAME'),2)
        pane = strtrim(sxpar(h, 'PANE'),2)
        window = strtrim(sxpar(h, 'WINDOW'),2)
        slit =  clip(strlowcase(clip(sxpar(h, 'SLITNAME', count=cts))))
        if prefchar eq 'b' then grname = clip(sxpar(h,'GRISNAME'))
        if prefchar eq 'r' then grname = clip(sxpar(h,'GRANAME'))
        filemode = ''
        if slit eq 'direct' and strpos(grname,'/') eq -1 then filemode = 'i'   ; should probably put into readlrisheaders
        if strpos(slit,'long') ge 0  and strpos(grname,'/') ge 0 then filemode = 's' ; longslit and grating
        if strpos(slit,'long') eq -1 and strpos(grname,'/') ge 0 then filemode = 'm' ; non-longslit and grating
        if filemode ne 'i' and filemode ne 's' and filemode ne 'm' then begin
           if cts eq 0 then begin
                 logprint, files[f], ': SLITNAME keyword absent; cannot distinguish image/spectrum; skipping...' 
           endif else begin
                 logprint, files[f], ': unusual combination of slit ('+clip(slit)+') and disperser ('+clip(grname)+'); skipping...'
           endelse
           ;if slit eq 'direct' and strpos(grname,'/') ge 0 then $
           ;      print, files[f], ' Spectroscopic slitless flats not supported yet, skipping for now...'
           prepfail = [prepfail, files[f]]
           continue
        endif

        ; If the user forgot to change the object name but moved to clearly different targets, update using TARGNAME
        if n_elements(newobj) gt 0 then delvarx, newobj        
        obj = clip(sxpar(h,'OBJECT'))
        targ = clip(sxpar(h,'TARGNAME'))
        radeg = 15*ten(clip(sxpar(h,'RA')))
        decdeg = ten(clip(sxpar(h,'DEC')))
        ltarg = strlowcase(targ)
        ;trapdoor = clip(sxpar(h,'TRAPDOOR'))
        if strpos(ltarg,'flat') eq -1 and strpos(ltarg,'stow') eq -1 and $            ; trapdoor eq 'open' and 
           strpos(ltarg,'dome') eq -1 and strpos(ltarg,'horizon') eq -1 and ltarg ne '' then begin
            if obj eq prevobj and targ ne prevtarg and (abs(decdeg-prevdecdeg) gt 0.1 or abs(radeg-prevradeg) gt 0.3) then begin
              newobj = (strsplit(targ,'_',/extract)) [0]
            endif else begin
              prevtarg = targ
              prevradeg = radeg
              prevdecdeg = decdeg
            endelse
          prevobj = obj
        endif

        if mode eq 'i' and filemode ne 'i' then continue
        if mode eq 's' and filemode ne 's' and filemode ne 'm' then continue
        if filemode eq 'i' then outname = outnameim
        if filemode eq 's' then outname = outnamesp
        if filemode eq 'm' then outname = outnamemos
        if file_test(mosworkingdir) eq 0 then spawn, 'mkdir '+mosworkingdir

        ;print, 'Preparing ', files[f]
;        catch, errorstatus
errorstatus = 0
        if errorstatus ne 0 then begin
           logprint, '   Could not read in file (unexpected error):'
           logprint, '      '+!error_state.msg
           ;print, '   Ignoring this file, consider adding to badfiles.txt'
           prepfail = [prepfail, files[f]]
           success = -10
        endif else begin
          lrisprepare, files[f], gain=[1,1,1,1], /merge, crop='auto', $
            outname=outname, namefixfiles=namefixfiles, seq=seq, /requirecompleteh, newobj=newobj, reducer=reducer, version=lpversion
          if file_test(outname) eq 0 then prepfail = [prepfail, files[f]]
        endelse
;        catch, /cancel 

     endif
  endfor

end




; -------------------------
pro autolrismakeimflat, chip=chip, camera=camera

   ; Construct imaging flatfields.   Looks at what science images actually need flat-fielding, then looks for flats
   ; in the appropriate configurations and combines them, choosing dome or sky based on how many of each or available.

   ; Currently it constructs a flat even if the ability to do so is rather marginal.  Ideally in these cases it should use
   ; a set of archival flats kept in a special directory (in fact it's quite possible that this would be better in general.)

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

    if strlen(chip) gt 1 then fchip = strmid(chip,0,1) else fchip = chip

    prefchar = strmid(camera,0,1)
    if prefchar eq 'r' and lrisversion eq 1 then xnormkey = 'AMP1' else xnormkey = 'AMPR1'

    files = file_search(imworkingdir+'p'+prefchar+wildchar+fchip+'.fits') ; camera dealt with here
    imh = readlrisheaders(files)
    w = where(imh.slit eq 'direct', ct)
    if ct eq 0 then return
    files = files[w]
    imh   = imh[w]
    isdomeflat = intarr(ct)
    isskyflat  = intarr(ct)
    isscience  = intarr(ct) 

    for f = 0, n_elements(files)-1 do begin
       if files[f] eq '' then continue
       if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
       h = headfits(files[f], /silent)

       if imh[f].trapdoor eq 'open' then begin    ;abs(domerelaz-90) lt 1 or abs(domerelaz+90) lt 1))
         if (abs(imh[f].elevation-45) lt 0.5 and abs(abs(imh[f].domerelaz)-90) lt 3) or strpos(imh[f].target,'cass') ge 0 then begin
             isdomeflat[f] = 1  ;if counts[f] gt 7000 and counts[f] lt 40000 then 
         endif else begin
             if ((imh[f].skycounts gt 9000 and camera eq 'blue') or $
                 (imh[f].skycounts gt 3500 and imh[f].filter eq 'U') or $  
                 (imh[f].skycounts gt 3500 and clip(imh[f].filter,2) eq 'NB') or $  ; narrowband
                 (imh[f].skycounts gt 6000 and imh[f].filter eq 'B') or $
                 (imh[f].skycounts gt 6000 and imh[f].skycounts/imh[f].exp gt 450. and camera eq 'red')    $ 
                ) and $
                  imh[f].skycounts lt 43000 and imh[f].exp le 90 and abs(imh[f].domerelaz) lt 3 $ 
                then isskyflat[f] = 1
             if abs(imh[f].domerelaz) lt 5.0 and ((imh[f].skycounts lt 40000) and (imh[f].skycounts le 15000 or (imh[f].filter eq 'I' and imh[f].skycounts le 30000) or (imh[f].filter eq 'RG850') or (imh[f].exp ge 60))) then isscience[f] = 1
         endelse
       endif

      if keyword_set(debug) then print, files[f], ' ', clip(imh[f].target,12), string(round(imh[f].exp),format='(I4)'), ' ', clip(imh[f].filter,5), ' ', clip(imh[f].dichroic,4), ' ', clip(imh[f].bin,3), '  ', rclip(fix(imh[f].skycounts),5), ' ', clip(imh[f].skycounts/imh[f].exp,7), ' ', fpr(imh[f].domerelaz,3.2), ' ', fpr(imh[f].elevation,2.2), '  ', isdomeflat[f] ? 'dom':'   ',' ', isskyflat[f]?'twi':'   ',' ',isscience[f]?'sci':'   '

    endfor


    sciindex = where(isscience, ct)
    if ct eq 0 then return
    scifilterlist  = unique(imh[sciindex].filter)
    for i = 0, n_elements(scifilterlist)-1 do begin
      filt = scifilterlist[i]

      dichlist = unique(imh[sciindex[where(filt eq imh[sciindex].filter)]].dichroic)
      for d = 0, n_elements(dichlist)-1 do begin
        dich = dichlist[d]
        if n_elements(dichlist) gt 1 then dichstr = dichlist[d] else dichstr = ''

        binlist = unique(imh[sciindex[where(filt eq imh[sciindex].filter and dich eq imh[sciindex].dichroic)]].bin)
        for b = 0, n_elements(binlist)-1 do begin
          bin = binlist[b]
          if n_elements(binlist) gt 1 then binstr = 'b'+binlist[b] else binstr = ''

        winlist = unique(imh[sciindex[where(filt eq imh[sciindex].filter and dich eq imh[sciindex].dichroic and bin eq imh[sciindex].bin)]].win)
        for wi = 0, n_elements(winlist)-1 do begin
          win = winlist[wi]
          if n_elements(winlist) gt 1 then winstr = 'w'+winlist[wi] else winstr = ''        

          outflatname = imworkingdir+'lris'+strmid(camera,0,1)+'flat'+filt+dichstr+binstr+winstr+fchip+'.fits'
          if file_test(outflatname) and overwrite eq 0 then continue ; flat exists already

          flatsuccess = 0

          ;print, filt, ' ', dich, ' ', bin
        
         ; Regular sky flat.
         if flatsuccess eq 0 then begin
            skyflats = where(isskyflat and imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win), ctsky)
            if ctsky ge 5 then begin
               logprint, filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsky),2), ' sky flats.'
               flatcombine, files[skyflats], outflatname, /removeobjects, type='sky', xnormkey=xnormkey ;  normzone=normzone
               flatsuccess = 1
            endif
         endif

        ; Dome flat.
        if flatsuccess eq 0 then begin
           if strmid(camera,0,1) eq 'b' then mincounts=900. else mincounts=1000.  ; formerly 5000 and 7000
           domeflats = where(isdomeflat and imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and $
                             imh.skycounts gt mincounts and imh.skycounts lt 44000., ctdome)
           if ctdome gt 5 then domeflats = domeflats[where(imh[domeflats].skycounts gt ( median(imh[domeflats].skycounts, /even)*0.6) < 9000.)]
           if ctdome ge 3 then begin
              logprint, filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctdome),2), ' dome flats'
              flatcombine, files[domeflats], outflatname, type='dome', xnormkey=xnormkey ; normzone
              flatsuccess = 1
           endif
        endif

         ; "Good" supersky flat [now avoided - supersky correction should be done in "reflat" portion]
        ctsupersky = 0
        if flatsuccess eq 0 and (lrisversion eq 1 and camera eq 'red' and (filt eq 'I' or filt eq 'RG850' or filt eq 'GG570')) ne 1 then begin ; don't supersky fringed frames
           ; First, see if we have enough frames to do a supersky flat.
            poss = where(imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and imh.skycounts gt 300 and imh.exp gt 10., ctposs)
            if ctposs gt 0 and filt ne 'U' then begin ;otherwise usually means some weird situation like the dichroic info is missing, etc.
               if ctposs gt 1 then $
                  medcounts = median(imh[poss].skycounts) $
               else $
                  medcounts = imh[poss].skycounts
               if camera eq 'red' then superskymincount = 1000 > medcounts/3.
               if camera eq 'blue' then superskymincount = 350 > medcounts/3.
               superskymaxcount = (medcounts * 5) < (30000. > medcounts*1.2) ; 
               superskyminexp = 10
               
               ;;;if filt eq 'g' or filt eq 'V' or filt eq 'R' then $
               superskyflats = where(imh.filter eq filt and imh.dichroic eq dich and imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and $
                                     imh.skycounts gt superskymincount and imh.skycounts lt superskymaxcount and imh.exp ge superskyminexp, ctsupersky)        ;    else  $
               ;superskyflats = where(imh.filter eq filt and                   imh.bin eq bin and wincontain(win, imh.win) and isdomeflat eq 0 and $
               ;                      imh.skycounts gt superskymincount and imh.skycounts lt superskymaxcount and imh.exp ge superskyminexp, ctsupersky)
               ;print, superskymincount, ' - ', superskymaxcount
               ;for i = 0, ctsupersky-1 do print, '  ', clip(removepath(files[superskyflats[i]]),20), clip(filters[superskyflats[i]],3), clip(imh[superskyflats[i]].dichroic,5), fix(counts[superskyflats[i]]), ' ', targets[superskyflats[i]]
               

 
               ; Exclude extended sources or anything else labeled as a bad flat

               extendedsources = [''] 
               if file_test('extendedsources.txt') then readcol, 'extendedsources.txt', extendedsources, format='a', comment='#', /silent
               badflattargets = [''] 
               if file_test('badflattargets.txt') then  readcol, 'badflattargets.txt',  badflattargets,  format='a', comment='#', /silent
               badflattargets = unique([badflattargets, extendedsources])
               notblank = where(badflattargets ne '', ct)
               if ct gt 0 then badflattargets = badflattargets[notblank]


               for bad = 0, n_elements(badflattargets)-1 do begin
                  if badflattargets[bad] eq '' then continue
                  w = where(imh[superskyflats].target ne badflattargets[bad], ctsupersky)
                  superskyflats = superskyflats[w]
               endfor
  
               if ctsupersky gt 1 then superskyfields = unique(imh[superskyflats].target) else superskyfields = ['']
               nsuperskyfields = n_elements(superskyfields)
               ; this criterion is a bit loose and depends on the number of very bright stars in the image and how big the dither steps were.
               ; eventually, should make flatcombine able to find bright (ultra-saturated) objects and add halo masks, and 
               ; treat big dithers like new fields.
               ;print, ctsupersky, nsuperskyfields
               ;if (ctsupersky ge  9 and nsuperskyfields ge 5) or (ctsupersky ge 17 and nsuperskyfields ge 4) or $
               ;   (ctsupersky gt 17 and nsuperskyfields gt 3) or  ctsupersky ge 21 then begin
               if (ctsupersky ge 21 and nsuperskyfields ge 5) or (ctsupersky ge 25 and nsuperskyfields ge 4) or $
                 (ctsupersky gt 29 and nsuperskyfields gt 3) or ctsupersky ge 33 then begin
                  logprint, filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsupersky),2), ' supersky flats (', clip(nsuperskyfields), ' unique fields)'
                  flatcombine, files[superskyflats], outflatname, /removeobjects, type='supersky', xnormkey=xnormkey
                  flatsuccess = 1
               endif
            endif
         endif

        ; marginal sky flat
        if flatsuccess eq 0 then begin
            logprint, 'WARNING - insufficient images for a good flat.  ('+clip(ctdome)+' dome, '+clip(ctsky)+' sky, '+clip(ctsupersky)+' science).'
            if ctsky ge 3 then begin
               logprint, filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsky),2), ' sky flats (WARNING - marginal!)'
               flatcombine, files[skyflats], outflatname, /removeobjects, type='sky', xnormkey=xnormkey
               flatsuccess = 1
            endif
        endif

        ; marginal supersky flat
        if flatsuccess eq 0 then begin
            if ctsupersky ge 5 then begin
               logprint, filt, '-band / ', 'D'+dich,' '+binstr+': ', strtrim(string(ctsupersky),2), ' supersky flats  (WARNING - EXTREMELY marginal)'
               logprint, 'Flat field for this filter/dichroic combination is likely to be extremely poor.'
               flatcombine, files[superskyflats], outflatname, /removeobjects, type='supersky', xnormkey=xnormkey
               flatsuccess = 1
            endif
        endif

        if flatsuccess eq 0 then begin
           logprint, 'Unable to produce a flat field for this setting: ', filt, '-band / ', 'D'+dich,' '+binstr+' '+winstr
           noproc = where(isscience and imh.filter eq filt and imh.dichroic eq dich and imh.binning eq bin and wincontain(win, wins), ctnoproc)
           if ctnoproc gt 0 then begin
              logprint, 'Will not be able to further process ', clip(ctnoproc), ' files without a flat from another source.'
              if ctnoproc lt 6 then $
              for j = 0, ctnoproc-1 do logprint, '   ', files[noproc[j]], imh[noproc[j]].exp, '  ', imh[noproc[j]].target
                   ; this seems to actually miss some sometimes, as shown 2013-05-10
           endif
        endif

       endfor
       endfor
     endfor  
  endfor


end






; -------------

pro autolrismakespecflat, chip=chip, camera=camera

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

    ; Construct spectroscopic flatfields.  Always uses dome flats unless "desparate", doesn't recognize 
    ; the new lamps yet or any sky flats.  Does a y-axis medianing to remove the wavelength response - this
    ; does leave residuals of the telluric lines (from air in the dome) and emission lines in the lamp.

    fchip = strmid(chip,0,1)
    prefchar = strmid(camera,0,1)

    files = file_search(spworkingdir2d+'p'+prefchar+wildchar+fchip+'.fits')

    imh = readlrisheaders(files)
    w = where(imh.slit ne 'direct', ct)
    if ct eq 0 then return
    files = files[w]
    imh   = imh[w]
    isdomeflat = intarr(ct)
    ishalogen  = intarr(ct)
    isskyflat  = intarr(ct)
    isscience  = intarr(ct) 

   for f = 0, n_elements(files)-1 do begin
       if files[f] eq '' then continue
       if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue

       if imh[f].trapdoor eq 'open' then begin 
         if (abs(imh[f].elevation-45) lt 0.5 and abs(imh[f].domerelaz) gt 3) or strpos(imh[f].target,'cass') ge 0 then begin
             isdomeflat[f] = 1  ;if counts[f] gt 7000 and counts[f] lt 40000 then 
         endif else begin
             if imh[f].skycounts gt 10000 and imh[f].skycounts lt 43000 and imh[f].exp lt 300 then isskyflat[f] = 1
             if abs(imh[f].domerelaz) lt 5.0 and (imh[f].skycounts lt 1000) then isscience[f] = 1
         endelse
       endif

       if imh[f].skycounts gt 500 and imh[f].skycounts lt 43000 and imh[f].trapdoor eq 'closed' and strmid(imh[f].lamps,10,1) eq '1' then ishalogen[f] = 1

       ;if isdomeflat[f] then $
       ;print, files[f], ' ', clip(imh[f].target,12), string(imh[f].exp,format='(I4)'), ' ', clip(imh[f].grname,9), ' ', imh[f].skycounts, imh[f].domerelaz, ' ', clip(imh[f].trapdoor,6), ' ', imh[f].lamps, ' ', clip(isdomeflat[f]), ' ', clip(ishalogen[f])
  
   endfor

   sciindex = where(isscience, ct)

   if ct eq 0 then return
   scigratinglist  = unique(imh[sciindex].grname)
   if size(scigratinglist,/tn) eq 'INT' then begin
       logprint, '  WARNING: Namespace issue, do you have conflicting packages installed?'
   endif
   for i = 0, n_elements(scigratinglist)-1 do begin
      grat = clip(scigratinglist[i])  ; the clip shouldn't be necessary
      gratstr = repstr(grat,'/','-')
      dichlist = unique(imh[sciindex[where(grat eq imh[sciindex].grname)]].dichroic)
      for d = 0, n_elements(dichlist)-1 do begin
        dich = dichlist[d]
        if n_elements(dichlist) gt 1 then dichstr = 'D'+dichlist[d] else dichstr = ''
        binlist = unique(imh[sciindex[where(grat eq imh[sciindex].grname and dich eq imh[sciindex].dichroic)]].bin)
        for b = 0, n_elements(binlist)-1 do begin
          bin = binlist[b]
          if n_elements(binlist) gt 1 then binstr = 'b'+binlist[b] else binstr = ''
          outflatname = spworkingdir2d+'lris'+strmid(camera,0,1)+'flat'+gratstr+dichstr+binstr+'.fits'
          ;print, 'Need ', outflatname
          if file_test(outflatname) and overwrite eq 0 then continue ; flat exists already
          flatsuccess = 0
          domeflats = where(isdomeflat and imh.grname eq grat and imh.dichroic eq dich and imh.bin eq bin and imh.skycounts gt 1000. and imh.skycounts lt 39000., ctdome)
          if ctdome eq 0 then begin
             logprint, 'No spectroscopic dome flats in this setting: ', scigratinglist[i], ' ', dichstr, ' ', binstr
             ; continue
          endif else begin

             dslits = imh[domeflats].slit
             wnot7 = where(imh[domeflats].slit ne 'long_0.7', ctnot7)
             if ct ge 3 then domeflats = domeflats[wnot7]
             if ct ge 3 then ctdome = ct

             dwindows = imh[domeflats].win
             winlist = unique(dwindows, count=nwin)
             mostcommonwin = winlist[(where(nwin eq max(nwin))) [0]] ; what if this is different from science windowing...?
             domeflats = domeflats[where(imh[domeflats].win eq mostcommonwin, ctdome)]
             if ctdome lt 3 then $
                domeflats = where(isdomeflat and imh.grname eq grat and imh.dichroic eq dich and imh.bin eq bin and imh.skycounts gt 100. and imh.skycounts lt 39000. and imh.win eq mostcommonwin, ctdome)



             ; THING TO CHANGE IN THE FUTURE - this counts estimate, the the hal one below, is not appropriate for the lowest-res blue grating. - fixed now?
             ;print, grat, ' ', dich,' ', bin, ' : ', ctdome
             if ctdome ge 3 then begin
                logprint, grat, ', D'+dich,': ', strtrim(string(ctdome),2), ' dome flats'
                h = headfits(files[domeflats[0]])
                if lrisversion eq 1 and camera eq 'red' then xdisc = (fix(strsplit(sxpar(h,'AMP2'),',',/extract))) [0] else xdisc = -1
                ; yresp0 = fix((strsplit(sxpar(h,'AMPR1'),',',/extract)) [0]) + 10  ; no longer used, it would seem
                ; yresp1 = yresp0 + 260
                ;if lrisversion eq 2 and camera eq 'red' then begin
                ;   yresp = findgen(580)  ; eeeew hard coded
                ;   yresp = yresp[where((yresp lt 182) or (yresp gt 256 and yresp lt 404) or (yresp gt 472 and yresp lt 577))]
                ;   flatcombine, files[domeflats], outflatname, type='dome', /response, yresp=yresp
                ;endif else 
               if camera eq 'blue' then begin
                  ; the blue side has some lines that a median response algorithm will miss 
                  ; finding and distinguishing from bad columns is tough, so just tell it where to search
                  ; needs to be updated for binned world and different gratings...
                  flatcombine, files[domeflats], outflatname, type='dome', /response, mincounts=10;;, lines=1399, linewidth=56  ; [1383,1412]
               endif else begin
                  flatcombine, files[domeflats], outflatname, type='dome', /response, mincounts=10, xdisc=xdisc
               endelse 
               flatsuccess = 1
             endif
          endelse

          if flatsuccess eq 0 then begin
            halflats = where(ishalogen and imh.grname eq grat and imh.dichroic eq dich and imh.skycounts gt 500. and imh.skycounts lt 39000., cthal)
            if cthal ge 2 then begin
               logprint, grat, ', D'+dich, ' ', bin, ': ', strtrim(string(cthal),2), ' halogen flats'
               xdisc = -1
               if lrisversion eq 1 and camera eq 'red' then begin
                 h = headfits(files[domeflats[0]])
                 xdisc = (fix(strsplit(sxpar(h,'AMP2'),',',/extract))) [0]
               endif
               logprint, 'WARNING - Halogen flats tend to be poor, next time get spectroscopic dome flats...'
               flatcombine, files[halflats], outflatname, type='halogen', /response, /indresp, /slitnorm, xdisc=xdisc
               flatsuccess = 1
            endif
          endif
          if flatsuccess eq 0 then begin
            logprint, 'Unable to produce a flat field for this combination: ', grat, ' & ', 'D'+dich+' ('+ bin + ').'
            noproc = where(isscience and imh.grname eq grat and imh.dichroic eq dich, ctnoproc)
            logprint, 'Will not be able to further process ', clip(ctnoproc), ' files without a flat from another source.'
            if ctnoproc lt 6 then $
               for j = 0, ctnoproc-1 do logprint, '   ', files[noproc[j]], imh[noproc[j]].exp, '  ', imh[noproc[j]].target
          endif

        endfor ; binning loop
      endfor   ; dichroic loop
    endfor     ; grating loop


end





; -------------------------
pro autolrisimflatten, chip=chip, camera=camera, ignoredichroic=ignoredichroic

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   fchip = strmid(chip,0,1)

   files = file_search(imworkingdir+'p'+camchar+wildchar+fchip+'.fits')
   if overwrite eq 0 then files = unmatched(files, 'f')
   if n_elements(files) eq 1 and files[0] eq '' then return

   ; un-flattened imaging files
   imh = readlrisheaders(files)
   w = where(imh.slit eq 'direct' and strmid(imh.camera,0,1) eq camchar and imh.exptime gt 0.5, ct)
   if ct eq 0 then return      ; return silently if nothing to process
   files = files[w]
   imh   = imh[w]

   ; flat field files
   flatfiles = file_search(imworkingdir+'*flat*.fits')
   flatimh   = readlrisheaders(flatfiles)
   w = where(flatimh.slit eq 'direct' and strmid(flatimh.camera,0,1) eq camchar, nflats)
   ws = where(abs(flatimh.elevation-45) ge 0.5 or abs(flatimh.domerelaz) gt 3, nsci)
   if nflats eq 0 and nsci gt 0 then begin   ; return loudly if something to process and no flats made yet
     logprint, 'No processed image flats found!  Cannot perform imaging flatfielding.'
     return
   endif
   flatfiles = flatfiles[w]
   flatimh   = flatimh[w]
   
   if keyword_set(ignoredichroic) then begin
      ; w = where(filter ne 'R' and filter ne 'V') could choose to require dichroic in these filters
      flatimh.dichroic = ''
      imh.dichroic = ''
   endif
  
   for f = 0, n_elements(files)-1 do begin
      if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
      outfile = fileappend(files[f], 'f')

      if imh[f].trapdoor eq 'closed' then continue    ; skip darks (and internal flats)

      if abs(imh[f].elevation-45) lt 0.5 and abs(imh[f].domerelaz) gt 3 then continue ; silently skip dome flats
      if strpos(imh[f].target,'cass') ge 0 then continue 
      if strpos(imh[f].target,'dome') ge 0 then continue     ; sometimes the telescope is pointed wrong; we still don't want these...
      if strpos(imh[f].target,'horizon') ge 0 then continue 

      if imh[f].skycounts gt 55000 then begin
         logprint,  '           ', files[f], ' is saturated.'
         continue
      endif

      ;;print, imh[f].skycounts, imh[f].target, imh[f].exptime, imh[f].skycounts gt 10000., strpos(strlowcase(imh[f].object), 'sky') ge 0

      ; We may actually want to keep these for use in the reflattening procedure 
      ;  (but perhaps mark them for further processing, or make this a function call.)
      if (imh[f].skycounts gt 30000. and imh[f].exptime lt 30) or $
         (imh[f].skycounts gt 20000. and imh[f].exptime lt 15 and imh[f].exptime gt 1.9) or $
         (imh[f].skycounts/imh[f].exptime gt 2000. and imh[f].filter ne 'RG850' and imh[f].exptime gt 1.9) or  $ 
         (imh[f].skycounts/imh[f].exptime gt 6000. and imh[f].filter eq 'RG850' and imh[f].exptime gt 1.9) or  $ 
         (imh[f].skycounts/imh[f].exptime gt 15000. and imh[f].exptime le 1.9) or $
         (imh[f].skycounts gt 10000. and imh[f].exptime gt 1.9 and imh[f].exptime lt 30 and $
          strpos(strlowcase(imh[f].object), 'flat') ge 0 or strpos(strlowcase(imh[f].object), 'sky') ge 0) or $
         (imh[f].skycounts gt 15000. and strpos(strlowcase(imh[f].target), 'flat') ge 0) or $
         (imh[f].skycounts gt 15000. and strpos(strlowcase(imh[f].target), 'sky') ge 0) then begin
            logprint, '           ', removepath(files[f]), '(',clip(imh[f].target), ', ', clipzero(imh[f].skycounts), 'ct, ', clip(round(imh[f].exptime)), 's)', $
                   ' is probably a sky flat: skipping'
            continue
      endif
      if imh[f].skycounts gt 40000. and imh[f].exptime gt 60 then begin
         if imh[f].satval gt 0 then satstr = '('+clip(imh[f].satval)+')' else satstr=''
         logprint, '  WARNING: ', files[f], ' has sky (', clip(round(imh[f].skycounts)), ') close to saturation '+satstr+'!'
      endif

      flatfileno = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic and  $
                         flatimh.binning eq imh[f].binning and flatimh.win eq imh[f].win, ct)
      if ct eq 0 then $
      flatfileno = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic and $
                         flatimh.binning eq imh[f].binning, ct)

      if ct eq 0 then begin
         flatfilenoothbin = where(flatimh.filter eq imh[f].filter and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic, ctwobin) ; and flatwins eq win
            ; Currently doesn't support flat-fielding with a different binning until I can be sure the case where both windowing and binning
            ; are wrong are dealt with

            if ctwobin gt 0 then binstr = imh[f].bin else binstr = '' ; .bin or .binning ?
            logprint, 'Flat field not found for ', removepath(files[f]), ' (', imh[f].filter+', '+'D'+imh[f].dichroic+ binstr,')' ; , ' / ', imh[f].win
            flatfail = [flatfail, files[f]]
            continue
      endif
      flatfile = flatfiles[flatfileno[0]]
      logprint, 'Flattening ', removepath(files[f]), ' using ', removepath(flatfile)

      flatproc, files[f], flatfile, flatminval=0.6, crop='auto', /quiet  ; add autocropping of low-signal zones

    endfor

end

; -------------

pro autolrisspecflatten, chip=chip, camera=camera

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   fchip = strmid(chip,0,1)

   files = file_search(spworkingdir2d+'p'+camchar+wildchar+fchip+'.fits')
   if overwrite eq 0 then files = unmatched(files, 'f')
   if n_elements(files) eq 1 and files[0] eq '' then return

   ; un-flattened data files
   imh = readlrisheaders(files)
   w = where(imh.slit ne 'direct' and strmid(imh.camera,0,1) eq camchar and imh.exptime gt 0.5, ct)
   if ct eq 0 then return      ; return silently if nothing to process
   files = files[w]
   imh   = imh[w]

   ; flat field files
   flatfiles = file_search(spworkingdir2d+'*flat*.fits')
   flatimh   = readlrisheaders(flatfiles)
   w = where(flatimh.slit ne 'direct' and strmid(flatimh.camera,0,1) eq camchar, nflats)
   if nflats eq 0 then begin   ; return loudly if something to process and no flats made yet
     logprint, 'No processed spectroscopy flats found!  Cannot perform spectroscopic flatfielding.'
     return
   endif
   flatfiles = flatfiles[w]
   flatimh   = flatimh[w]

       ;  Option to remove unused spec columns that seems to be no longer lused
       ;if total(ycrop eq [0,0]) eq 2.0 then begin  ; automatic cropping
       ;   flat = mrdfits(flats[f], /silent)
       ;   s = size(flat)
       ;   nx = s[1]
       ;   yprofile = total(flat[0:nx/5.,*], 1)/median(flat[0:nx/5.,*]) + $
       ;              total(flat[4*nx/5.:nx-1,*], 1)/median(flat[4*nx/5.:nx-1,*])
       ;              ; sum over x-axis, using the edge areas since we are liberally cropping out all rows with some
       ;              ; signal-less pixels and the spectrum is often tilted
       ;   signalzone = where(yprofile gt 0.86*median(yprofile)) ; hopefully there's no slit profile variation.
       ;   flatycrop[f,*] = [min(signalzone)+1,max(signalzone)-2]
       ;
       ;endif
       ;
       ; ycrop = median(flatycrop, dimension=1) ; assume it's the same for all gratings (and slits)
       ; this is probably not a good assumption!

 
   for f = 0, n_elements(files)-1 do begin
      if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
      outfile = fileappend(files[f], 'f')

      ; skip biases
      if imh[f].exptime lt 0.5 then continue
      ; skip halogens
      if imh[f].hal then continue ; internal halogen on

      ;print, files[f], elevation, domerelaz, target, counts, exptime, counts/exptime
    
      ; skip dome flats
      if total(strnum(strsplit(imh[f].lamps,',',/extract))) eq 0 and $   ; (but don't skip arcs, even if pointed at dome)
         imh[f].trapdoor eq 'open' then begin 
         if abs(imh[f].elevation-45) lt 0.5 and abs(imh[f].domerelaz) gt 3 then continue ; silently skip dome flats
         if strpos(imh[f].target,'cass') ge 0 then continue 
         if strpos(imh[f].target,'dome') ge 0 then continue     ; sometimes the telescope is pointed wrong; we still don't want these...
         if strpos(imh[f].target,'horizon') ge 0 then continue 
      endif
 
      flatfileno = where(flatimh.grname eq imh[f].grname and flatimh.chip eq imh[f].chip and flatimh.dichroic eq imh[f].dichroic and $
                         flatimh.binning eq imh[f].binning, ct) ;and flatimh.wins eq imh[f].win

      if ct eq 0 then begin
         logprint, 'Flat field not found for ', removepath(files[f]), ' (', strtrim(imh[f].grname,2), ', D', strtrim(imh[f].dichroic,2), ', ', imh[f].bin, ')'
         flatfail = [flatfail, files[f]]
         continue
      endif
      flatfile = flatfiles[flatfileno[0]]
      logprint, 'Flattening ', removepath(files[f]), ' using ', removepath(flatfile)
      flatproc, files[f], flatfile, flatminval=0.1, crop='auto', /quiet  ; note that the autocropping basically is overridden at the split phase.
                   ;crop=[0, ycrop[0], 0, ycrop[1]],  
    endfor

end


; ----------------------
pro autolrisreflatten, chip=chip, camera=camera, filters=filters

; Create a special super-sky flat specific to these exact settings.

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

    if strlen(chip) gt 1 then fchip = strmid(chip,0,1) else fchip = chip
    camchar = strmid(camera,0,1)

    files = file_search(imworkingdir+'fp'+camchar+wildchar+fchip+'.fits')
    imh = readlrisheaders(files)
    w = where(imh.slit eq 'direct', ct)
    if ct eq 0 then return
    files = files[w]
    imh = imh[w]
    reflatfiles = imworkingdir + 'i' + removepath(files)

               extendedsources = [''] 
               if file_test('extendedsources.txt') then readcol, 'extendedsources.txt', extendedsources, format='a', comment='#', /silent
               badflattargets = [''] 
               if file_test('badflattargets.txt') then  readcol, 'badflattargets.txt',  badflattargets,  format='a', comment='#', /silent
               badflattargets = unique([badflattargets, extendedsources])
               notblank = where(badflattargets ne '', ct)
               if ct gt 0 then badflattargets = badflattargets[notblank]

    ugr = unique(imh.gratf)       
    for g = 0, n_elements(ugr)-1 do begin
      if n_elements(filters) eq 0 then filts = unique(imh[where(imh.gratf eq ugr)].filter) else filts = filters
      for f = 0, n_elements(filts)-1 do begin

          wfg = where(imh.filter eq filts[f] and imh.gratf eq ugr[g], ctfg)
          if ctfg eq 0 then continue
          mednx = median([imh[wfg].nx])
          medny = median([imh[wfg].ny])  ; this is a lazy check, since it doesn't check the x0/y0 parameters.
          wfg = where(imh.filter eq filts[f] and imh.gratf eq ugr[g] and imh.nx eq mednx and imh.ny eq medny, ctfg)

          if ctfg eq 0 then continue

          unreflattened = 0
          for i = 0, ctfg-1 do begin   ; need to implement fileseq, etc., here
             if file_test(reflatfiles[wfg[i]]) eq 0 then unreflattened = 1
          endfor
          if unreflattened eq 0 and overwrite eq 0 then continue
          minfiles = 9
          if clip(filts[f]) eq 'G' then minfiles = 15
          if clip(filts[f]) eq 'V' then minfiles = 11  ; how many files before the last grating change needed?
          if clip(filts[f]) eq 'R' then minfiles = 10
          if clip(filts[f]) eq 'I' then minfiles = 8
          if clip(filts[f]) eq 'RG850' then minfiles = 6
          if ctfg lt minfiles then begin
             logprint, 'Not enough files in filter ', filts[f],' to reflat block of imaging starting at ', removepath(files[wfg[0]]), $
                    '  (',clip(ctfg), ' files)'
             continue
          endif
          ufields = unique(imh[wfg].target)  ; need to do a badflattarget/extendedsource check
          nfields = n_elements(ufields)
          goodufields = ['']
          wfggood = [-1]
          for u = 0, n_Elements(ufields)-1 do begin
             if total(ufields[u] eq badflattargets) eq 0 then begin
                goodufields = [goodufields, ufields[u]]
                wfggood = [wfggood, wfg[where(imh[wfg].target eq ufields[u])] ]   
             endif
          endfor

          ngoodfields = n_elements(goodufields)-1
          ngoodimages = n_elements(wfggood)-1
          if ngoodfields gt 0 then goodufields = goodufields[1:n_elements(goodufields)-1]
          skytblock = median(imh[wfggood].exptime, /even) lt 120 and median(imh[wfggood].ctrate, /even) gt 300
             ; seems to be a flag indicating sky flats.  
             ; formerly 'and skytblock eq 0' was required to trigger skipping -
             ;  we sometimes want to reflat twilight images?  deactivated for now, but consider re-adding...
          if (ngoodfields eq 0) or $
             (ngoodfields le 2 and ngoodimages le 13) or $  
             (ngoodfields le 3 and ngoodimages le 9) then begin
             logprint, 'Not enough fields in filter ', filts[f],' to reflat block of imaging starting at ', removepath(files[wfg[0]]), $
                    '  (',clip(ngoodimages), ' files, ', clip(nfields),' fields)'
             ;ugr[g]
             continue
          endif

          wfggood = wfggood[1:n_elements(wfggood)-1]
          wfggood = wfggood[sort(wfggood)]

          logprint, 'Creating special supersky reflats for ', clip(ctfg), ' ', filts[f]+'-band files between ', removepath(files[wfg[0]]), ' and ', removepath(files[wfg[ctfg-1]]), ' (', clip(nfields),' fields)'

          selfsupersky, files[wfg], outfilelist=reflatfiles[wfg]   ; include an "also" option to flat-field extra files
          
          ;fileseqstr = lrisfilenum(ugr[g], /str)
          ;outflatname = imworkingdir+'lrisrefl_'+filts[f]+'_'+fileseqstr+'.fits'
          ;if file_test(outflatname) eq 0 then begin
          ;   print, 'Creating special supersky reflat for ', clip(ctfg), ' files between ', removepath(files[wfg[0]]), ' and ', removepath(files[wfg[ctfg-1]])
          ;   flatcombine, files[wfg], outflatname, /removeobjects, type='supersky', xnormkey='AMPR1'
          ;endif

          ;for i = 0, ctfg-1 do begin    ; i for "improved" flatfield.
          ;   file = files[wfg[i]]
          ;   outfile = imworkingdir + 'i' + removepath(file)
          ;   if file_test(outfile) then continue
          ;   print, 'Reflattening ', file, ' -> ', outfile
          ;   if n_elements(flatim) le 1 then flatim = readfits(outflatname)
 
          ;   im = readfits(file, h)
          ;   im /= flatim
          ;   sxaddpar, h, 'REFLAT', outflatname, after='FLATTYPE'

          ;   writefits, outfile, im, h
          ;endfor
          ;flatim = 0

       endfor
    endfor


end



; -------------------------
pro autolrismakefringe

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

  if lrisversion eq 1 and makefringe ne 0 and camera eq 'red' then begin
    ffiles = file_search('f'+prefchar+wildchar+'?'+'.fits')
    fileinfo = replicate({filename:'', filter:'', counts:0., exptime:0., airmass:0., object:''}, n_elements(ffiles))
    fileinfo.filename = ffiles
    for f = 0, n_elements(fileinfo)-1 do begin
      if fileinfo[f].filename eq '' then continue
      h = headfits(fileinfo[f].filename, /silent)
      fileinfo[f].filter = strtrim(sxpar(h, filtkey),2)
      fileinfo[f].counts =  sxpar(h, 'COUNTS')
      fileinfo[f].exptime = sxpar(h, 'ELAPTIME')
      fileinfo[f].object  = sxpar(h, 'OBJECT')
      fileinfo[f].airmass = sxpar(h, 'AIRMASS')
      ;print, fileinfo[f].filename, ' ', fileinfo[f].filter, fileinfo[f].counts, fileinfo[f].exptime
    endfor
  
    fringablefilts = ['I','RG850']
    fringeexcludeobj = ['']  ;['GRB070412', '050412', 'GRB051109B', 'GRB070521', 'GRB100205A']
    for ff = 0, n_elements(fringablefilts)-1 do begin
       ffilt = fringablefilts[ff]
       fringefile = ffilt + 'fringe.fits'
       if file_test(fringefile) then continue
       thisfilt = where(fileinfo.filter eq ffilt, ct)
       if ct lt 3 then continue
       ffileinfo = fileinfo[thisfilt]
       thisdark = where(ffileinfo.exptime gt 60 and ffileinfo.counts lt 40000, ct)
       if ct lt 3 then continue
       ffileinfo = ffileinfo[thisdark]
       ctratethreshold = median((ffileinfo.counts/ffileinfo.exptime)/ffileinfo.airmass) * 1.8
       thisdark = where((ffileinfo.counts/ffileinfo.exptime)/ffileinfo.airmass le ctratethreshold)
       if ct lt 3 then continue
       ffileinfo = ffileinfo[thisdark]
       for nf = 0, n_elements(fringeexcludeobj)-1 do begin
          notforbiddenobj = where(ffileinfo.object ne fringeexcludeobj[nf], ct)
          if ct ge 1 then ffileinfo = ffileinfo[notforbiddenobj]
       endfor
       if ct lt 3 then continue

       logprint, 'Making ', ffilt+'-band fringe frame...'

       logprint, ffileinfo.filename

       mkfringe, ffileinfo.filename, outfile=fringefile, /removeobjects
    endfor
  endif


end




; -------------------------
pro autolrisrmfringe

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

  if lrisversion eq 1 and rmfringe ne 0 and camera eq 'red' then begin
    ffiles = file_search('f'+prefchar+wildchar+'?'+'.fits')
    if overwrite eq 0 then ffiles = unmatched(ffiles,'i')
    fileinfo = replicate({filename:'', filter:'', counts:0., exptime:0., object:''}, n_elements(ffiles))
    fileinfo.filename = ffiles
    for f = 0, n_elements(fileinfo)-1 do begin
      if fileinfo[f].filename eq '' then continue
      h = headfits(fileinfo[f].filename, /silent)
      fileinfo[f].filter = strtrim(sxpar(h, filtkey),2)
      fileinfo[f].counts =  sxpar(h, 'COUNTS')
      fileinfo[f].exptime = sxpar(h, 'ELAPTIME')
      fileinfo[f].object  = sxpar(h, 'OBJECT')
    endfor

    fringablefilts = ['I','RG850']
    for ff = 0, n_elements(fringablefilts)-1 do begin
       ffilt = fringablefilts[ff]
       fringefile = ffilt + 'fringe.fits'
       if file_test(fringefile) eq 0 then continue
       thisfringe = where(fileinfo.filter eq ffilt and fileinfo.counts/fileinfo.exptime lt 1000., ctfringe)

       if ctfringe ge 1 then begin
         fringablefiles = fileinfo[thisfringe].filename
         if fringablefiles[0] ne '' then $
           fringecorrect, fringablefiles, fringefile=fringefile, prefix='i';, /iterate, itersec=[50,880,200,-200]
         ; alternative method uses weights rather than iteration
         ;fringescale = filesi.counts
         ;continuumimg = where(filesi.counts/(filesi.exptime*filesi.airmass) gt 10, ctcontinuum)
         ;if ctcontinuum gt 0 then fringescale[continuumimg] = (filesi[continuumimg].exptime*filesi[continuumimg].airmass)/0.15
         ;fringecorrect, filesi.name, fringefile='Ifringe.fits', /iterate, itersec=[300,700,400,600]
      endif
    endfor
  endif

end




; -------------------------
pro autolriscrcleanim, chip=chip, camera=camera

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

   prefchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   files = choosefiles(prefchar+wildchar+chipchar+'.fits',imworkingdir+'ifp',imworkingdir+'fp')
   if overwrite eq 0 then files = unmatched(files,'z')

   for f = 0, n_elements(files)-1 do begin
      if files[f] eq '' then continue
      if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
      h = headfits(files[f])
      counts = sxpar(h,'COUNTS')
      exptime = sxpar(h,'ELAPTIME')
      target = sxpar(h,'TARGNAME')
      ;if counts gt 40000. then continue
      if counts gt 30000. and exptime lt 10. then continue
      if counts gt 20000. and exptime gt 5. and exptime lt 10. then continue
      if target eq '' then continue   ; blank target is certainly not a science object.
      if strpos(target,'flat') ge 0 then continue
      if strpos(target,'sky') ge 0 then continue
      if strpos(target,'twilight') ge 0 then continue

      logprint, 'Cleaning cosmic rays from ', removepath(files[f])
      zeal = 0.85
      if camera eq 'blue' then zeal = 0.75
      if (camera eq 'red' and lrisversion eq 1) then usamp=1 else usamp=0
      if camera eq 'red' and lrisversion gt 1 then zeal = 0.6 ;evidently should be the new default
      if n_elements(setzeal) eq 1 then zeal = setzeal
      pzap, files[f], zeal=zeal, usamp=usamp, /quiet   ; /weight,   no longer dealing with weight maps.  but ideally we should.
   endfor

end


; -------------------------
pro autolriscrcleanspec, chip=chip, camera=camera, doubleclean=doubleclean

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

camchar = strmid(camera,0,1)
chipchar = strmid(chip,0,1)

files = choosefiles(spworkingdir2d+'fp'+camchar+wildchar+chipchar+'.fits')

if overwrite eq 0 then files = unmatched(files,'z')

;if camchar eq 'b' then boxsize = 5
;if camchar eq 'r' then boxsize = 7

for f = 0, n_elements(files)-1 do begin
   if files[f] eq '' then continue
   if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue

   h = headfits(files[f])
   lampstr = clip(sxpar(h,'LAMPS'))
   lamps = strsplit(lampstr,',',/extract)
   ;;if total(lamps ne '0') gt 0 then continue ;  don't bother with lamps, arcs, etc.

   trapdoor = clip(sxpar(h,'TRAPDOOR'))
   if trapdoor eq 'closed' then continue

   binning = clip(sxpar(h,'BINNING'))
   bins = fix(strsplit(binning,',',/extract))

   ;if (bins[0] ge 2 or bins[1] ge 2) and boxsize ge 5 then boxsize -= 2   ; note the order is y,x for lris-red

   ; should use a faster and less dangerous algorithm for the blue side?

   logprint, 'Cleaning cosmic rays from ', removepath(files[f])
   doubleclean = 1
   if keyword_set(doubleclean) eq 0 or camchar eq 'b' then begin
       pzapspec, files[f], usamp=(camchar eq 'b'), ybin=(camchar eq 'r' and bins[0] gt 1), /mask
       ; formerly a bug had the binning always set (if LRIS-R) even if unbinned.
   endif else begin
       outname = spworkingdir2d+'z'+removepath(files[f])
       outmaskname = repstr(outname,'.fits','.mask.fits')
       pzapspec, files[f], 'temp.fits', 'temp.mask.fits', nsigma=19, ybin=(bins[0] gt 1), /mask
       pzapspec, 'temp.fits', outname, outmaskname, nsigma=16, inmask='temp.mask.fits', /usamp, /mask
       spawn, 'rm -f temp.fits temp.mask.fits'
   endelse

         ;, /nan  ; the nan is new and temporary

endfor


end


; ------------------
pro autolrissplit, chip=chip, camera=camera, mode=mode, nodistort=nodistort


   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

camchar = strmid(camera,0,1)
if n_elements(chip) gt 0 then chipchar = strmid(chip,0,1) else chipchar = ''

if mode eq 'im' then workingdir = imworkingdir
if mode eq 'sp' then workingdir = spworkingdir2d

if chipchar eq 'r' then rightonly=1
if chipchar eq 'l' then leftonly=1

files = choosefiles(camchar+wildchar+'.fits',workingdir+'ifp',workingdir+'fp')

for f = 0, n_elements(files)-1 do begin
   if files[f] eq '' then continue
   if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue

   rfile = repstr(repstr(files[f],'.fits','r.fits'),'if','f')
   lfile = repstr(repstr(files[f],'.fits','l.fits'),'if','f')

   if overwrite eq 0 then begin
      if file_test(rfile) and file_test(lfile) then continue
      if file_test(rfile) and keyword_set(rightonly) then continue
      if file_test(lfile) and keyword_set(leftonly) then continue
   endif

   lrissplit, files[f], rightonly=rightonly, leftonly=leftonly, renameas=repstr(files[f],'if','f'), nodistort=nodistort

endfor

end

; -------------------------
pro autolrisskysubtract, camera=camera, chip=chip, nocrclean=nocrclean

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

camchar = strmid(camera,0,1)
chipchar = strmid(chip,0,1)

if keyword_set(nocrclean) then $
   files = choosefiles(camchar+wildchar+chipchar+'.fits',spworkingdir2d+'zfp',spworkingdir2d+'fp') $
else $
   files = choosefiles(camchar+wildchar+chipchar+'.fits',spworkingdir2d+'zfp')

if overwrite eq 0 then files = unmatched(files,'s')

for f = 0, n_elements(files)-1 do begin
   if files[f] eq '' then continue
   if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue


   h = headfits(files[f])
   lampstr = clip(sxpar(h,'LAMPS'))
   lamps = strsplit(lampstr,',',/extract)
   ;;if total(lamps ne '0') gt 0 then continue ;  skip flats/arcs

   trapdoor = clip(sxpar(h,'TRAPDOOR'))
   if trapdoor eq 'closed' then continue

   dum = subtractsky2(files[f])

endfor

end

; -------------------------

pro autolriswavcal, camera=camera, chip=chip

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

    ; new functionality: order files as to always start with the standards
    ; visually display the extraction
 

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if chipchar eq 'l' then begin
      return
   endif

   arclistfile = refdatapath+'/'+'lrisarcall.list'   
   brightarclistfile = refdatapath+'/'+'lrisarcbright.list'


   ; Find useable arcs
   
   ; Get list of files which weren't crcleaned/skysubtracted (probable arcs)
   files = choosefiles(camchar+wildchar+chipchar+'.fits',spworkingdir2d+'fp', ct=ct)
   if ct gt 0 then files = unmatched(files, 'z', ct=ct)
   if ct gt 0 then files = unmatched(files, 's', ct=ct)
   if ct gt 0 then files = unmatched(files, 'sz', ct=ct)

   altfiles = choosefiles(camchar+wildchar+chipchar+'_arc.fits',spworkingdir1d+'fp', ct=ctalt)
   if ct eq 0 and ctalt gt 0 then files = altfiles  
   if ct eq 0 and ctalt gt 0 then ct = ctalt
     ; This is a kludgy way to do it.  The arc 2D files are looked at only if there are no arcs in the 2d/ directory.

   if ct eq 0 then return  ; No arcs at all have been taken yet
   

   ; future: copy these to 1d directory?  and also check there?

   ; Read in bad arc list

   if n_elements(badarclist) eq 0 then begin
     badarclist = ['']
     if file_test('badarc.txt') then begin
       nbarc = countlines ('badarc.txt')
       openr, 6, 'badarc.txt'
       badarcs = strarr(nbf)
       badarcstr = ''
       while not eof(6) do begin
          readf, 6, inline
          badarcstr += (inline + ' ')
       endwhile
       close, 6 
       badarclist = strlowcase(strsplit(badarcstr,/extract))
       notblank = where(badarcs ne '', ct)
     endif
   endif

   arcp = readlrisheaders(files, badarclist=badarclist)

   ; Remove the first member of identical consecutive doubles (probably unwarmed)   
   keep = bytarr(ct)+1
   for f = 1, ct-1 do begin
      if arcp[f].framenum-1 ne arcp[f-1].framenum then continue ; nonconsecutive
      if arcp[f].grname eq arcp[f-1].grname and arcp[f].binning eq arcp[f-1].binning and arcp[f].window eq arcp[f-1].window and $
         arcp[f].cwave eq arcp[f-1].cwave and arcp[f].dichroic eq arcp[f-1].dichroic and arcp[f].lamps eq arcp[f-1].lamps and $
         arcp[f].slit eq arcp[f-1].slit then $
         keep[f-1] = 0
   endfor
   arcp = arcp[where(keep)]
   files = files[where(keep)]

   ; Determine what are real arcs, with at least two lamps (one is not sufficient)
   if camchar eq 'b' then arcnlamps = arcp.hg + arcp.cd + arcp.zn   ; +arc.fe
   if camchar eq 'r' then arcnlamps = arcp.hg + arcp.neon + arcp.ar  ; + arc.kr + arc.xe
   isarc = (arcnlamps ge 2)             ; the factor of 2 on Ar is for broken-neon nights...

   w = where(isarc, narc)
   if narc eq 0 then return ; no arcs good enough to solve with
   arcfiles = files[w]
   arc = arcp[w]
   ;outarcfiles = spworkingdir1d+repstr(removepath(arcfiles),'.fits','_arc.spec')

   ; Which are already extracted?
   if overwrite eq 0 then begin
     outspec = spworkingdir1d+removepath(repstr(repstr(arcfiles,'_arc',''),'.fits','_arc.spec'))
     w = where(file_test(outspec) eq 0, narc) ; unsolved arcs
     if narc gt 0 then begin
       arcfiles = arcfiles[w]
       arc = arc[w]
     endif
   endif
   ; maybe sort by central wavelength...?
 

   ; Extract arcs

   if narc gt 0 then begin
      ; Sort to enable easier comparisons
       s = sort(arc.cwave)
      arc = arc[s]
      arcfiles = arcfiles[s]

      ; We now have our list arcs to solve.  Proceed to solve them.


      ; Extract the arc spectrum at the standard position (still comes from the old defaults)
 
      for f = 0, narc-1 do begin

          ; Extract a generic arc spectrum and save the solution.

          arcfile = arcfiles[f]
          ; outarcfile = outarcfiles[f]
          outarcfile = spworkingdir1d+repstr(removepath(repstr(arcfile,'_arc','')),'.fits','_arc.spec')
          if file_test(outarcfile) and (overwrite eq 0) then continue

          logprint, '  Extracting arc '+arcfile
          arch = headfits(arcfile)
          arcdata = mrdfits(arcfile, 0, /silent)

          s = size(arcdata)  ; Formerly would use the default trace position,
          ny = s[2]          ;  now just use 1/3 of the way up the chip +/- 16 unbinned pixels
          if camchar eq 'b'  then normaltracepos = defaulttraceyblue
          if camchar eq 'r'  then normaltracepos = defaulttraceyred
          y0 = (normaltracepos/arc[f].ybin-16/arc[f].ybin) > 2
          y1 = (normaltracepos/arc[f].ybin+16/arc[f].ybin) < ny-2
          arcdata = arcdata[*,y0:y1]
                  ;nans = where(finite(arcdata) eq 0, ct)
                  ;if ct gt 0 then arcdata[nans] = 0. ; could interpolate, but usually nans are just bad rows
                  ;farc = total(arcdata,2)  ; could use a median to exclude cosmics, but tilted lines will vanish...
          farc = median(arcdata, dimension=2) ;*(y1-y0)
          writespec, outarcfile, farc, arch

          if repstr(outarcfile,'.spec','.fits') ne arcfile then $
          spawn, 'cp -p '+arcfile+' '+repstr(outarcfile,'.spec','.fits')
              ; in the future, at this point will write out file with an astrometrically accurate header
      endfor
   endif

   earcfiles = file_search(spworkingdir1d+'fp'+camchar+'*_arc.spec', count=nearc) ; identical to outarcfiles on first run
   updated = 0  ; flag to indicate whether or not a new solution has been determined and recorded

   for f = 0, nearc-1 do begin
       earcfile = earcfiles[f]
       arcsp = readspec(earcfile, h)
       ; print, earcfile
       ; print, sxpar(h,'UTC')
       arcinfo = readlrisheaders(earcfile)
       solfile = repstr(repstr(earcfile,'.fits','.sol'),'.spec','.sol')
       fitsfile = repstr(solfile,'.sol','.fits') ; for copying the fits file over to the 1d/ directory on success
       if file_test(solfile) and overwrite eq 0 then continue
                                      ;'~/research/redux/arc/lrisall.list'
       lampstr = arcinfo.lamps
       fixlampstr, lampstr, badarclist
       lamps = strsplit(lampstr,',',/extract)
       while n_elements(lamps) lt 8 do lamps = [lamps, '']
       lampnames = ['']
       if fix(lamps[0]) then lampnames = [lampnames, 'Hg']
       if fix(lamps[1]) then lampnames = [lampnames, 'Ne']
       if fix(lamps[2]) then lampnames = [lampnames, 'Ar']
       if fix(lamps[3]) then lampnames = [lampnames, 'Cd']
       if fix(lamps[4]) then lampnames = [lampnames, 'Zn']
       if fix(lamps[6]) then lampnames = [lampnames, 'Kr']
       if fix(lamps[7]) then lampnames = [lampnames, 'Xe']
       if n_elements(lampnames) gt 1 then lampnames = lampnames[1:*] $
          else continue ; shouldn't happen unless every arc is broken

       ; reduce arc lists to just the arcs in the file
      
       readcol, arclistfile, wav, name, format='f,a'
       arclist = filterarclist(wav, name, lampnames)  ; returns a filtered list of wavs
       readcol, brightarclistfile, wav, name, format='f,a'
       brightarclist = filterarclist(wav, name, lampnames)

       if clip(sxpar(h,'DICHNAME')) eq '560' then begin
            if camchar eq 'b' then arclist = arclist[where(arclist lt 6000.)]
       ;     if camchar eq 'r' then arclist = arclist[where(arclist gt 5400.-300)]
       endif

       lampstr = ''
       for l = 0, n_elements(lampnames)-1 do lampstr += lampnames[l]
       logprint, '  Wavelength-calibrating arc ', removepath(earcfile), ' (',lampstr,', '+arcinfo.grname+')'


       nx = n_elements(arcsp)

       dispersion = arcinfo.dispersion
       cwav = arcinfo.cwave
       ; could do a try loop here

       if n_elements(initsol) gt 0 then delvarx, initsol

       uncrefwav = 0.3   ; systematic uncertainty floor on wavelength measurement
       if arcinfo.grname eq '300/5000' then uncrefwav = 0.6 ; severe blending makes this solution always bad...

       if camchar eq 'r' then nbrightline = 20
       if camchar eq 'b' then nbrightline = 16
       if camchar eq 'r' then order = 7
       if camchar eq 'b' then order = 5 ; order has been increased.  Will this be safe?

       errorstatus = 0
       if debug eq 0 then catch, errorstatus
       if errorstatus ne 0 then begin
          logprint, 'Unexpected error during wavelength calibration:'
          logprint, !error_state.msg
          continue
       endif

       centroidrad = 4
       if arcinfo.xbin ge 2 then centroidrad = 2
       
       ; need to set x0 for the unlikely situation that the CCD was windowed in x directon

       checkplotfile = repstr(earcfile,'.spec','.wavsol.ps') ; formerly from .fits
       outarclistfile = repstr(earcfile,'.spec','.list')     ;          from .fits to arc.list
       arcsol = solvearcspec(arcsp.flux, wav, refwav=arclist, brightrefwav=brightarclist, $
                  cwav=cwav*1., unccwav=cwav*0.05, dispersion=dispersion, uncdisp=0.075*dispersion, $
                  maxsep=nx*dispersion*0.5, uncrefwav=uncrefwav, initsol=initsol, nbrightline=nbrightline, $
                  minorder=order, maxorder=order, $    ; order=3 is a little more robust but will be off by an angstrom, including over dichroic
                  centroidrad=centroidrad, $
                  ndet=ndet, nmatch=nmatch, ngoodmatch=ngoodmatch, checkplotfile=checkplotfile, medabsresidual=medabsresidual) ; , outarclistfile=outarclistfile ; for user solutions?
       checkplotfile = ''
       catch, /cancel 


       validated = 1
       ;valerrstr = ''
       if n_elements(arcsol) ge 2 then begin
          checkcwav = arcsol[0]
          checkdisp = arcsol[1]
          if abs(checkcwav - cwav) gt cwav*0.05 then valerrstr = 0
          if abs(checkdisp - dispersion) gt dispersion*0.075 then validated = 0
          outwav = poly(findgen(nx)-nx/2., arcsol)
          dispx = outwav-shift(outwav,1)
          dispx[0] = dispx[1]
          avdisp = median(dispx)
          if abs(avdisp-dispersion) gt 0.075*dispersion then validated = 0
          okdisp = (dispx - avdisp)/avdisp lt 0.2*dispersion
          wokdisp = where(okdisp eq 0 and outwav gt 2900., ct)
          if ct gt 1 then validated = 0
          if validated eq 0 then begin
              logprint, '  Automatic wavelength solution failed validation check.  Not using this arc.  Were lamps not warm?'
          endif
       endif

       if camchar eq 'r' then residualthresh = 0.5
       if camchar eq 'b' then residualthresh = 0.8 ; formerly 1.7
       if camchar eq 'b' and arcinfo.grname eq '300/5000' then residualthresh=2.0
       if n_elements(arcsol) gt 2 and validated and nmatch gt 8 and ngoodmatch gt nmatch*0.25 and medabsresidual lt residualthresh then begin
           logprint, '   -> ', removepath(solfile)
           openw, 9, solfile
           for i = 0, n_elements(h)-1 do printf, 9, '#'+h[i]
           for i = 0, n_elements(arcsol)-1 do printf, 9, arcsol[i]
           close, 9  ; could print out a header (and even search it), but it's not really necessary now
           updated = 1
       endif else begin
           logprint, '  ERROR: Failed to solve arc.   Were the lamps warmed up, and is the central wavelength correct?' 
           ; could now just redo it until it works...?
            wavsolfail = [wavsolfail, earcfile]
           continue
       endelse
   endfor

   ; Finally, extract a "generic" solution that averages all the non-constant terms together to ensure consistency between frames.
   ; Also produce a summary of all the individual wavelength solutions.

   if updated then begin
    psopen, 'wavsol'+camchar+'.ps', xsize=11, ysize=8, /inches
    device, /color
    !p.font=0
    device, /times, font_index=17                                  
    colors = transpose([[0,0,0],$
                      [255,50,50],$ 
                      [120,120,40],$ 
                      [50,200,50],$ 
                      [50,190,190],$
                      [60,60,255],$
                      [140,40,140]])
    tvlct, colors
    cols = indgen(6)
    !p.multi = [0,1,3]

    solfiles = findfile(spworkingdir1d+'*fp'+camchar+'*_arc.sol')
    sh = readlrisheaders(solfiles)
    gr = unique(sh.grname) ; should also consider the possibility of highly different CWAV values
    for g = 0, n_elements(gr)-1 do begin
      ; print, gr[g]
      grstr = repstr(gr[g],'/','-')
      ;avfile = spworkingdir1d+'lris'+camchar+grstr+'_average.sol'
      ;print, avfile
      w = where(sh.grname eq gr[g], nsol)
      csolfiles = solfiles[w]
      csh = sh[w]
      for s = 0, nsol-1 do begin
         insol = grabcolumn(csolfiles[s],0)      
         if s eq 0 then order = n_elements(insol)-1
         if s eq 0 then solmatrix = fltarr(order+1, nsol) + !values.f_nan
         if n_elements(insol) ne (size(solmatrix)) [1] then logprint, 'Order mismatch in trying to measure a standard wavelength solution'
         for p = 0, order do insol[p] /= csh[s].xbin^p
         solmatrix[*,s] = insol  
      endfor
      avsol = fltarr(order+1)
      rangecwav = max(solmatrix[0,*])-min(solmatrix[0,*])  ; make sure this is not too high
      if rangecwav gt 20 then logprint, 'Warning - large offsets present in wavelength solutions'
      for p = 0, order do begin
         avsol[p] = mean(solmatrix[p,*])
      endfor
      ;if file_test(avfile) eq 0 or overwrite then begin
      ;  openw, 5, avfile
      ;  printf, 5, '# '+clip(nsol)+' solutions'
      ;  printf, 5, '# GRNAME = '+clip(gr[g])
      ;  for p = 0, order do begin
      ;     printf, 5, avsol[p]
      ;  endfor
      ;  close, 5
      ;endif

      if camchar eq 'r' and lrisversion eq 1 then begin
         nx = 2048
         x0 = 1024
      endif else begin 
         nx = 4096
         x0 = 2048
      endelse
      ; print, nx, x0, camchar, lrisversion

      ybars = [0.08, 0.38, 0.68, 0.98]
      pix = findgen(nx)
      if camchar eq 'b' and grstr eq '400-3400' then pix = pix[1300:*]
      avdisp = median( poly(pix-x0,avsol) - shift(poly(pix-x0, avsol),1) )
      for b = 0, 2 do begin
        !p.position = [0.08, ybars[2-b], 0.98, ybars[3-b]]
        if b eq 0 then yrange=[min(poly(pix-x0, avsol)), max(poly(pix-x0, avsol))]
        if b eq 1 then yrange=[-100,100]
        if b eq 2 then yrange=[-4 < min(poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp])), $
                               +4 > max(poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]))]
        ;plot, pix, poly(pix-x0, avsol)-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]), $
        ;              thick=4, xrange=[0,nx], /xsty, yrange=yrange, /ynoz, ytickformat='(I5)', charsize=2.0
        if b eq 2 then xtitle = 'Pixel (physical)' else xtitle = ''
        if b eq 1 or b eq 2 then ytitle = 'Wavelength residual (A)' else ytitle = 'Wavelength (A)'
        plot, [0], [0], xrange=[0,nx], /xsty, yrange=yrange, /ynoz, ytickformat='(I5)', charsize=2.0, xtitle=xtitle, ytitle=ytitle
        if b eq 0 then xyouts, 0.1, 0.93, /norm, gr[g], charsize=1.5
        if b eq 0 then xyouts, 0.2, ybars[2-b]+0.04, /norm, 'Wavelength curve'
        if b eq 1 then xyouts, 0.1, ybars[2-b]+0.04, /norm, 'Linear differential'
        if b eq 2 then xyouts, 0.1, ybars[2-b]+0.04, /norm, 'Linear differential after linear correction'
        ;if b eq 0 then print, avsol
        for s = 0, nsol-1 do begin
           ;if b eq 0 then print, solmatrix[*,s]
           oplot, pix, poly(pix-x0, solmatrix[*,s])-(b gt 0)*poly(pix-x0, [avsol[0], avdisp]) $
                       - (b eq 2)*(solmatrix[0,s]-avsol[0]), $
                       color=(s mod 5)+1
        endfor
      endfor
    endfor 

    psclose
    retitleplot, 'wavsol'+camchar+'.ps', (camchar eq 'b' ? 'Blue' : 'Red')+' Wavelength solutions'
    !p.multi = 0
   endif

end

; -----------

pro autolrissum, camera=camera, chip=chip

   ; Combine 2D spectra taken at a common orientation.

   ; Should probably only stack sequences - try looping through files, and with each file look ahead to see if the next one
   ; is of the same target and PA and configuration and similar exposure and no significant x-dither.

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)
   uwildchar = wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   if n_elements(date) gt 0 then uwildchar = clip(date)+strmid(uwildchar,6)

   files = choosefiles(camchar+uwildchar+chipchar+'.fits',spworkingdir2d+'szfp',spworkingdir2d+'sfp', ct=ct)
   if ct eq 0 then return
 
   fileh = readlrisheaders(files) ; should eventually put this in one of the conditionals below

   i = 0
   while i le n_elements(files)-1 do begin
      combfiles = [files[i]]
      ; look ahead, unless this is a short exposure
      ; print, fileh[i].filename, ' ', rclip(fix(fileh[i].exptime),6), '  ', fileh[i].target
      if fileh[i].exptime gt 100. then begin
        j = i+1
        while j le n_elements(files)-1 do begin
          ; print, '   ', fileh[j].filename, fileh[j].target, fileh[j].slitpa, fileh[j].grname, fileh[j].dichroic, fileh[j].cwave, fileh[j].gratf
          ; Note that files are being compared to the FIRST file, since j increments until the loop breaks but i remains fixed.
          if fileh[j].target ne fileh[i].target then break
          if abs((fileh[j].slitpa - fileh[i].slitpa)+360. mod 360) gt 1.0 then break
          if fileh[j].xbin ne fileh[i].xbin or fileh[j].ybin ne fileh[i].ybin then break
          if fileh[j].grname ne fileh[i].grname then break
          if fileh[j].dichroic ne fileh[i].dichroic then break
          if abs(fileh[j].cwave - fileh[i].cwave) gt 3. then break
          if fileh[j].gratf ne fileh[i].gratf then break  ; configuration changed, presumably a file sequence gap
          if (ten(fileh[j].ut) - ( ten(fileh[j-1].ut)+fileh[j-1].exptime/3600. )) gt 0.25 then break ; long gap - probably guiding was lost,
                                                                                               ; or the intent was to get spaced spectra
          slitparad = fileh[i].slitpa*!pi/180.
          cosdec = cos(fileh[i].decdeg*!pi/180.)
          raoff = ((fileh[j].radeg-fileh[i].radeg)*cosdec)*3600.
          decoff = (fileh[j].decdeg-fileh[i].decdeg)*3600.
          xoff = raoff*cos(slitparad)-decoff*sin(slitparad)  ; should be zero if slit remains on source
          yoff = raoff*sin(slitparad)+decoff*cos(slitparad)  ; dither step
          ;print, removepath(files[i]), ' ', removepath(files[j]), raoff, decoff, ' | ', xoff, yoff
          if abs(xoff) gt 1.5 or abs(yoff) gt 30. then break     ; off-slit or huge dither
          ;print, fileh[j].exptime, fileh[i].exptime
          ; could also check filter and some other things
          if fileh[j].exptime gt 60. and fileh[j].exptime gt fileh[i].exptime*0.2 then begin
             combfiles = [combfiles, files[j]]  
          endif
          j += 1
        endwhile
      endif else begin
        ; maybe consider doing something special if the next exposure is long on the same 'target'?
        j = i+1
      endelse

      nsum = n_elements(combfiles)
      outtarg = repstr(fileh[i].target,' ','')
      outtarg = repstr(fileh[i].target,'-','') ; ?
      outtarg = repstr(outtarg,'_S','|')
      outtarg = (strsplit(outtarg,'|',/extract)) [0]
      outtargstr = repstr(legalfilestr(outtarg),'_','')
      outtargstr = clip(clip(outtargstr,10))
      ;if strpos(strupcase(outtargstr),'GRB') eq -1 and $
      ;   strpos(strupcase(outtargstr),'PTFS') eq -1 then outtargstr = clip(clip(outtargstr,8))
                                                        
      datestr = strmid(repstr(fileh[i].date,'-',''),2,6)
      ; outpastr = string(upa[p],'(I03)')
      ; grnamestr = repstr(fileh[w[0]].grname,'/','.')
      filen0str = string(fileh[i].framenum,'(I04)')
      outfile = spworkingdir1d+'ls'+camchar+datestr+'_'+filen0str+chipchar+'_'+outtargstr+'_sum'+clip(nsum)+'.fits'
      outfile = repstr(outfile,'_sum1','')
      outfile = repstr(outfile,'_sum','_s')
      ; print, outfile, file_test(outfile)
      if file_test(outfile) and overwrite eq 0 then begin
         i = j
         continue
      endif
      competingoutfile = spworkingdir1d+'ls'+camchar+datestr+'_'+filen0str+chipchar+'_'+outtargstr+'*'+'.fits'
      checkcompeting = file_search(competingoutfile,count=ctcompete)
      if ctcompete gt 0 then begin
         for k = 0, ctcompete-1 do begin
            if checkcompeting[k] ne outfile then logprint, 'Deleting '+checkcompeting[k]
            if checkcompeting[k] ne outfile then spawn, 'rm -f '+checkcompeting[k]
         endfor
      endif

      if nsum eq 1 then begin
         if file_test(outfile) then spawn, 'rm -f '+outfile
         logprint, ' '+combfiles[0]+' '+clip('('+clipzero(round(fileh[i].exptime))+'s)',7)+'-> '+removepath(outfile)
         spawn, 'cp '+combfiles[0]+' '+outfile
      endif else begin
         combmasks = repstr(repstr(combfiles,'.fits','.mask.fits'),'sz','z')
         if total(file_test(combmasks)) lt n_elements(combmasks) then delvarx, combmasks ; if no mask files, skip masking...
         pixoffsets = intarr(nsum)
         xoffsets = fltarr(nsum) ; in arcsec
         ra0 = fileh[i].radeg
         dec0 = fileh[i].decdeg
         cosdec0 = cos(dec0*!pi/180.)
         slitparad = fileh[i].slitpa*!pi/180.
         for k = 1, nsum-1 do begin
           wk = where(files eq combfiles[k])
           ra = fileh[wk].radeg
           dec = fileh[wk].decdeg
           raoff = ((ra-ra0)*cosdec0)*3600.
           decoff = (dec-dec0)*3600.
           xoff = raoff*cos(slitparad)-decoff*sin(slitparad)  ; should be zero if slit remains on source
           yoff = raoff*sin(slitparad)+decoff*cos(slitparad)  ; dither step
           if abs(yoff) lt 0.01 then yoff = 0.0
           yoffpix = yoff/(0.135*fileh[wk].ybin)
           pixoffsets[k] = fix(round(yoffpix))
           xoffsets[k] = xoff
         endfor
         logprint, ' Combining 2D frames: '
         for k = 0, nsum-1 do begin
           if abs(xoffsets[k]) gt 0.01 then sstr = ' lateral offset: '+clip(fpr(xoffsets[k],4.2))+'"' else sstr = ''
           logprint, '   '+removepath(combfiles[k])+' '+rclip(fix(round(fileh[wk].exptime)),5)+'s'+' '+clip(pixoffsets[k])+'px'+sstr
         endfor
         logprint, clip('',20)+'   --> '+removepath(outfile)

         combinesum, combfiles, outfile, mask=combmasks, yshift=pixoffsets, /noweight, /silent
      endelse


      i = j
   endwhile

   return
end


; -----------

pro autolristrace, camera=camera, chip=chip

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

   ; This function does several things:
   ;  (1) a) Uses "trace2d" to measure the trace of every frame using ALL sources in each image, and writes out
   ;      .trace and .profile files containing the trace function and summed spatial profile.
   ;      b) Failing that, it uses the trace offsets from a similar observation to still write out the profile.
   ;  (2) Looks through profile files corresponding to 'bright' stars to measure where the trace normally resides.
   ;      These are saved to brightobjectpos.txt
   ;  (3) Finally, goes through all profile files to find the best guess of the science target position and aperture width.
   ;      These are saved to objectpos.txt

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)
   uwildchar = wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   if n_elements(date) gt 0 then uwildchar = clip(date)+strmid(uwildchar,6)

   if chipchar eq 'l' then return

   ; files = choosefiles(camchar+uwildchar+chipchar+'.fits',spworkingdir+'szfp',spworkingdir+'sfp', ct=ct)
   files = file_search(spworkingdir1d+'ls'+camchar+'*.fits', count=ct)
   if ct eq 0 then return

   tracefiles = strarr(n_elements(files))
   fileh = readlrisheaders(files)


   dum = file_search(spworkingdir1d+'ls'+camchar+'*.profile', count=ntrfilebefore)

   ; Trace spectra

   for f = 0, n_elements(files)-1 do begin
     outfile = repstr(files[f], '.fits', '.trace')
     profilefile = repstr(files[f], '.fits', '.profile')
     plotfile = repstr(files[f], '.fits', '.trace.eps')
     ;if (file_test(profilefile) and file_test(outfile)) and keyword_set(overwrite) eq 0 then continue
     if (file_test(profilefile) or file_test(outfile)) and keyword_set(overwrite) eq 0 then continue
     h = headfits(files[f])
     delvarx, xstart
     delvarx, xrange
     if camchar eq 'b' and sxpar(h,'GRNAME') eq '400/3400' then xstart=sxpar(h,'NAXIS1')*mean([0.65,0.9])
     if camchar eq 'b' and sxpar(h,'GRNAME') eq '400/3400' then xrange=sxpar(h,'NAXIS1')*[0.35,1.0]
     logprint, 'Measuring spatial profile and trace function for '+removepath(files[f])
     trace2d, files[f], xstart=xstart, xrange=xrange, tracefile=outfile, profilefile=profilefile, plotfile=plotfile
   endfor


   ; for any failed traces, use the trace function from another observation at same config, similar airmass/angle
   success = intarr(n_elements(files))
   for f = 0, n_elements(files)-1 do begin
      if file_test(repstr(files[f], '.fits', '.trace')) then success[f] = 1
   endfor
   fut = fltarr(n_elements(files))
   for f = 0, n_elements(files)-1 do fut[f] = ten(fileh[f].ut)
   for f = 0, n_elements(files)-1 do begin
     if success[f] eq 0 then begin
        profilefile = repstr(files[f], '.fits', '.profile')
        if (file_test(profilefile)) and keyword_set(overwrite) eq 0 then continue
           ; if profile exists but not in objectpos.txt, then objectpos.txt won't be written.  maybe don't want that?

        hf = fileh[f]
        wsimilar = where(success and hf.ybin eq fileh.ybin and hf.xbin eq fileh.xbin and $
                         hf.grname eq fileh.grname and abs(hf.cwave-fileh.cwave) lt 10., ct)
        if ct eq 0 then begin 
           logprint, '  '+removepath(files[f])+' has no strong traces and no similar observations found - no way to determine profile or trace.'
           continue
        endif
        ranksimilar = (hf.airmass-fileh[wsimilar].airmass)^2 + abs(fut[f]-fut[wsimilar])/24.
        s = sort(ranksimilar)
        tracefiles[f] = repstr(files[wsimilar[s[0]]],'.fits','.trace') ; includes the directory path
        logprint, '  Using trace file ', removepath(tracefiles[f]), ' for ', removepath(files[f])

        plotfile = repstr(files[f], '.fits', '.trace.eps')
        trace2d, files[f], xstart=xstart, settracefile=tracefiles[f], profilefile=profilefile, plotfile=plotfile
     endif
   endfor


   ; Find the preferred sources using the profiles

   brightobjfile = spworkingdir1d+'brightobjectpos.txt'
   objfile = spworkingdir1d+'objectpos.txt' ; consider adding camera...
   userobjfile = 'userobjectpos.txt'

   ; Get locations of bright objects (trace stars, usually standards)
   tfile = ['']
   if file_test(brightobjfile) eq 0 then begin
       openw, 10, brightobjfile
       printf, 10, '#filename                               ycen'
       close, 10
   endif else begin
      readcol, brightobjfile, tfile, tcenter, format='a,f', comment='#'
   endelse

   ; isstd = fileh.exptime lt 60. and fileh.maxcounts gt 5000.

   isstd = fileh.exptime le 61. and fileh.maxsource gt 4000. 
            ; and strpos(fileh.target,'HD') gt 0 or strpos(fileh.target,'Feige') gt 0... some other similar filtering?
   if total(isstd) le 1 then isstd = fileh.exptime le 100. and fileh.maxsource gt 4000.
   if total(isstd) le 1 then isstd = fileh.exptime le 120. and fileh.maxsource gt 1000.
          ; note: maxcounts is not always a reliable indicator...

   ; for j = 0, n_elements(fileh)-1 do print, fileh[j].filename, fileh[j].exptime, fileh[j].maxsource, isstd[j] ? ' * ' : '  '

   wstd = where(isstd, ct) ; probable standards (or other bright tracing objects) - might find some with the star at the edge, though
   for i = 0, ct-1 do begin
       if total(files[wstd[i]] eq tfile) gt 0 then continue  ; already measured the center
       ; measure a new center
       profilefile = repstr(files[wstd[i]],'.fits','.profile')
       if file_test(profilefile) eq 0 then continue ; no profile
       readcol, profilefile, y, prof, format='f,f', comment='#'
       maxpix = median(y[where(prof eq max(prof))], /even)
       if maxpix lt 20 or maxpix gt n_elements(y)-50 then continue ; not really the tracing star
        ; continuing will re-open profile files lots of times unnecessarily, but no big deal
       centroidpix = total(y[maxpix-5:maxpix+5]*prof[maxpix-5:maxpix+5])/total(prof[maxpix-5:maxpix+5])
       fitcenter = getlinecenter(y, prof, centroidpix, fitrad=10., width=width)
       if width gt 1.0 and width lt 20.0 then begin
          objectpos = fitcenter
       endif else begin
          objectpos = centroidpix
       endelse
       outstr = clip(removepath(files[wstd[i]]),35)+' '+fpr(objectpos,5.2)+' '+fpr(width,3.2)
       updatefile, brightobjfile, outstr
   endfor
   readcol, brightobjfile, tfile, tcenter, format='a,f'

   isstd = intarr(n_elements(files))  ; this will generally be the same as isstd, but need to recalc if loading from disk
   centerpos = fltarr(n_elements(files)) + !values.f_nan
   for i = 0, n_elements(tfile)-1 do begin
       match = where(tfile[i] eq removepath(fileh.filename), ctmatch)
       if ctmatch gt 0 then begin
         isstd[match] = 1   ; mark the bright standards
         centerpos[match] = tcenter[i]
       endif
   endfor

   ofile = ['']
   if file_test(objfile) eq 0 then begin
       openw, 10, objfile
       printf, 10, '#filename                               ycen    rad   background tracefile # (alternates) '
       close, 10
   endif else begin
       readcol, objfile, ofile, ocenter, orad, format='a,f,f', comment='#'
   endelse

   objfilechanged = 0
   ; Scan through the science targets and find the closest source to the average standard trace in the same configuration
   for i = 0, n_elements(files)-1 do begin
      filei = removepath(files[i])
      if total(filei eq ofile) gt 0 then continue  ; already measured the center

      profilefile = repstr(files[i], '.fits', '.profile')
      if file_test(profilefile) eq 0 then continue ; no profile, so there is no way to measure the center

      hi = fileh[i]
      ybin = hi.ybin

      wcam = where(fileh.camera eq hi.camera and isstd, ctcam)
      if ctcam eq 0 then begin
         logprint, '  No bright objects observed; not sure where to look for the trace.'
         avstdpos = 50./hi.ybin
         logprint, '  Guessing '+clipzero(avstdpos)
      endif else begin
        avstdpos = median(centerpos[wcam]*fileh[wcam].ybin, /even)/hi.ybin  ; default to search around the global average standard position
        wsimstd = where(fileh.xbin eq hi.xbin and fileh.ybin eq hi.ybin and fileh.grname eq hi.grname and isstd, ctsimstd)
        if ctsimstd ge 1 then avstdpos = median(centerpos[wsimstd], /even)
      endelse

      readcol, profilefile, y, prof, format='f,f', comment='#'
      nm = n_elements(prof)
      mprof = fltarr(nm)
      for k = 0, n_elements(mprof)-1 do mprof[k] = prof[k] - median([prof[(k-10)>0:(k-2)>0],prof[(k+2)<(nm-1):(k+10)<(nm-1)]])
      rms = median(abs(mprof-shift(mprof,1)),/even) ; some outlier rejection?
      if finite(rms) eq 0 or rms le 0. then begin
         logprint, ' WARNING: had difficulty estimating profile noise level.'
         rms = 1.
      endif
      smoothprof = convol(mprof,[0.5,1,1,1,0.5]/4.)
       
      ismax4 = smoothprof gt shift(smoothprof,-1) and smoothprof gt shift(smoothprof,1) and $
               smoothprof gt shift(smoothprof,-2) and smoothprof gt shift(smoothprof,2) and $
               smoothprof gt shift(smoothprof,-3) and smoothprof gt shift(smoothprof,3) and $
               smoothprof gt shift(smoothprof,-4) and smoothprof gt shift(smoothprof,4)
      peaks = where(ismax4 and $
                    smoothprof gt 3*rms and y gt 20/ybin and y lt n_elements(y)-40/ybin, ctpeak)
      if ctpeak eq 0 then begin
         peaks = where(ismax4 and $
                    smoothprof gt 2.5*rms and y gt 40/ybin and y lt n_elements(y)-80/ybin, ctweakpeak)
         if ctweakpeak eq 0 then begin
            logprint, clip(removepath(profilefile))+': found no sources.'
            continue
         endif
      endif
      peakcen   = fltarr(n_elements(peaks))
      peakwidth = fltarr(n_elements(peaks))  ; this is sigma.  fwhm = 2*sqrt(2*alog(2.))*sigma = 2.35482*sigma
      for p = 0, n_elements(peaks)-1 do begin
        maxpix = peaks[p]
        centroidpix = total(y[maxpix-5:maxpix+5]*prof[maxpix-5:maxpix+5])/total(prof[maxpix-5:maxpix+5])
        fitcenter = getlinecenter(y, prof, maxpix,      fitrad=16./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if abs(fitcenter-maxpix) gt 5. then $
          fitcenter = getlinecenter(y, prof, maxpix,      fitrad=30./ybin, peak=peakf, width=width, bkpoly=2, pix=pix,flux=flux,ffit=ffit)   
        if peakf lt 0 or width gt 20./ybin then $
          fitcenter = getlinecenter(y, prof, maxpix,      fitrad=20./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if peakf lt 0 or width lt 1. then $
          fitcenter = getlinecenter(y, prof, maxpix,      fitrad=10./ybin, peak=peakf, width=width, bkpoly=1, pix=pix,flux=flux,ffit=ffit)   
        if width gt 1.5/ybin and width le 30.0/ybin then begin          
          peakcen[p] = fitcenter
          peakwidth[p] = width
        endif else begin
          peakcen[p] = centroidpix
          peakwidth[p] = 6.0/ybin
        endelse
      endfor
      peaksig = (smoothprof[peaks]/rms) ; significance, not sigma
      peakdist = abs(peaks-avstdpos) 
      peakrank = (1./sqrt(abs(peaksig-2))) * ((peakdist*4/ybin) > 5.0)^2  ; lower=more likely to be peak, could explore optimal reweighting here
      ; peakrank = (1./abs(peaksig-2)) * (peakdist > 1.0)^2  ; older weighting
      ss = sort(peakrank) 
      slitwidtharcsec = strnum(strmid(hi.slit,1+strpos(hi.slit,'_')))
      slitwidth = slitwidtharcsec/(0.135*hi.ybin)
      if slitwidth gt 0. then slitfrac = erf(slitwidth/2./peakwidth[ss[0]]/sqrt(2.))  else slitfrac='?' ; assumes a gaussian PSF
      if slitwidth gt 0. then slitloss = 1./slitfrac else slitloss = '?'
      updatefile, profilefile, ['#SIGMA   = '+clip(peakwidth[ss[0]]),'#FWHM    = '+clip(peakwidth[ss[0]]*2.35),$
                                '#SLITFRAC= '+clip(slitfrac),'#SLITLOSS= '+clip(slitloss)]
      objectpos = peakcen[ss[0]]
      objectrad = peakwidth[ss[0]] * 2.0 ; the 'optimum' 1D value is ~1.40 sigma.  ; FWHM would be 2.35 sigma. (full not half!)
      if isstd[i] then objectrad += 2/ybin
      ; bakrad = [0.,0.]
      bakrad = objectrad + 6/ybin + [0,12/ybin]
        ; could do a test here to check for contaminators in the background 
      outstr = clip(filei,35)+' '+fpr(objectpos,5.2)+' '+fpr(objectrad,3.2) + '  ' $
              +clip(clip(fpr(bakrad[0],2.1))+'-'+clip(fpr(bakrad[1],2.1)),12) $
              +removepath(tracefiles[i])  $
              +'  # '
      for p = 1, (n_elements(peaks)-1) < 5 do outstr += clip(fpr(peakcen[ss[p]],4.1))+' '
      if objfilechanged eq 0 then logprint, 'Updating object position file...'
      logprint, '  ', outstr
      updatefile, objfile, outstr
      objfilechanged += 1
   endfor

   forcesameapertures = 0
   if forcesameapertures and (objfilechanged gt 0) then begin
      ; finally, go through the object position file and ensure the same radius(/background) is being used on both blue and red
      ; not tested yet, and probably  makes the file ugly
      oinfo = replicate({file:'', rad:0., bak:'', str:''}, countlines(objfile))
      openr, 1, objfile
      inline = ''
      i = 0
      while not eof(1) do begin
         readf, 1, inline
         inarr = strsplit(inline,/extract)
         oinfo[i].file = inarr[0]
         oinfo[i].rad = strnum(inarr[2])
         oinfo[i].bak = inarr[3]
         oinfo[i].str = inline
         i += 1
      endwhile
      oinfo = oinfo[0:i-1]
      sh = readlrisheaders(oinfo.file)
      for i = 0, n_elements(sh)-2 do begin
         for j = i+1, n_elements(sh)-1 do begin
            if sh[j].mjd gt sh[i].mjd and sh[j].mjd lt sh[u].mjdstop then begin
              if oinfo[i].rad eq oinfo[i].rad*(sh[i].ybin/sh[j].ybin) then continue
              oinfo[j].rad = oinfo[j].rad * sh[i].ybin/sh[j].ybin
              ; if sh[i].ybin eq sh[j].ybin then oinfo[j].bak = oinfo[i].bak ; only for same binning
              inarrpos = strsplit(oinfo[i].str)
              newoutstr = strmid(oinfo[i].str,0, inarrpos[2]) + fpr(objectrad,3.2) + strmid(oinfo[i].str,inarrpos[3])
              updatefile, objfile, newoutstr
           endif
         endfor
      endfor
   endif

   ; slit capture fraction is:  S[-w to +w]             e^(-x^2/2*sigma^2) dx  / S[-infinity to infinity] e^(-x^2/2*sigma)
   ;   z = (x/sigma)/sqrt(2)    S[-w/sigma/sqrt(2), +]  e^(-z^2)               / S[] e^(-z^2)
   ;                            erf(x/sigma/sqrt(2))  ??
   ; erf([slit half-width]/sigma/sqrt2) * erf([aperture radius]/sigma/sqrt2)
   ; slitloss = 1./(erf(slitwidth/2./sigma/sqrt(2.)) * erf(rad/sigma/sqrt(2.))

   dum = file_search(spworkingdir1d+'ls'+camchar+'*.profile', count=ntrfileafter) 

   profileplotfile = 'profiles'+camchar+'.ps' ; at top level
   if ntrfileafter gt ntrfilebefore or file_test(profileplotfile) eq 0 or objfilechanged gt 0 then begin
      ; print, ntrfileafter, ntrfilebefore, objfilechanged
      lrisplotprofiles, camera=camchar
      if file_test('profilesr.ps') and file_test('profilesb.ps') then lrisplotprofiles2
   endif

end


; -------------------------

pro autolrisextract, camera=camera, chip=chip, $
   date=date, yblue=yblue, yred=yred  ; for now these are hidden options

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisconfig, lrisautopath, refdatapath, defaultspath
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

    ; visually display the extraction
 

   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if chipchar eq 'l' then begin
      logprint, 'Automatic extraction of left chip not yet supported.'
      return
   endif
 
   nsuccess = 0


   ; search for science files

   uwildchar = wildchar  ; this is a temporary measure for now until generalized date filtering is implemented
   if n_elements(date) gt 0 then uwildchar = clip(date)+strmid(uwildchar,6)

   ;files = choosefiles(camchar+uwildchar+chipchar+'.fits',spworkingdir+'szfp',spworkingdir+'sfp', ct=ct)
   files = file_search(spworkingdir1d+'ls'+camchar+'*.fits', count=ct)

   ; check for unmatched
   if ct eq 0 then return
   fileh = readlrisheaders(files)

   ; search for arc solutions
   
   ; We want both a 'nightly' file (common to the configuration) and a 'local' file (possibly taken during the night)
   ; with the latter only used for the linear adjustment.

   arcsolfiles = choosefiles(camchar+uwildchar+chipchar+'_arc.sol',spworkingdir1d+'fp', ct=ct)
   if ct eq 0 then begin
      logprint, spworkingdir1d+'fp', camchar+uwildchar+chipchar+'.sol'
      logprint, 'No arc solutions found.  Cannot extract wavelength-calibrated spectra.'
      print
      return
   endif
   narcsol = n_elements(arcsolfiles)
   arcfitsfiles = spworkingdir1d+removepath(repstr(arcsolfiles,'.sol','.fits'))
   arcfitsfilesexist = bytarr(narcsol)
   for f = 0, narcsol-1 do begin
      if file_test(arcfitsfiles[f]) then arcfitsfilesexist[f] = 1 
      if arcfitsfilesexist[f] eq 0 then begin
         logprint, '    Data file ', arcfitsfiles[f], ' corresponding to ', removepath(arcsolfiles[f]), ' is missing - cannot use this solution.'
      endif
   endfor
   if total(arcfitsfilesexist eq 0) gt 0 then $
         logprint, '  Must rerun wavcal step to regenerate these files.'


   arch = readlrisheaders(arcsolfiles, badarclist=badarclist) ; arc metadata
   if camchar eq 'b' then arcnlamps = arch.hg + arch.cd + arch.zn    ; +arc.fe
   if camchar eq 'r' then arcnlamps = arch.hg + arch.neon + arch.ar  ; + arc.kr + arc.xe


   ; ----- Default trace position -----

   ; trace default positions specified in file
   if n_elements(yblue) gt 0 then defaulttraceyblue = yblue 
   if n_elements(yred) gt 0 then  defaulttraceyred  = yred 
   if camchar eq 'b'  then normaltracepos = defaulttraceyblue
   if camchar eq 'r'  then normaltracepos = defaulttraceyred

  
   ; ----- Load tracepos.dat to get custom trace positions --------

   ; old method
   ; if file_test('tracepos.dat') $
   ;    tr = loadtraceposfile('tracepos.dat', ct=ntr)

   objfile = spworkingdir1d+'objectpos.txt'
   userobjfile = 'userobjectpos.txt'

   if file_test(objfile) eq 0 then begin
      logprint, ' No object position file found: ', clip(objfile)
      logprint, ' This is required to extract the spectra.  Run trace step.'
      return
   endif

   tr = loadobjectposfile(objfile, ct=ntr)
   userset = intarr(ntr)
   usertr = loadobjectposfile(userobjfile, ct=nutr)
   for u = 0, nutr-1 do begin
      w = where(usertr[u].file eq tr.file, ctmatch) 
      if ctmatch gt 0 then begin
      if usertr[u].apname eq '' then begin
         ; regular aperture, override what's already in the file
         tr[w[0]] = usertr[u]
         userset[w[0]] = 1
      endif else begin
         ; special (named) aperture
         tr = [tr, usertr[u]]
         userset = [userset, 1]
      endelse
      endif
   endfor

   ; ------ Other final setups ------

   if camchar eq 'r' then grkey = 'GRANAME' else grkey = 'GRISNAME'
   if camchar eq 'r' then filtkey = 'REDFILT' else filtkey = 'BLUFILT'
   if camchar eq 'r' then fltseqkey = 'RFLTSEQF' else fltseqkey = 'BFILTSEQF'

   keywords = ['FILTER','EXPTIME','OBJECT','TARGNAME','ROTPOSN','DICHNAME','GRNAME',grkey,filtkey,'SLITNAME',$
               'INSTRUME','TELESCOP','OBSERVER','REDUCER','LPIPEVER',$
               'ELAPTIME','DATE','UTC','MJD-OBS','JD','JDMID','DATE_BEG','DATE_END',$
               'AZ','EL','RA','DEC','EQUINOX','AIRMASS', $
               'SUNELEV','MOONELEV','MOONDIST','SLITPA','ROTPPOSN','BINNING', $
               'CAMERA','FLATFLE','FLATTYPE','DICHSEQF',fltseqkey, $
               'FRAMENUM','DOMEPOSN','LAMPS','WINDOW','PANE', $               ; for readlrisheaders
               'NAXIS1','NAXIS2','LTV1','LTV2','XBIN','YBIN','PXSCALE','DISPERSN','SATURATE',$
               'COMB'+['0','1','2','3','4','5','6','7','8','9'], $
               'EXPT'+['0','1','2','3','4','5','6','7','8','9'], $
               'UT'+['0','1','2','3','4','5','6','7','8','9'], $
               'AIRMASS'+['0','1','2','3','4','5','6','7','8','9'], $
               'TOTALEXP', 'NSUM']    ; for readlrisheaders
   if camchar eq 'r' then keywords  = [keywords,'GRATSEQF','GRANSEQF','WAVELEN']
   keywords = [keywords, 'CWAVE'] 
   ; could do a uniqueness check to avoid duplicates...
   ;stds = where(fileh.exptime le 60. and fileh.maxcounts-fileh.skycounts gt 6000., complement=notstds, ctstd)
   ;if ctstd gt 0 then files = files[[stds, notstds]]   ; do the obvious standards first if possible


   ; -------- Now do the extraction ------------

   unextracted = ['']
   nunextracted = 0

   for f = 0, n_elements(files)-1 do begin
      filename = files[f]
      if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
      if filename eq '' then continue
      baseoutfile = fileappend(repstr(filename,'.fits', '.spec'),'e')
      ; if overwrite eq 0 and file_test(outfile) then continue

      ; theoretically this is inefficient
      hi = readlrisheaders(filename)

      spatbin = hi.ybin

      if normaltracepos le 0 then begin
        ; if normaltracepos eq 0 then print, 'Default trace position not set!'  ; this is now OK, sort of...
        if normaltracepos lt 0 then logprint, "Trace position can't be negative?"
      endif else begin
         defaultysearch = normaltracepos/spatbin + 22./spatbin*[-1,1]   ; doesn't make much difference when tracing is actually performed.
      endelse

      if n_elements(defaultysearch) gt 0 then ysearch = defaultysearch ; otherwise specextract default, i.e. the whole chip
      if ntr gt 0 then begin
         ta = where(tr.file eq removepath(files[f]), ct)
         ;if ct ge 2 then begin
         ;   print, 'WARNING:  Multiple entries for ', files[f], ' in tracepos.dat'
         ;   ; in the future, this is OK - write multiple spectra out to extract multiple sources.  ; need to create a second loop
         ;endif
         if ct eq 0 then begin
            ; print, 'error message'
            nunextracted += 1
            unextracted = [unextracted, files[f]]
            continue
            ;print, '         Need to rerun trace step, or specify manually?'
         endif
      endif else begin
         logprint, 'No trace data... the pipeline theoretically supports continuing but please run the trace step to enable good behavior.'
         continue
      endelse
  
      for tt = 0, ct-1 do begin

          if n_elements(ycent) gt 0 then delvarx, ycent
          if n_elements(bakdist) gt 0 then delvarx, bakdist
          if n_elements(bakwidth) gt 0 then delvarx, bakwidth
          if n_elements(tracefile) gt 0 then delvarx, tracefile
          if n_elements(traceycent) gt 0 then delvarx, traceycent
          if n_elements(aprad) gt 0 then delvarx, aprad
          if n_elements(fixtrace) gt 0 then delvarx, fixtrace
          if n_elements(ysearch) gt 0 then delvarx, ysearch
          if n_elements(readdsky) gt 0 then delvarx, readdsky

          ; print, '  User-specified trace constraints for ', tr[t].file
          ;  help, tr[t], /str
          t = ta[tt]
          if tr[t].apname ne '' and tr[t].apname ne '0' then outfile = repstr(baseoutfile,'.spec','-'+tr[t].apname+'.spec') else outfile = baseoutfile
          if overwrite eq 0 and file_test(outfile) then continue
          if tr[t].yfix gt 0 then begin
             ycent = tr[t].yfix
             fixtrace = 1
          endif
          if tr[t].ymin ge 0 and tr[t].ymax gt 0 then begin
             ysearch = [tr[t].ymin, tr[t].ymax] 
          endif 
          if max(tr[t].bakmin) ge 0 and max(tr[t].bakmax) ge 0 then begin 
             bakdist = tr[t].bakmin            ;(tr[t].bakmin+tr[t].bakmax)/2.  ; I must have redefined this...?
             bakwidth = (tr[t].bakmax-tr[t].bakmin)
          endif
          if tr[t].tracefile ne '' then    tracefile  = spworkingdir1d+tr[t].tracefile 
          if tr[t].tracefileypos gt 0 then traceycent = tr[t].tracefileypos
          if tr[t].aprad gt 0 then         aprad      = tr[t].aprad
          if tr[t].addsky gt 0 then        readdsky    = 1
          if userset[t] then logprint, '  User-specified trace constraints: ' + $ 
                                   ' y='+clipzero(ycent) + $
                                   ' rad='+clipzero(aprad) + $
                                   ((total(bakdist) gt 0 or total(bakwidth) gt 0) ? ' bak='+$
                                       clip(fpr(tr[t].bakmin[0],2.1))+'-'+clip(fpr(tr[t].bakmax[0],2.1))+'|'+$
                                       clip(fpr(tr[t].bakmin[1],2.1))+'-'+clip(fpr(tr[t].bakmax[1],2.1)) : '') + $
                                   tr[t].apname eq '' ? '' : '  ['+tr[t].apname+']'
                                  ; ' bak='+clip(traceycent) + $
           logprint, 'Extracting ', removepath(filename), ' ('+clip(hi.target)+')'

           ;'~/research/Useful/skyspec/gident_most.dat'
          skylist = refdatapath + '/' + 'skylines.dat'

          if n_elements(arcfile) gt 0 then delvarx, arcfile
          if n_elements(skyfile) gt 0 then delvarx, skyfile
          if n_elements(initsol) gt 0 then delvarx, initsol
          if n_elements(extractarcfile) gt 0 then delvarx, extractarcfile
          setkey = ['']
          setval = ['']

          ; find matching arcs
          matches = where(arch.grname eq hi.grname and $  ; and arch.binning eq hi.binning and arch.window eq hi.window 
                    (abs(arch.cwave-hi.cwave) lt 10.) and $  
                    (arch.dichroic eq hi.dichroic or camchar eq 'r'), ct)

          if ct eq 0 then begin
             logprint, 'WARNING: Found no useable arcs in the same configuration as ', removepath(filename), ' ('+hi.target+')'
             logprint, '     gr='+hi.grname, ' bin='+repstr(hi.binning,',','x'), ' win='+hi.window, ' cwav='+clip(hi.cwave), ' dich='+clip(hi.dichroic)
             ;   print, filename, ' ', gr, ' ', binning, ' ', window, ' ', cwav, ' ', dich
             printarcs = 0
             if printarcs then for a = 0, n_elements(arc)-1 do begin
                logprint, arcsolfiles[a], ' gr=', arch[a].grname, ' bin=', arch[a].binning, ' win=', arch[a].window, ' cwav=', clip(arch[a].cwave), ' dich=', clip(arch[a].dichroic)
             endfor
             logprint, '  Cannot wavelength calibrate.  Skipping...'
             wavsolfail = [wavsolfail, filename]
             continue
             ;print, 'Spectrum will not be wavelength calibrated.'
          endif 


          ; get the 'initial' arc from the start of the night, only used for blue side
          ; for now just look at the first with all lamps, but could see if there is a way to decide on the best one

          w = where(arcnlamps[matches] eq max(arcnlamps[matches]) and abs(arch[matches].cwave-hi.cwave) lt 0.1, ct) 
             ; a better check might be arch[matches].granglef eq hi.granglef
          if ct eq 0 then begin
              logprint, ' WARNING: no arc was taken at this precise grating angle.  This can lead to imperfect wavelength/flux calibration.'
              w = where(arcnlamps[matches] eq max(arcnlamps[matches]), ct) 
          endif
          winit = matches[w[0]]


          ; get the 'local' arc, possibly taken during the night
 
          archere = 0
          w = where(abs(arch[matches].radeg-hi.radeg) lt 0.1 and abs(arch[matches].decdeg-hi.decdeg) lt 0.1 and $
                    abs(arch[matches].mjd-hi.mjd) lt 10./60./24., ct) 
                 ; first see if an arc was taken here and now
          if ct gt 0 then begin
             wlocal = max(w)
             archere = 1
          endif else begin
              ; get an arc not near horizon, if we can.
             w = where(arch[matches].airmass lt 3.0, ct)
             if ct gt 0 then matches = matches[w]

             ; as many lamps as possible
             w = where(arcnlamps[matches] eq max(arcnlamps[matches]), ct) 
             matches = matches[w]

             ; at this exact central wavelength, if possible
             w = where(abs(arch[matches].cwave-hi.cwave) lt 0.1, ct)
             if ct gt 0 then matches = matches[w]

             ; as close to this elevation as possible.  [ really should look at drive angle as well if not zenith ]
             w = where(abs(arch[matches].airmass - hi.airmass) eq min(abs(arch[matches].airmass-hi.airmass)))
             wlocal = matches[max(w)]

             ; could also write info about the fit to the solfile and choose the best-fitting arc.

          endelse
          delvarx, w

          ; Need a tool for wavelength refinement using only the mercury lamp (for blue morning twilight standards).
          ; Probably best done within autowavcal

          arcinitfitsfile = arcfitsfiles[winit]
          arcinithi = arch[winit]
          solinitfile = arcsolfiles[winit]
      
          arcfitsfile = arcfitsfiles[wlocal]
          archi = arch[wlocal]
          solfile = arcsolfiles[wlocal]

          arcsol = grabcolumn(solfile,0)
          for o = 0, n_elements(arcsol)-1 do arcsol[o] *= (hi.xbin/archi.xbin)^o
          if solinitfile ne solfile then begin
             ; use a common set of higher-order terms
             logprint, '  Loading arc solutions from '+removepath(solinitfile)+' and '+removepath(solfile)
             arcinitsol = grabcolumn(solinitfile,0)
                   for o = 0, n_elements(arcinitsol)-1 do arcinitsol[o] *= (hi.xbin/arcinithi.xbin)^o
             initsol = [arcsol[0], arcinitsol[1:*]]
             ; print, 'Reference cwave='+clip(arcinitsol[0])+',  local cwave='+clip(arcsol[0])+', diff = '+clip(arcsol[0]-arcinitsol[0])
          endif else begin
             initsol = arcsol
             logprint, '  Loading arc solution from '+removepath(solfile)
          endelse

          setkey = [setkey, 'ARCFILE0', 'SOLFILE0', 'ARCFILE', 'SOLFILE']
          setval = [setval, arcinitfitsfile, solinitfile, arcfitsfile, solfile]


          dispersion = hi.dispersion
          cwav = hi.cwave

          if camchar eq 'r' and hi.exptime le 9.0 then  skylist = ''   ; don't sky-refine short exposures
          if camchar eq 'b' and hi.exptime le 16.0 then skylist = ''  ;

          if camchar eq 'b' and hi.sunelev gt -11 then skylist = '' ; don't sky-refine twilight exposures.  
          if camchar eq 'r' and hi.sunelev gt -6  then skylist = ''
          if skylist ne '' then skyfile = 1 ; 
           ; if skylist eq '' and archere then begin
               ; in development
           ;     if file_test(arcfile) then skyfile = arcfitsfile 
           ;     skylist = lines
           ; endif
                                   ;This is conservative, better solar-subtraction will enable improvement.

          if camchar eq 'r' then maxnudge = 220.
          if camchar eq 'b' then maxnudge = 3.    ; actually 10 more than this, this is for the search

          uncrefwav = 0.3   ; still needed?

          if hi.grname eq '400/3400' then xsearch=hi.nx*[0.65,0.9] else xsearch=-1 ; works as default

          if hi.exptime le 120. and n_elements(aprad) eq 0 then begin
               ; This appears to be deprecated code from before the standards were incorporated into objectpos.txt.
               ; It isn't triggered unless the standard is missing from that file (which probably triggers errors earlier)
               standardpath = refdatapath
               readstandard, hi.target, /check, result=isstandard, path=standardpath+'/'
               if isstandard eq 0 and hi.exptime le 80. then begin  ; catch objects with nonstandard targnames
                  readstandard, hi.object, /check, result=isstandard, path=standardpath+'/'
               endif  
               if isstandard then aprad = 2.0/hi.pxscale
          endif

          errorstatus = 0
          if debug eq 0 then catch, errorstatus
          if errorstatus ne 0 then begin
             logprint, 'Unexpected error during spectrum extraction:'
             logprint, !error_state.msg
             success = -10
          endif else begin
             if n_elements(tracefile) gt 0 then logprint, '    Trace file: ', tracefile
             specextract, filename, ycent, xsearch=xsearch, $ 
             ysearch=ysearch, aprad=aprad, bakdist=bakdist, bakwidth=bakwidth, tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace, $
             skyfile=skyfile, skylist=skylist, $
             initsol=initsol, cwav=cwav, dispersion=dispersion, unccwav=cwav*0.04, uncdisp=0.075*dispersion, uncrefwav=uncrefwav, $
             outfile=outfile, keywords=keywords, minorder=7, maxorder=7, success=success, maxnudge=maxnudge, setkey=setkey, setval=setval, $
             readdsky=readdsky
             ; arcfile=extractarcfile, arclist=arclist, brightarclist=brightarclist - don't solve at-position arcs any more...
             ; probably lots of other keywords are no longer needed too: disp, unccwav, uncdisp, uncrefwav...
          endelse
          catch, /cancel 
          print

          ; success:  1=good, 2=questionablewavsol, -1=filenotfound, -2=badfile, -3=tracefail, -4=wavsolfail
          if success lt 0 and success ne -4 then extractfail = [extractfail, filename]
          if success eq -4 then wavsolfail = [wavsolfail, filename]
          if success eq 2 then wavsolwarn = [wavsolwarn, filename]
          if success gt 0 then nsuccess += 1
      endfor  ; sub-apertures
   endfor   ; fits files

   catch, errorstatus
   if errorstatus ne 0 then begin
        logprint, '   Error opening display window.'
        logprint, '   Turning off window display.'
        displayoff = 1
   endif

   if nunextracted gt 0 then logprint, '  ', clip(nunextracted)+' '+strupcase(clip(camera))+' spectra have no trace data and cannot be extracted.'

   if nsuccess gt 0 then begin
      ; update the plots if any new extractions performed
      if displayoff eq 0 then lrisplottraces, cam=camera   
      lrisplotprofiles, camera=camchar  ; user could have changed the object positions.
   endif

   catch, /cancel 

end


; -------------------------

pro autolrisresponse, camera=camera, chip=chip, files=files, notelluric=notelluric

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   common lrisconfig, lrisautopath, refdatapath, defaultspath

   standardpath = refdatapath
    
   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if n_elements(files) eq 0 then $
       files = file_search(spworkingdir1d+'els'+camchar+'*.spec', count=ct)

   if files[0] eq '' and n_elements(files) eq 1 then return

   imh = readlrisheaders(files)

   ; identify flux calibrators.

   wfluxstd = [-1]

   nstdobs = 0
   for f = 0, n_elements(files)-1 do begin
      if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue

      outrespfile = repstr(files[f],'.spec','.response')
      if file_test(outrespfile) and overwrite eq 0 then continue

      ; Check if target is in standards archive based on its name/position, and get its standardized name
      readstandard, imh[f].target, pos=[imh[f].radeg, imh[f].decdeg], stdname=stdname, /check, result=result, path=standardpath+'/'

      if result eq 0 then continue ; not a standard

      obs = readspec(files[f], h, /timenorm, cutoff=3000.)

      ; cwav = sxpar(h,'CWAV')   ; not needed here
      nwav = sxpar(h,'NWAV',count=ctnwav)      
      if ctnwav gt 0 and nwav eq 0 then continue

      nsaturate = sxpar(h,'NSATURAT')
      if nsaturate gt 3 then continue

      name = stdname


      ; 'obs' is the observation of the standard.
      ; 'ref' is the reference of the standard.
      logprint, '     Calibrator  ', removepath(files[f]), ' ', stdname, ' @ air=', fpr(imh[f].airmass,1.3)

      wav = obs.wav        ; may want to consider oversampling this in the future?
      obsflux = obs.flux
    
      readstandard, stdname, refwav_in, refflux_in, unit='flambda/ang', path=standardpath+'/', stdname=stdnameconfirm
      ;print, '  Read in ', imh[f].target, ' as ', stdname, ' ', stdnameconfirm
      ;print, min(wav), max(Wav), min(refwav_in), max(refwav_in)
      if max(refwav_in) lt max(wav) then begin
         if max(refwav_in) lt max(wav)-100 then begin
            logprint, 'WARNING: archival standard spectrum for '+clip(stdname)+' does not cover full wavelength range!'
            logprint, '         Extrapolating to the red with a Rayleigh-Jeans law, but caution should be used...'
         endif
         addwav = max(wav) +  100 * findgen(1 + fix((max(wav)-max(refwav_in))/ 100))
         addflux = refflux_in[n_elements(refflux_in)-1]* (addwav/refwav_in[n_elements(refflux_in)-1])^(-4.)
         refwav_in = [refwav_in, addwav]
         refflux_in = [refflux_in, addflux]
      endif
      refflux = interpol2(refflux_in, refwav_in, wav)

      ; remove lines
      lookupstandardlines, stdname, lines, linev, linesnt, linevnt ; get list of absorption lines in this standard star
      removelines, wav, obsflux, obsflux_nl, lines=lines, vel=linev, /interpol      
      removelines, wav, refflux, refflux_nl, lines=linesnt, vel=linevnt, /interpol  ; (tellurics already removed from ref)
 
      ; response defined as: observed count rate (e/s) per f_lambda
      response =  obsflux_nl / refflux_nl   ; instrument+atmosphere response except tellurics

      ; airmass correction, currently not in use.  Note also in telluric section.
      ; airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
      ; response *= 2.51^(-(refair-stdair[0])*airtermarr)
      ; wsignal = where(response gt 0.01*max(response), ct)
      ; if ct le 1 then begin
      ;      print, 'ERROR: Something wrong with flux calibration...'
      ;      continue
      ; endif

      ; here would go the dichroic corrections, etc., to smooth out some of the complex wiggles before fitting.

      ; fit a polynomial function to the response to smooth out noise
      bad = where(response lt median(response)*1e-4, ct, complement=good) ; just in case any negatives snuck through
      if ct gt 0 then response[bad] = min(response[good])
      par = [0,0,0]
      n = n_elements(wav)
      c = median(wav)
      rightbuf = 3
      if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then rightbuf=25
      if clip(sxpar(h,'INSTRUME')) eq 'LRISBLUE' and clip(sxpar(h,'DICHNAME')) eq '500' and clip(sxpar(h,'GRISNAME')) eq '600/4000' then rightbuf=280 ; total hack
      r = median(response[5:n_elements(wav)-rightbuf])
      for o = 3, 36 do begin  ; previously 18
        par = polyfitg((wav[5:n-rightbuf]-c)/1e3, median(alog((response[5:n-rightbuf]/r)),3), o, guess=[par, 0], chi2=chi2) 
               ; in theory I shouldn't be using the interpolated regions for this however bad things happen with the fit if I don't
        ;print, o, ':', par
        ;print, o, chi2
      endfor
      modelresponse = r*exp(poly((wav-c)/1e3, par))

      ; telluric correction
      removelines, wav, obs.flux, obsflux_nlt, lines=linesnt, vel=linevnt, /interpol  ; "t" version keeps tellurics
      if keyword_set(notelluric) then begin
         telluricresponse=1 
      endif else begin
         telluricresponse = obsflux_nlt / obsflux_nl  ; response due solely to tellurics
         ; telluricresponse = 1 < (telluricresponse)^(refair/stdair[0]) ; airmass correction
      endelse

      openw, 1, outrespfile
      printf, 1, '# AIRMASS  = '+clip(imh[f].airmass)
      printf, 1, '# wav          obsflux          refflux          response         telluricresponse'
      for i = 0, n_elements(wav)-1 do $
         printf, 1, wav[i], obsflux[i], refflux[i], modelresponse[i], telluricresponse[i]
      close, 1   

      for pl=0,1 do begin ;0=screen, 1=eps
        if displayoff and pl eq 0 then pl = 1

        xsize = 600

        if pl eq 0 then begin
           errorstatus = 0
           catch, errorstatus
           if errorstatus ne 0 then begin
              logprint, ' X-windows display not working: turning off display'
              displayoff = 1
              catch, /cancel
              continue
           endif else begin
              window, 0, xsize=xsize, ysize=800
           endelse
           catch, /cancel
           ;           -                  511          124890        391482905
           cols = [rgb(255,255,255), rgb(255,1,0), rgb(218,231,1), rgb(25,142,85), rgb(255,255,0), rgb(0,128,0)]
        endif else begin
           plotfile = repstr(files[f],'.spec','_response.eps')
           psopen, plotfile, xsize=(xsize/100), ysize=8, /inches, /encaps
           device, /color
           !p.font=0
           device, /times, font_index=17  ; the stupid new default plot axis format for scientific notation
                                          ; sets it to times even if not requested!!! can't use other fonts :(
           colors = transpose([[0,0,0],$
                      [255,0,0],$ 
                      [100,100,0],$ 
                      [0,200,0],$ 
                      [160,160,0],$  
                      [10,255,10]])
           tvlct, colors
           cols = indgen(6)
        endelse
        !p.multi = [2,1,2]

        xl = 0.1
        xr = 0.98
        resp = response
        telresp = telluricresponse
        name = stdname
        file = files[f]
 
        xrange = [min(wav), max(wav)]

        !p.position = [xl, 0.8, xr, 0.99]
        if pl eq 1 then title = '!17' else title=''
        plot, [0], [0], xrange=xrange, yrange=[0, 1.02*max([refflux_nl])], /xstyle, /ystyle, title=title
        if n_elements(origwav) eq 0 then delvarx, origwav
        oplot, refwav_in, refflux_in, color=cols[1]
        oplot, wav, refflux_nl
        xyouts, xr-(xr-xl)*0.35, !p.position[3]-0.02, /norm, title+name+' (database)'
 
        !p.position = [xl, 0.5, xr, 0.78]
        plot, [0], [0], yrange=[0, max([obs.flux,obsflux_nl])], xrange=xrange, /xstyle, title=title
        ; could add time-normalized right axis
        oplot, obs.wav, obs.flux, color=cols[1] ; noninterpolated
        oplot, wav, obsflux_nl                  ; interpolated (no difference...)
        xyouts, xr-(xr-xl)*0.65, !p.position[3]-0.02, /norm, title+name+' ('+removepath(file)+')'
        !p.multi = [2,1,2]

        !p.position = [xl, 0.20, xr, 0.48]
        plot, [0], [0], xrange=xrange, yrange=[0, max(response)*1.05], /xsty, /ysty, title=title;, /ylog
        ; if nmatchstd gt 1 then oplot, wav, resp*telresp, color=cols[3]
        oplot, wav, response
        oplot, wav, modelresponse*telluricresponse, color=cols[2]
        xyouts, xl+(xr-xl)*0.05, !p.position[3]-0.02, /norm, title+'Solved response', color=cols[4]
        xyouts, xl+(xr-xl)*0.05, !p.position[3]-0.035, /norm, title+'Measured response', charsize=0.85

        if pl eq 1 then psclose
        if pl eq 1 then retitleplot, plotfile, removepath(files[f])+' response'

      endfor ; plot loop
      !p.multi = 0
      !p.position = 0


   endfor

end

; ------------------------

; -------------------------

pro autolrisfluxcal, camera=camera, chip=chip, files=files, notelluric=notelluric

   ; New feature needed - need to try to fluxcal using only standards taken in the same spectroscopic block
   ; (LRIS is extremely bad about returning to the desired grating angle...)


   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   common lrisconfig, lrisautopath, refdatapath, defaultspath

   standardpath = refdatapath


   camchar = strmid(camera,0,1)
   chipchar = strmid(chip,0,1)

   if n_elements(files) eq 0 then $
       files = file_search(spworkingdir1d+'els'+camchar+'*.spec', count=ct)
;      files = choosefiles(camchar+wildchar+chipchar+'.spec',spworkingdir1d+'eszfp',spworkingdir1d+'esfp')
   nf = n_elements(files)
   if nf eq 1 and files[0] eq '' then return

   imh = readlrisheaders(files)

   names = imh.target
   isfluxcal = intarr(nf)
   issci = intarr(nf)
   cwavs = fltarr(nf)  ; measured value

   for f = 0, nf-1 do begin
      h = headspec(files[f])
      cwavs[f] = sxpar(h,'CWAV') ; the solved/measured value is CWAV, while CWAVE is from header (dumb convention, need to change...)

      readstandard, imh[f].target, pos=[imh[f].radeg, imh[f].decdeg], stdname=stdname, /check, result=result, path=standardpath+'/'
      ; print, imh[f].target, imh[f].radeg, imh[f].decdeg, result
      if result then names[f] = stdname 
     
      outrespfile = repstr(files[f],'.spec','.response')
      if file_test(outrespfile) then begin
         isfluxcal[f] = 1  ; response file means it's a usable standard
      endif else begin
         ; Either a bad standard, or a science file
         if result eq 0 then issci[f] = 1
      endelse
   endfor



   wfluxcal = where(isfluxcal, ctfluxcal)
   wscience = where(issci, ctsci)
   if ctsci eq 0 then begin
      logprint, 'Found no science observations!'
      return
   endif
   if ctsci gt 0 and ctfluxcal eq 0 then begin
      logprint, 'Found no LRIS-'+strupcase(camchar)+' response files.  Run response step or acquire standard-star observations.'
      ;print, 'Found no LRIS-'+strupcase(camchar)+' standard-star observations!'
      ;print, '  If standards were taken, check that ', standardpath, ' is a valid directory containing the standard-star library.'
      return
   endif

   nsci = n_elements(wscience)
   
   greatstandard = (strpos(names,'G191') ge 0) or (strpos(names,'BD+28') ge 0) or (strpos(names,'Feige34') ge 0) or $
                   (strpos(names,'Feige67') ge 0) or (strpos(names,'Feige110') ge 0 )  

   for s = 0, nsci-1 do begin
      wsci = wscience[s] ; scalar
      filename = files[wsci]

      outfile = fileappend(filename,'c')

      if file_test(outfile) and overwrite eq 0 then continue

      sameconfig = imh[wfluxcal].filter eq imh[wsci].filter and imh[wfluxcal].grname eq imh[wsci].grname and $
                   imh[wfluxcal].slit eq imh[wsci].slit and imh[wfluxcal].dichroic eq imh[wsci].dichroic
                ; always required
      airdiff = abs(imh[wfluxcal].airmass-imh[wsci].airmass)
      timediff = abs( imh[wfluxcal].mjd - imh[wsci].mjd)/24.

       ; Try to determine appropriate standards to use.
       ; other things to check would be whether or not the wavelength solution is precise, how bright twilight was,
       ; and the actual wavelength solution 
       ; Using hard cuts like this is non-optimal - should instead try a goodness metric.
       ; Also, should try to choose the same standard for both observations of the same target...

         matchstandards = where(sameconfig and timediff lt 3. and airdiff lt 0.35 and greatstandard[wfluxcal] and $ 
                                abs(cwavs[wfluxcal]-cwavs[wsci] lt 10.), nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and timediff lt 5. and airdiff lt 0.35 and greatstandard[wfluxcal] and $ 
                                abs(cwavs[wfluxcal]-cwavs[wsci] lt 15.), nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and timediff lt 5. and airdiff lt 0.6 and greatstandard[wfluxcal] and $ 
                                abs(cwavs[wfluxcal]-cwavs[wsci] lt 15.), nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and abs(cwavs[wfluxcal]-cwavs[wsci] lt 15.), nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and abs(cwavs[wfluxcal]-cwavs[wsci] lt 50.), nmatchstd)
      if nmatchstd eq 0 then $
         matchstandards = where(sameconfig and abs(cwavs[wfluxcal]-cwavs[wsci] lt 200.), nmatchstd)
      ; progressively search for standards with less and less similar central wavelength (since note that LRIS shifts)

      if nmatchstd eq 0 then begin
         logprint, 'No matching standard for ', removepath(files[wsci]), ' ('+imh[wsci].grname+', '+imh[wsci].filter+', D'+clip(imh[wsci].dichroic)+')'
         fluxcalfail = [fluxcalfail, files[wsci]]
         continue
      endif

      ; actually read in the science frame
      sci = readspec(files[wsci], h, /timenorm, cutoff=3000.)
      wav = sci.wav
      nwav = sxpar(h,'NWAV',count=ctnwav)
      if ctnwav gt 0 and nwav eq 0 then begin  
         logprint, filename, ' has no wavelength calibration.'
         continue
      endif

      ; a slit loss correction would probably go here.

      sciair = imh[wsci].airmass
      logprint, 'Flux-calibrating ', removepath(files[wsci]), ' (',imh[wsci].target,', '+imh[wsci].grname+')  airmass=', fpr(sciair,1.4)
      stdair = imh[wfluxcal[matchstandards]].airmass

         ;for n2 = 3, 20 do print,( 911.760 * 1./(1./n1^2 - 1./n2^2) ) / (1.0 + 2.735182E-4 + 131.4182 / vac^2 + 2.76249E8 / vac^4)

      elev = 4200.
      airp = 2^((4200-elev)/4570.)*[0.12947983, -0.13458371, 0.21833111, -0.48222307, $
                                    0.41800934, -0.11447838, 0.0032198187, -0.010178335, 0.0047375054]

      ; shut off the multiple standard matching by uncommenting this block
      ; if nmatchstd gt 1 then begin
      ;    wstd = wfluxcal[matchstandards[(where(stdair eq min(stdair))) [0]]]
      ;    nmatchstd = 1
      ; endif

      if nmatchstd eq 1 then begin
         ;if abs(sciair-stdair[0]) gt 0.1 then begin
         ;   print, '     Warning: Only one standard-star observation in this configuration.'
         ;   print, '     Airmass:  science ('+fpr(sciair,2.2)+'), standard ('+fpr(stdair[0],2.2)+')'
         ;endif

         wstd = wfluxcal[matchstandards[0]]

         logprint, '     Calibrator  ', removepath(files[wstd]), ' ', names[wstd], ' @ air=', fpr(imh[wstd].airmass,1.3)

         if abs(sciair-stdair[0]) gt 0.25 then begin
            adiff = abs(sciair-stdair[0])
            if adiff lt 0.4 then adj = 'Significant'
            if adiff ge 0.4 and adiff lt 0.65 then adj = 'Large'
            if adiff gt 0.65 and adiff lt 1.0 then adj = 'VERY large'
            if adiff gt 1.0 then adj = 'EXTREMELY large'
            logprint, '     Warning:  '+adj+' airmass difference between science target ('+clip(fpr(sciair,2.2))+ $ 
                                   ') and standard ('+clip(fpr(stdair[0],2.2))+')'
         endif

         responsefile = repstr(files[wstd],'.spec','.response')
         readcol, responsefile, rwav, rcounts, rflux, rresponse, rtelluricresponse, format='f,f,f,f,f', comment='#', /silent
         rresponse *= (imh[wsci].xbin*1.0/imh[wstd].xbin)*(imh[wsci].ybin*1.0/imh[wstd].ybin)
         response = interpol2(rresponse, rwav, wav)
         telluricresponse = interpol2(rtelluricresponse, rwav, wav) < 1

         ; airmass correction (using a standard function)
         airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
         ; print, ' airmass difference' , sciair-imh[wstd].airmass
         response *= 2.51^(-(sciair-imh[wstd].airmass)*airtermarr) 
         telluricresponse = telluricresponse^(sciair/imh[wstd].airmass)

         sxaddpar, h, 'FLUXCAL1', removepath(files[wstd])
         sxaddpar, h, 'FLUXNAM1', names[wstd]
         sxaddpar, h, 'FLUXAIR1', imh[wstd].airmass

      endif else begin
        if nmatchstd eq 2 then begin
            wlo = wfluxcal[matchstandards[(where(stdair eq min(stdair))) [0]]]
            whi = wfluxcal[matchstandards[(where(stdair eq max(stdair))) [0]]]
        endif

        if nmatchstd ge 3 then begin
           below = where(stdair lt sciair, ctbelow)
           above = where(stdair gt sciair, ctabove)
           ; try to bracket the source if possible
           if ctbelow ge 1 and ctabove gt 1 then begin
              wlo = wfluxcal[matchstandards[below[(where(stdair[below] eq max(stdair[below]))) [0]]]]
              whi = wfluxcal[matchstandards[above[(where(stdair[above] eq min(stdair[above]))) [0]]]]
           endif else begin
              wlo = wfluxcal[matchstandards[(where(stdair eq min(stdair))) [0]]]
              whi = wfluxcal[matchstandards[(where(stdair eq max(stdair))) [0]]]
           endelse
        endif
 
        logprint, '     Calibrators ', removepath(files[wlo]), ' ', names[wlo], ' @ air='+fpr(imh[wlo].airmass,1.3)
        logprint, '                 ', removepath(files[whi]), ' ', names[whi], ' @ air='+fpr(imh[whi].airmass,1.3)

        responsefilelo = repstr(files[wlo],'.spec','.response')
        responsefilehi = repstr(files[wlo],'.spec','.response')

        readcol, responsefilelo, rwavlo, rcountslo, rfluxlo, rresponselo, rtelluricresponselo, format='f,f,f,f,f', comment='#'
        readcol, responsefilehi, rwavhi, rcountshi, rfluxhi, rresponsehi, rtelluricresponsehi, format='f,f,f,f,f', comment='#'
        rresponselo *= (imh[wsci].xbin*1.0/imh[wlo].xbin)*(imh[wsci].ybin*1.0/imh[wlo].ybin)
        rresponsehi *= (imh[wsci].xbin*1.0/imh[whi].xbin)*(imh[wsci].ybin*1.0/imh[whi].ybin)

        responselo = interpol2(rresponselo, rwavlo, wav)
        responsehi = interpol2(rresponsehi, rwavhi, wav)
        telluricresponselo = interpol2(rtelluricresponselo, rwavlo, wav) < 1
        telluricresponsehi = interpol2(rtelluricresponsehi, rwavhi, wav) < 1

        ; airmass correction (using a standard function)
        airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
        responselo *= 2.51^(-(sciair-imh[wlo].airmass)*airtermarr) 
        responsehi *= 2.51^(-(sciair-imh[whi].airmass)*airtermarr) 
        telluricresponselo = telluricresponselo^(sciair/imh[wlo].airmass)
        telluricresponsehi = telluricresponsehi^(sciair/imh[whi].airmass)

 
        ; average the response curves together with a normalization.
         ; calculate weights based on airmass similarity
        weightlo = 1./abs(sciair-imh[wlo].airmass)
        weighthi = 1./abs(sciair-imh[whi].airmass)
        weightlo /= (weightlo+weighthi)
        weighthi = 1.-weightlo ; won't have to do any division below since this is enforced
         ; normalize and then average response
        normlo = percentile75(responselo)
        normhi = percentile75(responsehi)  
        norm = (normlo*weightlo + normhi*weighthi)
        responselo *= norm/normlo
        responsehi *= norm/normhi
        response = (responselo*weightlo + responsehi*weighthi)
         ; average telluric response
        telluricresponse = (telluricresponselo*weightlo + telluricresponsehi*weighthi)

        sxaddpar, h, 'FLUXCAL1', files[wlo]
        sxaddpar, h, 'FLUXCAL2', files[whi]
        sxaddpar, h, 'FLUXAIR1', imh[wlo].airmass
        sxaddpar, h, 'FLUXAIR2', imh[whi].airmass
        sxaddpar, h, 'FLUXNAM1', names[wlo]
        sxaddpar, h, 'FLUXNAM2', names[whi]
        sxaddpar, h, 'FLUXWT1',  weightlo
        sxaddpar, h, 'FLUXWT2',  weighthi
      endelse

      wsignal = where(response gt 0.01*max(response), ct)
      if camchar eq 'r' then if max(sci[wsignal].pix) eq max(sci.pix) then $
        wsignal = wsignal[0:n_elements(wsignal)-2] ; last pixel is usually bad on red
      if ct le 1 then begin
           logprint, 'ERROR: Something wrong with flux calibration...'
           continue
      endif

      outspec = sci[wsignal]

      outspec.resp = response[wsignal] * telluricresponse[wsignal]
      outspec.flux /= outspec.resp
      outspec.sky  /= outspec.resp
      outspec.unc  /= outspec.resp

      writespec, outfile, outspec, h

   endfor

end

; -----------------------------

pro autolriscombine, camera=camera, live=live

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

   common lrisconfig, lrisautopath, refdatapath, defaultspath

   standardpath = refdatapath

   ; automatic combination of 1D spectra of a common source in a common setting.

   ; this function is largely unnecessary now that things are summed in 2D space.
   ; However, it's still useful if observations aren't summed and you want them to be summed.

   ; assumes the central wavelength wasn't wildly adjusted between exposures.
 
   ; live mode doesn't do anything yet, and hasn't been tested at any rate.

   camchar = strmid(camera,0,1)
   chipchar = 'r'

   ;files = choosefiles(camchar+wildchar+chipchar+'.spec',spworkingdir1d+'ceszfp',spworkingdir1d+'cesfp')
   files = file_search(spworkingdir1d+'cels'+camchar+'*.spec', count=ct)
   nf = n_elements(files)

   if nf eq 1 and files[0] eq '' then return

   targets = strarr(nf)
   igrs = strarr(nf)
   ; worry about different cwavs?  probably not since it can shift slightly.

   for f = 0, n_elements(files)-1 do begin
     filename = files[f]
     if n_elements(fileseq) gt 0 then if total(filenum(files[f]) eq fileseq) eq 0 then continue
     h = headspec(filename)
     target = clip(sxpar(h,'TARGNAME'))
     if strpos(target,'_S') gt 0 then begin  ; strip the offset star from the target info
        ustriptarget = strmid(target,0,strpos(target,'_', /reverse_search))
        ;if ustriptarget ne 'HD' and ustriptarget ne 'BD' and ustriptarget ne 'PTF' then 
        target = ustriptarget
     endif
     targets[f] = target
     inst = clip(sxpar(h,'INSTRUME'))
     gr = clip(sxpar(h,'GRNAME'))
     incamchar = 'u' ; unknown
     if inst eq 'LRIS' then incamchar = 'r'
     if inst eq 'LRISBLUE' then incamchar = 'b'
     igrs[f] = incamchar + gr ; this is sort of obsolete now, but allows this to work with camera='?'
   endfor

   targetlist = unique(targets)
   ntargets = n_elements(targetlist)

   for t = 0, ntargets-1 do begin 
      target = targetlist[t]
      readstandard, target, /check, result=isstandard, path=standardpath+'/'
      if isstandard then continue ; don't stack observations of flux standards

      wtarget = where(targets eq target, nexp)
      
      ; determine intra-color normalization

      igrlist = unique(igrs[wtarget])
      ngr = n_elements(igrlist) 

      minwav = fltarr(ngr)
      maxwav = fltarr(ngr)
      for g = 0, ngr-1 do begin
         igr = igrlist[g]
         cam = strmid(igr,0,1)
         gr  = strmid(igr,1)
               ; make this a legal filename.  should really create a funciton to do this.
         ;outfile =  spworkingdir1d+'vc'+repstr(repstr(repstr(target,'_',''),'*',''),'?','')+'_'+repstr(igr,'/','-')+'.spec'

         wtargigr = where(targets eq target and igrs eq igr, ngexp)
         if ngexp le 1 then continue
         outfile = files[wtargigr[0]]
         for si = 2, 10 do outfile = repstr(outfile,'_s'+clip(si)+'.spec','.spec')
         outfile = repstr(outfile,'.spec','_sa.spec')
         if file_test(outfile) and overwrite eq 0 then continue
         logprint, outfile
            

         fluxnormval = fltarr(ngexp)
         totexp = 0.
         logprint, ' Combining ', format='($,A)'
         c = 0
         for i = 0, ngexp-1 do begin
            f = wtargigr[i]
            logprint, removepath(files[f])+' ', format='($,A)'
            s = readspec(files[f], h)
            inexp = sxpar(h,'ELAPTIME')

            if i eq 0 then sums = s
            if i gt 0 then begin
               sums = addspec(sums, s, w1=totexp, w2=inexp)
            endif
            totexp += inexp

            ; header info 
            if i eq 0 then begin
               outh = h  
            endif

            sxaddpar, outh, 'CFILE'+clip(i+1), files[f], i eq 0 ? 'summed 1D spectrum file name' : ''

            ;if ngexp gt 1 then begin
            sxaddpar, outh, 'APRAD'+clip(i+1), sxpar(h,'APRAD')
            sxaddpar, outh, 'YEXTRAC'+clip(i+1), sxpar(h,'YEXTRACT')
            sxaddpar, outh, 'EXPTIME'+clip(i+1), inexp
            ;endif

            for cc = 0, 9 do begin
                combcc = sxpar(h, 'COMB'+clip(cc+1), count=ctcc)
                ; if cc eq 0 then print, cc+1, combcc, ctcc
                if ctcc gt 0 then begin
                   sxaddpar, outh, 'COMB'+clip(c+1), combcc, removepath(files[f]), after='COMB'+clip(c)
                   sxaddpar, outh, 'EXPT'+clip(c+1), sxpar(h,'EXPT'+clip(cc+1)), after='EXPT'+clip(c)
                   sxaddpar, outh, 'AIRMASS'+clip(c+1), sxpar(h,'AIRMASS'+clip(cc+1)), after='AIRMASS'+clip(c)
                   sxaddpar, outh, 'UT'+clip(c+1), sxpar(h,'UT'+clip(cc+1)), after='UT'+clip(c)
                   c += 1
                endif
            endfor


         endfor
         sxaddpar, outh, 'NSUM', ngexp
         sxaddpar, outh, 'DATE_END', sxpar(h, 'DATE_END')
         sxaddpar, outh, 'AIRMASS', mean([sxpar(h, 'AIRMASS') , sxpar(outh,'AIRMASS')])
         sxaddpar, outh, 'AZ', mean([sxpar(h, 'AZ') , sxpar(outh,'AZ')])
         sxaddpar, outh, 'EL', mean([sxpar(h, 'EL') , sxpar(outh,'EL')])
         sxaddpar, outh, 'EXPTIME',  totexp
         sxaddpar, outh, 'TOTALEXP', totexp
         logprint
         logprint, '           -> ', removepath(outfile)

         writespec, outfile, sums, outh

      endfor

   endfor

end



pro autolrisconnect, files=files, noscale=noscale, live=live

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

   ; this can combine the red file with multiple different blue files.  Need to come up with a workaround.
   ; Probably scan through the blues and identify overlapping red matches, checking them off.

   if n_elements(files) eq 0 then $
;      files = file_search(spworkingdir1d+'vc*.spec')
       files = file_search(spworkingdir1d+'cels*.spec', count=ct)
   files = files[sort(files)]
   nf = n_elements(files)

   if nf eq 1 and files[0] eq '' then return

   targets = strarr(nf)

   for f = 0, n_elements(files)-1 do begin
     filename = files[f]
     h = headspec(filename)
     target = clip(sxpar(h,'TARGNAME'))
     if strpos(target,'_S') gt 0 then begin
        posstarget = strmid(target,0,strpos(target,'_S'))  ; offset stars
        if posstarget ne 'HD' and posstarget ne 'BD' and posstarget ne 'PTF' then target = posstarget
     endif
     targets[f] = target 
   endfor

   targetlist = unique(targets)
   ntargets = n_elements(targetlist)

   for t = 0, ntargets-1 do begin 
      target = targetlist[t]

      outfile = spfinaldir+'lris'+target+'.spec'
      outfile = repstr(repstr(repstr(repstr(outfile,' ','_'),'*',''),'?',''),'/','')   ; make filename legal
      if file_test(outfile) and overwrite eq 0 then continue

      wtarget = where(targets eq target, nexp)

      vcfiles = files[wtarget] ; all observations of this target
      hinfo = readlrisheaders(vcfiles)
      exptimes = float(hinfo.exptime)
      cams = hinfo.camera
      utstart = fltarr(n_elements(hinfo))
      utstop  = fltarr(n_elements(hinfo))
      for i = 0, n_elements(hinfo)-1 do utstart[i] = ten(hinfo[i].ut)
      for i = 0, n_elements(hinfo)-1 do utstop[i]  = ten(hinfo[i].ut)+hinfo[i].elaptime/3600. ; use mj-end??

      wb = where(cams eq 'blue', ctblue)
      wr = where(cams eq 'red', ctred)

      if ctblue eq 0 or ctred eq 0 then begin
         logprint, ' No match for: '
         for j = 0, nexp-1 do logprint, '   '+removepath(vcfiles[j])+' '+fpr(hinfo[j].exptime,5.1)+'s'+'  '+hinfo[j].grname+'  '+clip(fix(hinfo[j].cwave))
         continue
      endif

      if nexp gt 2 then begin 
          ; Multiple exposures on the same target weren't summed together.
          ; Could be different epochs, different orientations, or different targets in the same field.

          ; check for 1D sums (_sa means sum of everything)
          wb = where(cams eq 'blue' and strpos(vcfiles,'_sa.spec') gt 0, ctb)
          wr = where(cams eq 'red' and strpos(vcfiles,'_sa.spec') gt 0, ctr)
          if ctb eq 1 and ctr eq 1 then begin
              vcfiles = vcfiles[[wb[0], wr[0]]]  ; skip all the stuff below (don't actually sum individuals, just the sum)
              nexp = 2
          endif
      endif
      if nexp gt 2 then begin
          ; Didn't do 1D sums (for at least one set), so we have to figure out how to pair things up.
          if ctb eq 0 then wb = where(cams eq 'blue', ctb)
          if ctr eq 0 then wr = where(cams eq 'red', ctr)
          logprint, '  Multiple summed exposures exist for '+target+' ('+clip(ctb)+' blue, '+clip(ctr)+' red)'
          if ctb ne ctr then begin
             logprint, '  Warning:  Some exposures will not be connected.'
             logprint, '            Sum in 1D space, or delete unwanted frames?'
          endif
          for i = 0, (ctb < ctr)-1 do begin
             ; Things get complicated if there are different numbers of spectra even at this stage.
             ; Loop over the shorter list and match each with the longest overlapping exposure on the longer list.

             eoutfile = repstr(outfile,'.spec','.'+clip(i)+'.spec')
             eplotfile = spworkingdir1d+'rescale_'+target+clip(i)+'.eps'
             if file_test(eoutfile) and overwrite eq 0 then continue

             if ctb le ctr then ii = wb[i] else ii = wr[i]
             if ctb le ctr then wc = wr    else wc = wb
             ;print, removepath(vcfiles[ii]), '   ', utstart[ii], utstop[ii]
             ;for j = 0, n_Elements(wc)-1 do print, '   ', clip(removepath(vcfiles[wc[j]]),30), utstart[wc[j]], utstop[wc[j]]
             overlap = where((utstart[wc] gt utstart[ii] and utstart[wc] lt utstop[ii]) or $
                             (utstop[wc]  gt utstart[ii] and utstop[wc]  lt utstop[ii]), ctover)
             if ctover eq 0 then begin
                logprint, '  Having difficulty matching multiple spectra of '+target
                continue
             endif 
             wcoverlap = wc[overlap]
             wbest = wcoverlap[ (where(exptimes[wcoverlap] eq max(exptimes[wcoverlap]))) [0] ]

             if (ctb le ctr) then lrisconnect, vcfiles[ii], vcfiles[wbest], eoutfile, eplotfile, displayoff=displayoff $
                             else lrisconnect, vcfiles[wbest], vcfiles[ii], eoutfile, eplotfile, displayoff=displayoff
          endfor
          continue
      endif 
      if nexp eq 2 then begin   
          lrisconnect, vcfiles[0], vcfiles[1], outfile, spworkingdir1d+'rescale_'+target+'.eps', displayoff=displayoff
      endif
      if nexp eq 1 then begin ; currently this is impossible
          logprint, ' Only one side present for '+vcfiles[0]+' - trivial connection' ;, format='($,A)'
          logprint, '             -> '+outfile
          ; I guess we could modify connect to be compatible with only one file again to get the headers...
          if file_test(outfile) and overwrite then spawn, 'rm -f '+outfile
          spawn, 'cp '+vcfiles[0]+' '+outfile
     endif

   endfor  ; target loop

end


; -------------------------
pro autolrisastrometry, camera=camera, chip=chip, nocrclean=nocrclean, nodistort=nodistort

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail
   
    achip = strmid(chip,0,1)
    prefchar = strmid(camera,0,1)

    if keyword_set(nocrclean) eq 0 then begin
       zffiles = choosefiles(prefchar+wildchar+achip+'.fits',imworkingdir+'zifp',imworkingdir+'zfp')
       if lrisversion eq 1 and camera eq 'red' then $
         allzffiles = choosefiles('?'+wildchar+'.fits',imworkingdir+'zifp',imworkingdir+'zfp') else $
         allzffiles = choosefiles('?'+wildchar+'?'+'.fits',imworkingdir+'zifp',imworkingdir+'zfp')
    endif else begin
       zffiles = choosefiles(prefchar+wildchar+achip+'.fits',imworkingdir+'ifp',imworkingdir+'fp')
       allzffiles = choosefiles('?'+wildchar+'?'+'.fits',imworkingdir+'ifp',imworkingdir+'fp')
    endelse

    ; "all" files include the other camera and chips, and is for the initial catalog lookup.

    if n_elements(allzffiles) eq 1 and removepath(allzffiles[0]) eq '' then return

    allhinfo = readlrisheaders(allzffiles)
    allfiletargets = strarr(n_elements(allhinfo))
    for f = 0, n_elements(allzffiles)-1 do begin
       if allzffiles[f] eq '' then continue
       h = headfits(allzffiles[f], /silent)
       allfiletargets[f] = repstr(repstr(strtrim(allhinfo[f].target,2),' ', '_'),'/','_')  ; spaces cause barfing in filenames
    endfor

    alltargets = unique(allfiletargets)
    catalogs = ['sdss','usno','2mass']
   
    ; Get catalog reference data for every field (avoids having to do repeated lookups later)
    ;   and reduces risk that SDSS will start rejecting the catalog queries)
    ;   Note that this is chip/camera-independent    

    for t = 0, n_elements(alltargets)-1 do begin
      wthistarget = where(allfiletargets eq alltargets[t], ntargfiles)
      if ntargfiles eq 0 then continue ; shouldn't happen
      for c = 0, n_elements(catalogs)-1 do begin
         cat = catalogs[c]
         ; saves the number of images in the filename in case more appear later
         targname = alltargets[t]  ; name already repaired
         outcatfile = imworkingdir+targname+'.'+string(ntargfiles,'(I03)')+'.'+cat+'.ccat'
         if file_test(outcatfile) eq 0 then begin
            ; allow for a pointing error of up to 2'.  (another ~1-2' also OK but only partial overlap)
            logprint, '  Looking up '+strupcase(cat)+' catalog file for '+alltargets[t]+$
                   ' ('+clip(ntargfiles)+' files)' ;, format='($,A)'
            ; print, allzffiles[wthistarget] ; debug
            ; continue                       ; debug
            medra = median([allhinfo[wthistarget].radeg], /even)
            meddec = median([allhinfo[wthistarget].decdeg], /even)
            sepra = (max(allhinfo[wthistarget].radeg)-min(allhinfo[wthistarget].radeg))*cos(meddec*!pi/180.)
            sepdec = max(allhinfo[wthistarget].decdeg)-min(allhinfo[wthistarget].decdeg)
            if sepra gt 60*20 or sepdec gt 60*20 then begin
                 logprint, '   WARNING - Files of target '+alltargets[t]+' are quite widely spaced'
            endif
            if sepra gt 60*60 or sepdec gt 60*60 then begin
                 logprint, '   ERROR   - Files of target '+alltargets[t]+' are more than a degree apart!'
                 continue
            endif
            printcatalog, allzffiles[wthistarget], cat, edgebuffer=-889, /unc, out=outcatfile
            ; Sometimes the catalog is on the edge of an SDSS field - trash it if nothing within 1' of the object.
            if file_test(outcatfile) and cat eq 'sdss' then begin
               readcol, outcatfile, ccra, ccdec, /silent, format='d,d'
               cdist  = 3600 * sqrt( ((ccra-medra)*cos(meddec*!pi/180.))^2 + (ccdec-meddec)^2 )
               mindist = min(cdist)
               if mindist gt 60 then begin
                  logprint, '  SDSS appears to only partially cover this field; not saving the file.'
                  spawn, 'rm -f '+outcatfile
               endif
            endif
            if file_test(outcatfile) then begin
               ; delete any earlier versions with fewer images included
               others = file_search(imworkingdir+targname+'.*.'+cat+'.ccat')
               for j = 0, n_elements(others)-1 do if others[j] ne outcatfile then spawn, 'rm -f '+others[j]

               ; also write out a region file (doesn't need the number)
               readcol, outcatfile, ccra, ccdec, /silent, format='d,d'
               outcatregfile = imworkingdir+targname+'.'+cat+'.ccat.reg'
               ;print, '  Converting '+clip(outcatfile)+' to region file '+clip(outcatregfile)
               openw, 1, outcatregfile
               printf, 1, 'global color=green width=1'
               printf, 1, 'fk5'
               for i = 0, n_elements(ccra)-1 do begin
                  printf, 1, 'point('+clip(ccra[i])+','+clip(ccdec[i])+') # point=boxcircle'
               endfor
               close, 1
            endif else begin
               if cat eq 'sdss' then begin
                   openw, 1, outcatfile
                   printf, 1, '# SDSS query unsuccessful - probably not an SDSS field'
                   close, 1
                   ;print, ' (unsuccessful)'
               endif else begin
                   logprint, '  WARNING - unsuccessful - problem with internet connection?' ;, format='($,A)'
               endelse
            endelse
         endif
      endfor
    endfor



    ; Solve every image (of this chip and camera) directly against SDSS/USNO

    hinfo = readlrisheaders(zffiles)

    if overwrite eq 0 then zffiles = unmatched(zffiles,'a')  ; catalogs should always be the same; only do this now

    for f = 0, n_elements(zffiles)-1 do begin
       if zffiles[f] eq '' then continue
       outfile = fileappend(zffiles[f],'a')
       if file_test(outfile) and overwrite eq 0 then continue
       h = headfits(zffiles[f], /silent)
       pa = strtrim(string((float(sxpar(h, 'ROTPOSN'))+0) MOD 360),2)
       exptime = sxpar(h,'ELAPTIME')
       counts = sxpar(h,'COUNTS')
       if counts gt 30000 and exptime le 10 or counts gt 20000 and exptime le 5 or counts gt 15000 and exptime le 4 then begin
          logprint, zffiles[f], ' is in bright twilight; not solving astrometry.'
          continue
       endif

       targname = repstr(strtrim(sxpar(h,'TARGNAME'),2),' ', '_')  ; spaces cause barfing in filenames
       if targname eq '' then continue
       if strpos(targname,'flat') ge 0 then continue

       ; Find the "best" catalog
       refcatfile = ''
       backuprefcatfile = ''
       for c = 0, n_elements(catalogs)-1 do begin
          ; in case there are multiple versions, find the newest
          refcatfiles = file_search(imworkingdir+targname+'.*.'+catalogs[c]+'.ccat')
          tryrefcatfile = refcatfiles[n_elements(refcatfiles)-1]
          if tryrefcatfile eq '' then logprint, "  ERROR - can't find the catalog file...?"
          nline = countlines(tryrefcatfile)
          ;print, '  ', tryrefcatfile, nline
          if nline gt 4+1 then begin
             if refcatfile ne '' and backuprefcatfile eq '' then backuprefcatfile = tryrefcatfile
             if refcatfile eq '' then refcatfile = tryrefcatfile
          endif
       endfor

       sat = sxpar(h,'SATURATE')
       if sat eq 0 then sat = 60000
       if refcatfile eq '' then catstr = '' else catstr = ' -c '+refcatfile
       if refcatfile eq '' then backupcatstr = '' else backupcatstr = ' -c '+backuprefcatfile
       for i = 0, 2 do begin
          if i eq 0 and strpos(refcatfile,'sdss') eq -1 then continue  ; skip the run without saturated stars on USNO/2MASS
          if i eq 0 then $
             acommand =  autoastrocommand+' '+zffiles[f] + catstr+' -upa 1.0' + ' -x '+clip(sat)+' -q'
          if i eq 1 then $
             acommand =  autoastrocommand+' '+zffiles[f] + catstr+' -upa 1.0' + ' -q'
          if i eq 2 then $
             acommand =  autoastrocommand+' '+zffiles[f] + backupcatstr+' -upa 1.0' + ' -q'
          logprint, acommand
          spawn, acommand
          if file_test(outfile) then begin
              sip2pv, outfile, /modify  ; could skip this if not SIP
              break  ; don't do the run with saturated stars if the first one succeeds
          endif
       endfor
       if file_test(outfile) eq 0 then begin
          fullastrofail = [fullastrofail, zffiles[f]]
          continue
       endif
 
    endfor

    ; future improvement: should go through fullastrofail list and align against neighboring images.






    ; Align images against each other, including distortion terms up to 4th order.
    ; 4th does not actually appear to be quite high enough, but the SIP conversion code only works to 4th.
    ; This is essentially a separate step, but still within lrisautoastrometry for now.

    ; For non-mosaic fields (the only capability at present), start with a central image 
    ; and work outward.


    afiles = choosefiles(prefchar+wildchar+achip+'.fits',imworkingdir+'azifp',imworkingdir+'azfp') 

    if keyword_set(nocrclean) eq 0 then begin
       afiles = choosefiles(prefchar+wildchar+achip+'.fits',imworkingdir+'azifp',imworkingdir+'azfp')
    endif else begin
       afiles = choosefiles(prefchar+wildchar+achip+'.fits',imworkingdir+'azifp',imworkingdir+'azfp')
    endelse
    if overwrite eq 0 then uafiles = unmatched(afiles,'e') else uafiles = afiles
    if n_elements(uafiles) eq 1 and uafiles[0] eq '' then return

    ahinfo = readlrisheaders(afiles, /crval)     ; all the headers get read in every time, I guess this is OK.


    targetstorealign = unique(ahinfo.target) ; for now
    distort = keyword_set(nodistort) eq 0
    
    for t = 0, n_elements(targetstorealign)-1 do begin
       if targetstorealign[t] eq '' then continue

       w = where(ahinfo.target eq targetstorealign[t], ct)
       if ct eq 0 then logprint, '????'
       tahinfo = ahinfo[w]
       tafiles = afiles[w]

       uatfiles = unmatched(tafiles,'e')
       if n_elements(uatfiles) eq 1 and uatfiles[0] eq '' then continue
       if max(tahinfo.exp) lt 60. then continue  ; don't bother to realign e.g. standards
       logprint
       logprint, 'Aligning images of '+targetstorealign[t]

       cosdec = cos(median([tahinfo[w].decdeg])*!pi/180.)    
       dr = (tahinfo[w].radeg - median([tahinfo[w].radeg],/even))*cosdec*3600.
       dd = tahinfo[w].decdeg - median([tahinfo[w].decdeg],/even)*3600.
       dc = sqrt(dr^2 + dd^2)
       score = dc ;+ sqrt(((ahinfo[w].exptime < 100) > 20.) - 100)

       firstimagei = (where(score eq min(score))) [0] ; index
       firstimagename = tafiles[ firstimagei ]
       alignstatus = intarr(ct)
       alignstatus[firstimagei] = 1

       hfirst = headfits(firstimagename)
       acatfile = sxpar(hfirst,'ASTR_CAT')
       logprint, '  Starting realignment with '+removepath(firstimagename)+' vs. catalog '+removepath(acatfile)
       if file_test(acatfile) eq 0 then acatfile = imworkingdir+acatfile
       acatnum  = sxpar(hfirst,'ASTR_NUM') ; if this is high enough could re-distort here too?
       outfile = fileappend(firstimagename,'e')
       if file_test(acatfile) then $
          realignimage, firstimagename, acatfile, mindist=0.5, redistort=0  ; consider turning on redistort for SDSS with 30+ matches
       if file_test(outfile) eq 0 then begin
          logprint, 'Not re-aligning the first image.'
          spawn, 'cp -p '+firstimagename+' '+outfile
       endif else begin
          if distort then sip2pv, outfile, /modify
       endelse

       for i = 1, ct-1 do begin ; 0 is already done
          aligned = where(alignstatus eq 1, naligned, complement=unaligned, ncomplement=nunaligned)
          if naligned eq 0 or nunaligned eq 0 then begin
            logprint, 'Everything is aligned?'
            continue
          endif
          ;print, 'unaligned: ', tafiles[unaligned]
          ;print, 'aligned: ', tafiles[aligned]
          mind = 60.    ; don't trust this process for large offsets
          bestd = 180.*3600.
          bestj = -1
          bestk = -1
          for j = 0, n_elements(unaligned)-1 do begin
             for k = 0, n_elements(aligned)-1 do begin
                d = sqrt( ((tahinfo[unaligned[j]].radeg -tahinfo[aligned[k]].radeg)*cosdec)^2 + $  ; meridian cross?
                           (tahinfo[unaligned[j]].decdeg-tahinfo[aligned[k]].decdeg)^2 ) * 3600.
                ;print, removepath(tafiles[unaligned[j]]), ' - ', removepath(tafiles[aligned[k]]), ' ', d
                if d lt mind then begin
                   bestj = j
                   bestk = k
                   mind = d
                endif
                if d lt bestd then bestd = d
             endfor
          endfor

          if bestj ge 0 then begin
             infile =  tafiles[unaligned[bestj]]
             reffile = fileappend(tafiles[aligned[bestk]],'e')
             outfile = fileappend(infile,'e')
             logprint, '  Aligning '+removepath(infile)+' against '+removepath(reffile)+' (sep = '+clip(bestd)+')'
             realignimage, infile, reffile, redistort=distort
             ;;spawn, 'cp -f '+fileappend(tafiles[unaligned[bestj]],'e')+' '+tafiles[unaligned[bestj]]
             if file_test(outfile) and distort then sip2pv, outfile, /modify
             alignstatus[unaligned[bestj]] = 1
          endif else begin
             logprint, '  Remaining images of this field are too far apart for relative alignment (minsep = '+clip(bestd)+')'
          endelse
       endfor

    endfor



end



; -------------------------
pro autolrisphotometry, camera=camera, chip=chip, photometric=photometric

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

   common lrisconfig, lrisautopath, refdatapath, defaultspath

   if n_elements(chip) eq 0 then chipchar = '?' else chipchar = strmid(chip,0,1)
   camchar = strmid(camera,0,1)

   files = choosefiles(camchar+wildchar+chip+'.fits',  imworkingdir+'eazfp', imworkingdir+'azfp',imworkingdir+'eafp',imworkingdir+'afp')
   if n_elements(files) eq 1 and files[0] eq '' then return

   if camera eq 'blue' then istr = 'lrisb'
   if camera eq 'red'  then istr = 'lrisr'

   calibratenight, instrument='LRIS', istr=istr, files, workingdir=imworkingdir, photometric=photometric, archivezpt=refdatapath+'/zeropoints.dat'

end




; -------------------------
pro autolrisphotcat

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion

   ; Last step of the photo pipeline combines red/blue left/right and needs to be run separately.

   ; merge all tcal files (red AND blue, left AND right) together!

   tcalfiles = [file_search(imworkingdir+'a*f*'+'?'+wildchar+'?'+'.tcal'), file_search(imworkingdir+'a*f*'+'?'+wildchar+'?'+'.tcaz')]
   w = where(tcalfiles ne '', ct)
   if ct eq '' then return

   tcalfiles = tcalfiles[w]
   tcalinfo = readcatheader(tcalfiles, count=ct)
   
   printed = 0
   targets = unique(tcalinfo.target)
   for t = 0, n_elements(targets)-1 do begin
       wt = where(tcalinfo.target eq targets[t])
       targetoutcat = imfinaldir+'stars'+repstr(targets[t],' ','_')+'.cat'
       if file_test(targetoutcat) then continue  ; could check contents as well.
       if printed eq 0 then logprint, 'Combining photometric catalogs...'
       printed = 1
       logprint, '     ', targets[t]
       mergecat, tcalfiles[wt], out=targetoutcat  ; would also be nice to write out the origin of the calib, etc.
   endfor


  

end


; need to re-enable use of photometric scaling

; -------------------------
pro autolrisstack, camera=camera, chip=chip, live=live, final=final

   common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
   common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

  if file_test('default.swarp') eq 0 then $
     spawn, swarpcommand+' -d > default.swarp'

  if n_elements(chip) eq '' then chip = '*'
  chipchar = strmid(chip,0,1)
  prefchar = strmid(camera,0,1)

  latesttarget = ''
  if keyword_set(live) then begin   ; if "live" reduction, find what target was most recently read out
    rawfiles = file_search(datadir+prefchar+wildchar+chipchar+'.fits')
    if n_elements(rawfiles) gt 1 then begin
       rawfiles = rawfiles[sort(rawfiles)]                ; this is not a perfect solution for truly continuous running.
       latestfile = rawfiles[n_elements(rawfiles)-1]      ; the appearance of raw-data may outrun the data processing.
       h = headfits(latestfile)                          
       if sxpar(h,'SLITNAME') eq 'DIRECT' then $
         latesttarget = repstr(strtrim(sxpar(h,'TARGNAME'),2), ' ', '_')
    endif
  endif

  azffiles = file_search(imworkingdir+'ea*'+prefchar+wildchar+chipchar+'.fits')
  realfiles = where(azffiles ne '', ct)

  nfiles = ct
  if nfiles eq 0 then return ; can't stack if there are no astrometry files
  if nfiles ge 1 then azffiles = azffiles[realfiles]

  ; make a custom structure for this (but, in principle could update readlrisheaders)
  info = replicate({filename:'', target:'', exp:0., filter:'', satval:0., skyval:0., skyrate:0., airmass:0., seepix:0., zpt:0., $
                    rzpt:0., trans:0., reltrans:0., utstart:'', utend:'', jdstart:0.d, jdend:0.d, mjdstart:0.d, mjdend:0.d, $
                    sunelev:0., moonelev:0.}, nfiles)

  for i = 0, nfiles-1 do begin
     h = headfits(azffiles[i], /silent)
     info[i].filename = removepath(azffiles[i])
     info[i].target = repstr(strtrim(sxpar(h,'TARGNAME'),2), ' ', '_')
     info[i].exp    = round(sxpar(h,'EXPTIME')*10.)/10.  ; round to 0.1 sec
     info[i].filter = sxpar(h,'FILTER')
     info[i].satval = sxpar(h,'SATURATE')
     info[i].skyval = sxpar(h,'SKYCTS')
     info[i].skyrate= sxpar(h,'CTRATE')
     info[i].airmass= sxpar(h,'AIRMASS')
     info[i].seepix = sxpar(h,'SEEING')
     info[i].zpt    = sxpar(h,'ZEROPT')
     info[i].rzpt   = sxpar(h,'RZPT')
     info[i].trans  = sxpar(h,'TRANS')
     info[i].reltrans=sxpar(h,'RELTRANS')
     info[i].utstart =sxpar(h,'UTC')
     info[i].utend   =sxpar(h,'UTC-END')
     info[i].jdstart =sxpar(h,'JD')
     info[i].jdend   =info[i].jdstart+info[i].exp/(24*3600.)
     info[i].mjdstart=double(sxpar(h,'MJD'))
     info[i].mjdend  =info[i].mjdstart+info[i].exp/(24*3600.)
     info[i].sunelev =sxpar(h,'SUNELEV')
     info[i].moonelev=sxpar(h,'MOONELEV')
   endfor  

   keepkeystr = 'OBJECT,TARGNAME,TELESCOP,INSTRUME,CURRINST,FILTER,DICHNAME,SLITNAME,'+$
                'GRISNAME,GRANAME,GRANGLE,UT,UTC,JD,MJD_OBS,MJD-OBS,ST,'+$
                'DATE,DATE-OBS,AIRMASS,AZ,RA,DEC,EL,HA,TELFOCUS,ROTPOSN,DOMEPOSN,CURRINST,OBSERVER,REDUCER,LPIPEVER'+$
                'DOMEPOSN,ADCWAVE0,ADCWAVE1,AMPMODE,CCDGAIN,CCDSPEED,NUMAMPS'+$
                'DATE_BEG,BINNING,TRAPDOOR,LAMPS,FLIMAGIN,FLSPECTR'+$
                'FLATFLD,FLATTYPE,NSUPFLT,CALFILT,MOONDIST,SUNELEV,MOONELEV'
   keepkeys = strsplit(keepkeystr, ',', /extract)


   if file_test('extendedsources.txt') eq 0 then begin
      extendedsources = [''] 
   endif else begin
      readcol, 'extendedsources.txt', extendedsources, format='a', /silent
      notblank = where(extendedsources ne '', ct)
      if ct gt 0 then extendedsources = extendedsources[notblank]
   endelse


   targets = unique(info.target)
   w = where(targets ne latesttarget, ct)  ; if we are still integrating on this target, don't stack it yet
   if ct eq 0 then return
   targets = targets[w]

   for t = 0, n_elements(targets)-1 do begin
     target = targets[t]
     thistarget = where(info.target eq target, cttarg)
     if cttarg eq 0 then continue
     thistargetfilts = unique(info[thistarget].filter)
     for l = 0, n_elements(thistargetfilts)-1 do begin
        filter = thistargetfilts[l]
        thistargfilt = where(info.target eq target and info.filter eq filter, cttargfilt)
        if cttargfilt eq 0 then continue

        outfile       = imworkingdir + 'coadd' + strtrim(target,2) +'_'+ strtrim(filter,2) + '.' + chipchar + '.fits'
        outweightfile = imworkingdir + 'coadd' + strtrim(target,2) +'_'+ strtrim(filter,2) + '.' + chipchar + '.weight.fits'
        if file_test(outfile) and overwrite eq 0 then continue

        if max(info[thistargfilt].exp) lt 59 then continue
        ;expthresh = (median([info[thistargfilt].exp], /even) * 0.5) > (max(info[thistargfilt].exp) * 0.19) ; formerly median*0.8
        ;ctratethresh = 2 * median([info[thistargfilt].skyrate], /even)
        ;ww = where(info[thistargfilt].reltrans gt 0.2)
        ;;reltransthresh = 0.5*median([info[thistargfilt[ww]].reltrans], /even)
        ;expxtransthresh =  0.3*max([info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)]) > $
        ;                   0.7*median([info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)], /even)
        ; should do a full sort by contribution and determine what actually helps/hurts the S/N...
        ;print, info[thistargfilt[ww]].reltrans * sqrt(info[thistargfilt[ww]].exp)
        ;print, expxtransthresh

        tinfo = info[thistargfilt]
        relvar = tinfo.skyrate / (tinfo.exp * tinfo.reltrans^2)  ; scaled variance of the sky in this image
        badphot = where(finite(relvar) eq 0, ct)
        if ct gt 0 then logprint, 'WARNING: ',clip(ct),' files do not have usable photometric header information.'
        if ct gt 0 then relvar[badphot] = tinfo[badphot].skyrate / tinfo[badphot].exp   ; images for which photometry failed
        
        ord = sort(relvar)                                    ; lowest to highest variance (best to worst)
        prevcumnoise = !values.f_infinity
        for j = 0, cttargfilt-1 do begin                         ; could also incorporate seeing into this...
           cumnoise = sqrt( total( relvar[ord[0:j]]) ) / (j+1)
           ;;print, tinfo[ord[j]].filename, relvar[ord[j]], cumnoise
           if cumnoise gt prevcumnoise then break
           prevcumnoise = cumnoise
        endfor
        nstack = j
        stacki = thistargfilt[ord[0:j-1]]
        stacki = stacki[sort(stacki)]
        nexclude = cttargfilt-nstack
        if nexclude gt 0 then excludei = thistargfilt[ord[j:cttargfilt-1]]

        ;stacki = where(info.target eq target and info.filter eq filter and $
        ;               info.exp gt expthresh and info.skyrate le ctratethresh and $
        ;               info.reltrans*sqrt(info.exp) ge expxtransthresh, nstack)
        ;excludei = where((info.target eq target and info.filter eq filter) and $
        ;               (info.exp le expthresh or info.skyrate gt ctratethresh or $
        ;               info.reltrans*sqrt(info.exp) lt expxtransthresh), nexclude)
        ;if nstack lt 2 then continue
        if nstack eq 0 then continue
        ;;if nstack eq 1 then print, 'Only one deep exposure:  trivial stack'
        stacklist = azffiles[stacki]
        if nexclude gt 0 then excludelist = azffiles[excludei]
        totalexp = total(info[stacki].exp)
        medianexp = median([info[stacki].exp], /even)
        medair = median([info[stacki].airmass], /even)
        minair = min(info[stacki].airmass)
        maxair = max(info[stacki].airmass)
        minseeingpix = min(info[stacki].seepix)
        maxseeingpix = max(info[stacki].seepix)
        medseeingpix = median([info[stacki].seepix], /even)
        mintrans = min(info[stacki].trans)
        maxtrans = max(info[stacki].trans)
        medtrans = median([info[stacki].trans], /even)
        minreltrans = min(info[stacki].reltrans)
        maxreltrans = max(info[stacki].reltrans)
        medreltrans = median([info[stacki].reltrans], /even)        
        if mintrans gt 0 then transratio = maxreltrans/minreltrans else transratio = !values.f_nan

        if nstack gt 1 then stdtrans = stdev(info[stacki].reltrans) else stdtrans = 0

        stackcmd = swarpcommand+' '    ;'swarp '
        for s = 0, nstack-1 do begin
           if s eq 0 then stackcmd = stackcmd + stacklist[s] 
           if s gt 0 then stackcmd = stackcmd + ',' + stacklist[s] 
        endfor
        for s = 0, nstack-1 do begin  ; this looks like an opaque way to convert zfpXXX.weight.fits to azfpXXX.wei...
           weightfilename = strmid(stacklist[s],0,strlen(stacklist[s])-5-0)  + '.weight.fits'
           weightexists = file_test(weightfilename)
           if weightexists eq 0 then begin
              slashpos = strpos(weightfilename,'/',/reverse_search)
              weightfilenameinit = imworkingdir + strmid(weightfilename,slashpos+2)  
              weightexists = file_test(weightfilenameinit)
              if weightexists then begin
                  logprint, 'mv '+weightfilenameinit+' '+weightfilename
                  spawn, 'mv '+weightfilenameinit+' '+weightfilename
              endif
           endif
        endfor

        stackcmd = stackcmd + ' -IMAGEOUT_NAME ' + outfile + ' -WEIGHTOUT_NAME ' + outweightfile
        stackcmd = stackcmd + ' -WEIGHT_TYPE NONE'  ; weighting is now turned off.
        ;if nstack gt 1 then begin ; used to be 3...
        ;  stackcmd = stackcmd + ' -WEIGHT_TYPE MAP_WEIGHT'
        ;endif else begin
        ;  stackcmd = stackcmd + ' -WEIGHT_TYPE BACKGROUND'
        ;endelse

        ;stackcmd = stackcmd + ' -VERBOSE_TYPE QUIET'

        dum = where(target eq extendedsources, extct)
        if extct gt 0 then stackcmd += ' -BACK_SIZE 2048'

        if nstack gt 1 then begin
          if transratio le 1.07 or finite(transratio) eq 0 then fluxscales = medianexp/info[stacki].exp $
                                                           else fluxscales = medianexp/(info[stacki].exp*info[stacki].reltrans)
        endif else begin
          fluxscales = 1
        endelse

        if nstack gt 1 then begin
          stackcmd = stackcmd + ' -FSCALE_DEFAULT '
          for s = 0, nstack-1 do stackcmd = stackcmd + clip(fluxscales[s]) + ','   ; multiplicative factor...
        endif
 
        indmodzeropts = info[stacki].zpt + 2.5*alog10(fluxscales*(info[stacki].exp/medianexp))  ; divide the exptime ratios back out since zeropt includes this...
        outzeropt = median([indmodzeropts], /even)

          ; Too many keywords causes pyfits to be unable to load the mosaic.  It's rather freakish.
          ; Also, these all have the same irritating conversion to scientific notation with absurd sigdigits.
          ; make this a string and add to the swarpscikey iter below.
        ;stackcmd = stackcmd + ' -BACK_SIZE 256'

        if (file_test(outfile) eq 0) or overwrite then begin
           if nstack eq 1 then $ ; used to be 3 for some bizarre reason?
             logprint, 'Warning - only ', clip(nstack), ' exposures; not flagging bad pixels.'
           logprint, 'Stacking ', target, ', ', clip(filter),'-band:'
           for s = 0, nstack-1 do logprint, ' ', removepath(stacklist[s]), ' x ', clip(fluxscales[s],5), $
                                         ' (exp =', fpr(info[stacki[s]].exp,4.1) + 's', $
                                         ', rtrans = ', clip(info[stacki[s]].reltrans,4), $
                                         ', zpt = ', fpr(info[stacki[s]].zpt,2.3), $
                                         ', skyr = ', clip(info[stacki[s]].skyrate,4) + 'ct/s)'
           if nexclude gt 0 then logprint, '   Excluding:'
           for x = 0, nexclude-1 do logprint, ' ', removepath(excludelist[x]), '   ', clip('',5), $
                                         ' (exp =', fpr(info[excludei[x]].exp,4.1) + 's', $
                                         ', rtrans = ', clip(info[excludei[x]].reltrans,4), $
                                         ', zpt = ', fpr(info[excludei[x]].zpt,2.3), $
                                         ', skyr = ', clip(info[excludei[x]].skyrate,4) + 'ct/s)'
           logprint, stackcmd
           spawn, stackcmd  ; do the stack
           data = mrdfits(outfile, 0, h, /silent)
           weightdata = readfits(outweightfile, /silent)
           wnodata = where(weightdata eq 0, ct)
           if ct gt 0 then data[wnodata] = !values.f_nan  ; dangerous to keep these at zero like swarp prefers
           swarpscikeys = ['CRPIX1','CRPIX2','EXPTIME','GAIN']
           h0 = headfits(stacklist[0])
           for k = 0, n_elements(swarpscikeys)-1 do sxaddpar, h, swarpscikeys[k], sxpar(h, swarpscikeys[k], comment=cmt), cmt 
           for k = 0, n_elements(keepkeys)-1 do     sxaddpar, h, keepkeys[k],     sxpar(h0, keepkeys[k],     comment=cmt), cmt 

           ;dates = unique(info[stacki].date, ctd)
           ;if ctd gt 0 then for d = 0, ctd-1 do sxaddpar, h, 'DATE'+clip(d), dates[d]  ; in case imaged on multiple nights
           sxaddpar, h, 'UTSTART', info[stacki[0]].utstart
           sxaddpar, h, 'UTSTOP',  info[stacki[nstack-1]].utend
           sxaddpar, h, 'MJDSTART', info[stacki[0]].mjdstart
           sxaddpar, h, 'MJDSTOP',  info[stacki[nstack-1]].mjdend
           sxaddpar, h, 'JDSTART', info[stacki[0]].jdstart
           sxaddpar, h, 'JDSTOP',  info[stacki[nstack-1]].jdend
           ;sxaddpar, h, 'TELAPSED', 
           sxaddpar, h, 'NSTACK', nstack, '# of exposures'
           sxaddpar, h, 'AIRMASS', medair, 'Median exposure airmass', after='NSTACK'
           sxaddpar, h, 'AIRMIN', minair, 'Minimum exposure airmass'
           sxaddpar, h, 'AIRMAX', maxair, 'Maximum exposure airmass'
           sxaddpar, h, 'EXPOSURE', medianexp, 'Effective rescaled exposure time'
           sxaddpar, h, 'SWARPEXP', sxpar(h, 'EXPTIME'), 'Maximum equivalent exposure time (s)'
           sxaddpar, h, 'EXPTIME', medianexp, 'Effective rescaled exposure time', after='EXPOSURE' ; this is relevant for ZEROPT - overwrites SWARP value
           sxaddpar, h, 'TOTALEXP', totalexp, 'Total summed integration time'
           sxaddpar, h, 'MAXEXP', max(info[stacki].exp), 'Length of longest exposure'
           sxaddpar, h, 'MINEXP', min(info[stacki].exp), 'Length of shortest exposure'
           sxaddpar, h, 'SATURATE', min(info[stacki].satval-info[stacki].skyval)
           sxaddpar, h, 'MEDSKY', median([info[stacki].skyval], /even)
           sxaddpar, h, 'SUNELEV', max(info[stacki].sunelev), 'Highest sun elevation'
           sxaddpar, h, 'MOONELEV', max(info[stacki].moonelev), 'Highest moon elevation'
           sxaddpar, h, 'ZEROPT', outzeropt, 'Effective image zeropoint of stack'
           sxaddpar, h, 'RZPT', median([info[stacki].rzpt], /even), 'Aperture radius for zeropoint (pixels)'
           if minseeingpix gt 0 then begin
              sxaddpar, h, 'SEEING', medseeingpix, 'Median seeing in pixels'
              sxaddpar, h, 'SEEMIN', minseeingpix
              sxaddpar, h, 'SEEMAX', maxseeingpix
              if camera eq 'red' and lrisversion eq 1 then medseeingarcsec = medseeingpix*0.210 $
                                                      else medseeingarcsec = medseeingpix*0.135
              sxaddpar, h, 'SEEARCSC', medseeingarcsec, 'Median seeing in arcsec'
           endif
           if finite(transratio) then begin
              sxaddpar, h, 'TRANSRAT', maxreltrans/minreltrans, 'Transmission ratio of max/min images'
              sxaddpar, h, 'TRANSSTD', stdtrans, 'Std. dev. of relative transmission'
           endif
           for s = 0, nstack-1 do  sxaddpar, h, 'IMAGE'+clip(s), info[stacki[s]].filename
           for s = 0, nstack-1 do  sxaddpar, h, 'IMEXP'+clip(s), info[stacki[s]].exp
           for s = 0, nstack-1 do $
               if info[stacki[s]].zpt gt 0 then sxaddpar, h, 'ZEROPT'+clip(s), info[stacki[s]].zpt
           for s = 0, nstack-1 do $
               if info[stacki[s]].seepix gt 0 then sxaddpar, h, 'SEEING'+clip(s), info[stacki[s]].seepix
           for s = 0, nstack-1 do $
               if info[stacki[s]].trans gt 0 then sxaddpar, h, 'TRANS'+clip(s), info[stacki[s]].trans
           for s = 0, nstack-1 do $
               if info[stacki[s]].reltrans gt 0 then sxaddpar, h, 'RTRANS'+clip(s), info[stacki[s]].reltrans
           if nstack gt 1 then for s = 0, nstack-1 do sxaddpar, h, 'FSCALE'+clip(s), fluxscales[s]
           get_date, now
           sxdelpar, h, ['SOFTNAME','SOFTVERS','SOFTDATE','SOFTAUTH','SOFTINST','AUTHOR']
           hist = 'Processed by lpipe '+now
           sxaddhist, hist, h, /COMMENT
           mwrfits, data, outfile, h, /create ; add this header info

           ;;print, autoastrocommand+' '+outfile+' -q'
           ;;spawn, autoastrocommand+' '+outfile+' -q'
           ;;                 ; do another round of precise astrometry on the stack (probably unnecessary now)

           ; In the future, there will be a final alignment step.

        endif


     endfor
  endfor


     ; Combine left and right coadded images to produce final l+r images

  if keyword_set(final) then begin   ; only run this for whatever chip comes next (otherwise if r comes first it will assume l doesn't exist)

     ; Honestly this is a little dangerous as the photometric scaling may be slightly different especially on cloudy nights where the "reference"
     ; image is not guaranteed to be the same for the right and left chips.
     ; Would probably be best to actually open up the coadd files and get the list of images and weights and completely re-stack using that
     ; information.
     ; However, we are sticking with this for now because it is faster.  Just be careful on nonphotometric nights.

     coaddfilesr = file_search(imworkingdir+'coadd*r.fits')
     coaddfilesl = file_search(imworkingdir+'coadd*l.fits')
     if coaddfilesr ne [''] then coaddfilesr = coaddfilesr[where(coaddfilesr ne '')] ;else continue
     if coaddfilesl ne [''] then coaddfilesl = coaddfilesl[where(coaddfilesl ne '')] ;else continue

     ; match every r with an l and coadd
     for f = 0, n_elements(coaddfilesr)-1 do begin
        rfilename = coaddfilesr[f]
        if rfilename eq '' then continue
        lfilename = repstr(rfilename, 'r.fits','l.fits')
        outfileidir = repstr(repstr(rfilename, '.r.fits', '.fits'),'coadd','lris')
        outfilefdir = imfinaldir+repstr(repstr(removepath(outfileidir),'_RG850','_z'),'_I','_i')
        if file_test(outfilefdir) eq 0 or overwrite then begin
           match = where(coaddfilesl eq lfilename, ct)
           ;if ct eq 0 then logprint, 'No match for ', rfilename, ' among ', coaddfilesl, ' (looking for ', lfilename, ')'
           if ct ge 1 then begin
              stackcmd = swarpcommand + ' ' + rfilename + ',' + lfilename
              rweightname = repstr(rfilename, '.fits','.weight.fits') 
              lweightname = repstr(lfilename, '.fits','.weight.fits') 
              ;if file_test(rweightname) eq 0 then spawn, 'mv ' + strmid(rweightname,1) + ' ' + rweightname
              ;if file_test(lweightname) eq 0 then spawn, 'mv ' + strmid(lweightname,1) + ' ' + lweightname
              logprint
              logprint, 'Combining ', rfilename, ' and ', lfilename
              stackcmd = stackcmd + ' -IMAGEOUT_NAME ' + outfileidir
              stackcmd = stackcmd + ' -WEIGHTOUT_NAME ' + 'weight.temp.fits'
              stackcmd = stackcmd + ' -WEIGHT_TYPE MAP_WEIGHT'

              logprint, stackcmd
              spawn, stackcmd
              data = mrdfits(outfileidir, 0, h, /silent)
              weightdata = readfits('weight.temp.fits', /silent)
              spawn, 'rm -f weight.temp.fits' ; don't need it anymore
              wnodata = where(weightdata eq 0, ct)
              if ct gt 0 then data[wnodata] = !values.f_nan
              swarpscikeys = ['EXPTIME','GAIN']  ; formerly also included 'CRPIX1','CRPIX2', which was insane
              hr = headfits(rfilename)
              hl = headfits(lfilename)
              for k = 0, n_elements(swarpscikeys)-1 do sxaddpar, h, swarpscikeys[k], sxpar(hr, swarpscikeys[k], comment=cmt), cmt 
              for k = 0, n_elements(keepkeys)-1 do     sxaddpar, h, keepkeys[k],     sxpar(hr, keepkeys[k],     comment=cmt), cmt 
              for i = 0, 100 do begin
                key = 'IMAGE'
                rimagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, 'RIMAGE'+clip(i), clip(rimagekeyn)
              endfor
              for i = 0, 100 do begin
                key = 'IMAGE'
                limagekeyn = sxpar(hl, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, 'LIMAGE'+clip(i), clip(limagekeyn)
              endfor
              for i = 0, 100 do begin
                key = 'IMEXP'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              for i = 0, 100 do begin
                key = 'SEEING'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              for i = 0, 100 do begin
                key = 'ZEROPT'
                imagekeyn = sxpar(hr, key+clip(i), comment=cmt, count=ct)
                if ct gt 0 then sxaddpar, h, key+clip(i), imagekeyn
              endfor
              sxdelpar, h, ['SOFTNAME','SOFTVERS','SOFTDATE','SOFTAUTH','SOFTINST','AUTHOR']
              get_date, now
              hist = 'Processed by lpipe '+now
              sxaddhist, hist, h, /COMMENT
              mwrfits, data, outfileidir, h, /create
              spawn, 'mv '+outfileidir+' '+outfilefdir
           endif else begin

             spawn, 'cp -f '+rfilename+' '+outfilefdir

           endelse
        endif
     endfor
  endif

end



; -------------------------


pro lpipe, datadirectory=datadirectory, modestr=modestr, camerastr=camerastr, chipstr=chipstr, files=files, filters=filters, gratings=gratings, targets=targets, start=start, stop=stop, only=only, step=step, nocrclean=nocrclean, noreflat=noreflat, notelluric=notelluric, redo=redo, nofringe=nofringe, continuous=continuous, photometric=photometric, help=help, date=date, nodisplay=nodisplay, distort=distort, ignoredichroic=ignoredichroic, debugmsg=debugmsg, nophotometry=nophotometry, configure=configure, reducername=reducername
;   modestr      - Mode (imaging or spectroscopy)
;   camerastr    - Camera to process (red or blue)
;   chipstr      - Chip to process (left or right)
;   files        - File numbers to process
;   filters      - Filters to process
;   gratings     - Gratings to process
;   targets      - Targets to process
;   start        - Start with this step, skipping previous ones
;   stop         - End with this step, skipping subsequent ones
;   only         - Do only this step
;   step         -  (completely identical to only, takes precedence)
;   redo         - Overwrite any existing products
;   noreflat     - Do not perform a secondary sky-flat division
;   nocrclean    - Do not zap cosmic rays
;   notelluric   - Do not remove telluric absorption
;  (continuous   - Keep running, assimilating new data as it appears)
;  (live         - Don't stack/combine the last target)
;   photometric  - Use zeropoint solution to calibrate non-SDSS fields
; XX quick        - Set defaults to: no cr-cleaning, faster sky subtraction
; XX  nofringe     - Skip fringe production and correction for LRIS-R1
;   date         - String to indicate specific night/run to process
;   nodisplay    - Turn off plotting to the screen (in case of display issues)
;   distort      - Use nonlinear astrometric distortion terms
;   ignoredichroic - Neglect dichroic differences when applying flatfield
;   configure    - Write out the configuration file
;   reducername  - Specify who you are


; Load default parameters and interpret user arguments.

common lrisauto, autoastrocommand, swarpcommand, datadir, imworkingdir, spworkingdir2d, spworkingdir1d, imfinaldir, spfinaldir, lrisversion, wildchar, overwrite, fileseq, filterstr, gratingstr, targetstr, defaulttraceyred, defaulttraceyblue, displayoff, debug, reducer, lpversion
common lrisfail, prepfail, flatfail, catastrofail, relastrofail, fullastrofail, extractfail, wavsolfail, wavsolwarn, fluxcalfail

lpversion = '2017.05'
if n_elements(reducername) gt 0 then reducer = reducername
!quiet = 1

print, 'LPIPE version '+lpversion+' - new this version: output logging'


; 'combine' has been deprecated
; 'makefringe', 'rmfringe' - come after flatten
steps = ['prepare', 'makeflat', 'flatten', 'reflatten', 'split', 'crclean', 'skysubtract', $
         'wavcal', 'sum', 'trace', 'extract', 'response', 'fluxcal', 'combine', 'connect', $
         'astrometry', 'photometry', 'stack']
stephelp = ['Converts multi-fits to simple fits, bias subtracts, adds header info',$
            'Creates flat fields',$
            'Divides by flat fields',$
            'Create customized super-skyflats for each exposure (set noreflat=0)',$
            'Separates right and left chips',$
            'Removes cosmic rays',$
            'Subtracts sky lines from 2D spectra', $
            'Wavelength-calibrates arc spectra', $
            'Sums 2D spectra of the same target', $
            'Traces spectra', $
            'Extracts spectra', $
            'Solves response function of flux standards', $
            'Flux-calibrates spectra including telluric correction', $
            'Combines 1D spectra', $
            'Connects blue/red spectra', $
            'Solves for astrometric offset/rotation of images', $
            'Solves for zeropoint and transmission variations of images', $
            'Co-adds images']
   stepstr = ''
   for i = 0, n_elements(steps)-1 do stepstr += ' '+steps[i]

if keyword_set(help) then begin
   print, '  LPIPE - LRIS automatic processing pipeline'
   print, '  Command-line options:'
   print, "      mode   = 'i' for imaging only, 's' for spectroscopy only, default is both"
   print, "      camera = 'b' for blue only, 'r' for red only, default is both"
   print, "      chip   = 'both' to process both chips, default is right only after flat-fielding"
   print, "      step   = step to execute individually (see information below)"
   print, "      start  = step to start with"
   print, "      stop   = step to end with"
   print, "      datadir= location of raw data subdirectory"
   print, "      /redo  - overwrite subsequent files (otherwise pipeline resumes where it left off)"
   print, "  To control tracing of individual 2D spectra, run setapertures"
   print, "  To remove problematic files from processing, create the file 'badfiles.txt'"
   print, "  Additional tools:"
   print, "      lriscorrectheaders - run to repair missing values from headers before processing"
   print, "      lriscat            - run to print out a catalog of your data files"
   print, '  Steps are:'
   for i = 0, n_elements(steps)-1 do print, '      '+clip(steps[i],12)+' '+stephelp[i]
   print, '  For additional help and more options, go to http://www.astro.caltech.edu/~dperley/programs/lpipe.html'
   return
endif
if n_elements(step) eq 1 then begin  ; print list of steps for convenience
  if clip(step[0]) eq '1' then begin
    str = ' '
    for i = 0, n_elements(steps)-1 do str += ' '+clip(steps[i])
    print, str
    return
   endif
endif

close, /all


if n_elements(step) gt 0 and n_elements(stop) gt 0 then if step ne stop then begin
   print, 'Both step and stop are set: conflicting requests.'
   return
endif
if n_elements(step) gt 0 and n_elements(start) gt 0 then if step ne start then begin
   print, 'Both step and stop are set: conflicting requests.'
   return
endif

if n_elements(datadirectory) gt 0 then datadir = datadirectory
if n_elements(datadir) gt 0 then begin
  if strlen(datadir) gt 0 and strmid(datadir,strlen(datadir)-1,1) ne '/' then datadir = datadir + '/'
endif

if n_elements(files) then fileseq = strseq(files)
if n_elements(filterstr) then filterseq = strsplit(filters,',',/extract)
if n_elements(targetstr) then targetseq = strsplit(targets,',',/extract) ; this doesn't do anything
if n_elements(gratingstr) then gratingseq = strsplit(gratings,',',/extract)

if n_elements(noreflat) eq 0 then noreflat = 1  ; default is off


redo = keyword_set(redo)
nocrclean = keyword_set(nocrclean)
noreflat = keyword_set(noreflat)
notelluric = keyword_set(notelluric)
nodistort = keyword_set(distort) eq 0
nofringe = keyword_set(nofringe)
nophotometry = keyword_set(nophotometry)
debug = keyword_set(debugmsg)


cd, current=pwd
dirtree = strsplit(pwd,'/',/extract,count=nd)
lastdir = dirtree[nd-1]
if lastdir ne '' then lastdir += '/'  ; whether or not a slash is on the end is very
                                      ; confusing, need to rethink this.
if lastdir eq 'imredux/' or lastdir eq 'spredux/' or lastdir eq 'spredux1d/' or lastdir eq 'spredux2d/' then begin
     print, 'Currently in a reduction subdirectory.'
     print, "Type cd, '..' and rerun."
  ; could reinterpret this as run with the mode set to whatever this directory is...
     return
endif



; --- Process mode options ---

;  By default do both imaging and spectroscopy.
if n_elements(modestr) eq 0 then begin
   modestr = 'is' 
   if n_elements(step) gt 0 then begin  ; for certain mode-specific steps also set modestr
     if step eq 'skysubtract' or step eq 'extract' or step eq 'fluxcal' $
        or step eq 'combine' or step eq 'connect' $
        then  modestr = 's'
     if    step eq 'astrometry' or step eq 'photometry' or step eq 'stack' $
        then  modestr = 'i'
   endif
endif
modestr = strlowcase(modestr)
if modestr eq 'is' or modestr eq 'i,s' or modestr eq 'im,sp'  then modes = ['im', 'sp']
if modestr eq 'si' or modestr eq 's,i' or modestr eq 'sp,im'  then modes = ['sp', 'im']
if n_elements(modes) eq 0 then begin
   if strmid(modestr,0,1) eq 'i' then modes = ['im']
   if strmid(modestr,0,1) eq 's' then modes = ['sp']
endif
if n_elements(modes) eq 0 then begin
  print, 'Cannot recognize mode request: ', modestr
  return
endif

; Set up / check for the config file
if keyword_set(configure) then autolrissetup  ; this is normally not run now

; Load settings from the config file
autolrisdefaults, modes=modes

displayoff = keyword_set(nodisplay)
if keyword_set(date) then wildchar = clip(date+'??????',6)

; --- Process camera options ---

if n_elements(camerastr) eq 0 then camerastr = 'rb'
camerastr = strlowcase(camerastr)
if camerastr eq 'br' or camerastr eq 'b,r' then cameras = ['blue', 'red']
if camerastr eq 'rb' or camerastr eq 'r,b' then cameras = ['red', 'blue']
if n_elements(cameras) eq 0 then begin
   if strmid(camerastr,0,1) eq 'b' then cameras = ['blue']
   if strmid(camerastr,0,1) eq 'r' then cameras = ['red']
endif
if n_elements(cameras) eq 0 then begin
  print, 'Cannot recognize camera request: ', camerastr
  return
endif

; --- Process chip options
if n_elements(chipstr) eq 0 then chipstr = 'r'
chipstr = strlowcase(chipstr)
if chipstr eq 'lr' or chipstr eq 'l,r' then chips = ['l', 'r']
if chipstr eq 'rl' or chipstr eq 'r,l' or chipstr eq 'both' or chipstr eq 'b' then chips = ['r', 'l']
if n_elements(chips) eq 0 then begin  
   if strmid(chipstr,0,1) eq 'l' then chips = ['l']
   if strmid(chipstr,0,1) eq 'r' then chips = ['r']
endif
if n_elements(chips) eq 0 then begin
  print, 'Cannot recognize chip request: ', chipstr
  return
endif

; --- Process step options
if n_elements(start) gt 0 then begin
   w = (where(steps eq start, ct)) [0]
   if ct eq 0 then begin
      print, "Invalid starting step '", start, "'
      print, "Must be one of: ", steps
      return
   endif
   steps = steps[w:*]
endif
if n_elements(stop) gt 0 then begin
   w = (where(steps eq stop, ct)) [0]
   if ct eq 0 then begin
      print, "Invalid stopping step '", stop, "'
      print, "Must be one of: ", steps
      if n_elements(start) gt 0 then print, 'Note that start is also set.'
      return
   endif
   steps = steps[0:w]
endif
if n_elements(step) gt 0 then only = step
if n_elements(only) gt 0 then begin
   w = (where(steps eq only, ct)) [0]
   if ct eq 0 then begin
      print, "Invalid step '", only, "'
      print, "Must be one of: ", steps
      if n_elements(start) gt 0 then print, 'Note that start is also set.'
      if n_elements(stop) gt 0 then print, 'Note that stop is also set.'
      return
   endif
   steps = steps[w]
endif

; For imaging, check if autoastrometry, sextractor, swarp are installed and functioning

if total(modes eq 'im') ge 1 then begin
   if total(steps eq 'astrometry') gt 0 or total(steps eq 'photometry') gt 0 or total(steps eq 'stack') gt 0 then begin
      if file_test('temp.txt') then spawn, 'rm -f temp.txt'
      cmd = getsexpath()+' -d '+' > temp.txt'
      spawn, cmd
      c = countlines('temp.txt')
      if c eq 0 then begin
         print, 'Error: Sextractor is not installed or not configured.'
         print, '   Cannot run image alignment steps.'
         print, "   Configure or set mode='s' or stop='crclean'"
         return
      endif
   endif

   if total(steps eq 'stack') gt 0 then begin
      if file_test('temp.txt') then spawn, 'rm -f temp.txt'
      cmd = swarpcommand+' -d > temp.txt'
      spawn, cmd
      c = countlines('temp.txt')
      if c eq 0 then begin
         print, 'Error: Swarp is not installed or not configured.'
         print, '   Cannot run image coadds.'
         print, "   Configure or set mode='s' or stop='photometry'"
         return
      endif
   endif

   if total(steps eq 'astrometry') or total(steps eq 'stack') gt 0 then begin
      if file_test('temp.txt') then spawn, 'rm -f temp.txt'
      cmd = autoastrocommand+' > temp.txt'
      spawn, cmd
      c = countlines('temp.txt')
      if c eq 0 then begin
         print, 'Error: Autoastrometry is not installed or not configured.'
         print, '   Cannot run image alignment steps.'
         print, "   Configure or set mode='s' or stop='crclean'"
         return
      endif
      if c eq 22 then begin
         print, 'Error: Python is not configured.'
         print, '   Cannot run image alignment steps.'
         print, "   Configure or set mode='s' or stop='crclean'"
      endif
   endif
endif


; Investigate the data and rename files to modern format if necessary.
;   (no longer automatic - user should do this themselves using lrisrename now.)
;noldfiles = countfiles(datadir+'lred*.fits') + countfiles(datadir+'lblue*.fits')
; ;+ countfiles(datadir+'rfoc*.fits') + countfiles(datadir+'bfoc*.fits')
;if noldfiles gt 0 then begin
;   print, 'Image file names are in old format.  Updating...'
;   autolrisrename
;endif


lrisversion = getlrisversion(datadir=datadir, /quiet)
if lrisversion le 0 then lrisversion=getlrisversion(datadir=spworkingdir2d, /quiet)
if lrisversion le 0 then lrisversion=getlrisversion(datadir=spworkingdir1d, /quiet)
if lrisversion le 0 then lrisversion=getlrisversion(datadir=imworkingdir, /quiet)
if lrisversion eq 0 then begin
  check1 = file_search(spworkingdir1d+'*.fits', count=ct1)
  check2 = file_search(spworkingdir1d+'*.spec', count=ct2)
  if ct1+ct2 eq 0 then begin
     print, '  Found no files to process.'
     print, "  If raw files are in a different directory, try specifying: datadir='/path/to/files/'"
     return
  endif else begin
     lrisversion = 3   ; assume this for now, could check later if it actually affects anything
  endelse
endif

if redo then overwrite = 1 else overwrite = 0

; if keyword_set(nocrclean) eq 0 then flag = 1  this doesn't do anything

if n_elements(modes) eq 1 then begin  ; remove redundant steps if mode is set to only do imaging or spectroscopy
   spsteps = where(steps ne 'astrometry' and steps ne 'photometry' and steps ne 'stack',ct)
   if modes[0] eq 'sp' and ct gt 0 then steps = steps[spsteps]
   imsteps = where(steps ne 'trace' and steps ne 'wavcal' and steps ne 'fluxcal' and steps ne 'combine' and steps ne 'connect',ct)
   if modes[0] eq 'im' and ct gt 0 then steps = steps[imsteps]
endif


nsteps = n_elements(steps)
nmodes = n_elements(modes)
ncameras = n_elements(cameras)
nchips = n_elements(chips)

npipeiter = 1L
if keyword_set(continuous) eq 0 then maxpipeiter = 1L else maxpipeiter = 100000L

logprint, logfile='lpipelog.txt'
logprint, ' --- LPIPE version '+lpversion+' started UT '+tstamp(/ut)+ ' ---', /logonly
if n_elements(start) gt 0 then logprint, ' start='+clip(start), /logonly
if n_elements(stop) gt 0 then logprint, ' stop='+clip(stop), /logonly
if n_elements(step) gt 0 then logprint, ' step='+clip(step), /logonly
if n_elements(modestr) gt 0 then if strlen(modestr) lt 2 then logprint, ' mode='+clip(modestr), /logonly
if n_elements(camerastr) gt 0 then if strlen(camerastr) lt 2 then logprint, ' cam='+clip(camerastr), /logonly
; could print arguments and such...

while npipeiter le maxpipeiter do begin
for istep = 0, nsteps-1 do begin
   instep = steps[istep]
   logprint, instep

   for icam = 0, ncameras-1 do begin
      camera = cameras[icam]
      ca = strmid(cameras[icam],0,2)

      if instep eq 'prepare' then begin
         ; the mode setting is passed on to the routine itself.
         autolrisprepare, modestr=modestr, camstr=ca
      endif

      for imode = 0, nmodes-1 do begin
         mo = strmid(modes[imode],0,2)
         if mo eq 'im' then begin   
            if instep eq 'makeflat' then autolrismakeimflat, cam=camera, chip=''
            if instep eq 'flatten'  then autolrisimflatten,  cam=camera, chip='', ignoredichroic=ignoredichroic
            if lrisversion eq 1 and nofringe eq 0 and cameras[icam] eq 'red' then begin
               if instep eq 'makefringe' then autolrismakefringe, chip=ch, cam=ca
               if instep eq 'rmfringe'   then autolrisrmfringe, chip=ch, cam=ca
            endif
            if instep eq 'reflatten' and noreflat eq 0 then autolrisreflatten, chip='', cam=ca, filter=['RG850','I','R','V','G']
         endif
         if mo eq 'sp' then begin
            if instep eq 'makeflat' then autolrismakespecflat, cam=camera, chip=''
            if instep eq 'flatten'  then autolrisspecflatten,  cam=camera, chip=''
         endif

         if instep eq 'split' and (lrisversion gt 1 or cameras[icam] eq 'blue') then begin  ; and n_elements(chips) eq 2 
            if n_elements(chips) eq 1 then ch = chips[0] else ch = ''
            autolrissplit, cam=camera, mode=mo, ch=ch, nodistort=nodistort
         endif

         for ichip = 0, nchips-1 do begin
            if lrisversion eq 1 and camera eq 'red' then ch = '' $
                                                    else ch = strmid(chips[ichip],0,1)

            if nocrclean eq 0 and instep eq 'crclean' then begin
               if mo eq 'im' then autolriscrcleanim,    cam=camera, chip=ch
               if mo eq 'sp' then autolriscrcleanspec,  cam=camera, chip=ch
            endif 
 
            if mo eq 'im' then begin   
               if instep eq 'astrometry' then autolrisastrometry, chip=ch, cam=camera, nocrclean=nocrclean, nodistort=nodistort
               if camera eq 'blue' then bl = 1 else bl = 0
               if camera eq 'red' then  re = 1 else re = 0
               ;if n_elements(chips) eq 1 then begin
                  if instep eq 'photometry' and nophotometry eq 0 then autolrisphotometry, chip=ch, camera=camera, photometric=photometric
               ;endif else begin    ; eventually need to re-enable this?
               ;   if instep eq 'photometry' and ichip eq 1 then autolrisphotometry, camera=camera
               ;endelse
               if instep eq 'stack'      then autolrisstack, chip=ch, cam=camera, live=live, final=(ichip eq nchips-1)
            endif
            if mo eq 'sp' then begin   
               if instep eq 'skysubtract' then autolrisskysubtract, chip=ch, cam=camera, nocrclean=nocrclean
               if instep eq 'wavcal' then autolriswavcal, chip=ch, cam=camera
               ;if instep eq 'trace' then autolrisfindtrace, chip=ch, cam=camera
               if instep eq 'sum' then autolrissum, chip=ch, cam=camera
               if instep eq 'trace' then autolristrace, chip=ch, cam=camera
               if instep eq 'extract' and ch ne 'l' then autolrisextract, chip=ch, cam=camera
            endif
         endfor
         if mo eq 'sp' then begin ;  not really necessary
            if instep eq 'response' then autolrisresponse, chip='r', cam=camera
            if instep eq 'fluxcal' then autolrisfluxcal, chip='r', cam=camera, notelluric=notelluric ; could move this in the chip loop
            if n_elements(step) gt 0 then $
            if instep eq 'combine' and step eq 'combine' then autolriscombine, camera=camera, live=live ; only if specifically requested
         endif
       
      endfor ; mode

   endfor ; camera

   
   ; Camera has ceased to be relevant for the following

   for imode = 0, nmodes-1 do begin

      if mo eq 'im' and instep eq 'photometry' then autolrisphotcat ; OK that this happens after stacking.

      if mo eq 'sp' and n_elements(cameras) ge 2 and instep eq 'connect' then autolrisconnect, live=live

      ; a redblue imaging alignment procedure would also go here      

   endfor

endfor ; step



logprint

nprepfail = n_elements(prepfail)-1
if nprepfail ge 1 then begin ;element 0 is blank
  logprint
  logprint, 'Unable to prepare the following files:'
  for f = 1, n_elements(prepfail)-1 do begin
    h = headfits(prepfail[f])
    logprint, clip(removepath(prepfail[f]),25),' ',clip(sxpar(h,'SLITNAME'),10),' ', string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ',clip(sxpar(h,'TARGNAME'),16), ' ',clip(sxpar(h,'OBJECT'),16), ' ', repstr(repstr(clip(sxpar(h,'BINNING')),',','x'),' ','')
  endfor
endif

nflatfail = n_elements(flatfail)-1
if nflatfail ge 1 then begin ;element 0 is blank
  logprint
  logprint, 'Unable to flat-field the following images:'
  for f = 1, n_elements(flatfail)-1 do begin
    h = headfits(flatfail[f])
    gg = sxpar(h,'GRNAME')
    if clip(gg) eq '0' then gg = sxpar(h,'SLITNAME')
    logprint, clip(removepath(flatfail[f]),25), string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ', clip(sxpar(h,'FILTER'),5), ' ', clip(sxpar(h,'DICHNAME'),7),' ',clip(gg,12), ' ',clip(sxpar(h,'TARGNAME'),16), ' ', repstr(clip(sxpar(h,'BINNING')),',','x')
   ;, clip(sxpar(h,'WINDOW')), ' ', clip(sxpar(h,'PANE'))
  endfor
  flats = [file_search(imworkingdir+'*flat*.fits'),file_search(spworkingdir2d+'*flat*.fits')]
  nflats = total(flats ne '')
  if nflats gt 0 then begin
    flats = flats[where(flats ne '')]
    logprint, 'Available flats are:'
    for f = 0, nflats-1 do begin
       h = headfits(flats[f], /silent)
       gg = sxpar(h,'GRNAME')
       if clip(gg) eq '0' then gg = sxpar(h,'SLITNAME')

logprint, clip(removepath(flats[f]),25), string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ', clip(sxpar(h,'FILTER'),5), ' ', clip(sxpar(h,'DICHNAME'),7),' ',clip(gg,12), ' ',clip(sxpar(h,'TARGNAME'),16), ' ', repstr(clip(sxpar(h,'BINNING')),',','x')
   ;, clip(sxpar(h,'WINDOW')), ' ', clip(sxpar(h,'PANE'))
     endfor
   endif else begin
     logprint, 'No processed flat-fields exist!  Check inputs or copy raw flat-field files to data directory.'
   endelse
endif

nefail = n_elements(extractfail)-1
if nefail ge 1 then begin
  logprint, 'Unable to model trace and extract 1D spectrum for the following 2D spectra:'
  for f = 1, n_elements(extractfail)-1 do begin
    h = headfits(extractfail[f])
    logprint, clip(removepath(extractfail[f]),32), rclip(sxpar(h,'ELAPTIME'),5)+'s', ' ', clip(sxpar(h,'GRNAME'),9), ' ', clip(sxpar(h,'TARGNAME'),16)
  endfor
endif
nwfail = n_elements(wavsolfail)-1
if nwfail ge 1 then begin
   logprint, 'Unable to wavelength-calibrate the following spectra:'
   for f = 1, n_elements(wavsolfail)-1 do begin
      h = ['']
      if strpos(wavsolfail[f],'.spec') gt 0 then h = headspec(wavsolfail[f])
      if strpos(wavsolfail[f],'.fits') gt 0 then h = headfits(wavsolfail[f])
      logprint, clip(removepath(wavsolfail[f]),32), rclip(sxpar(h,'ELAPTIME'),5)+'s', ' ', clip(sxpar(h,'GRNAME'),9), ' ', clip(sxpar(h,'TARGNAME'),16)
   endfor
endif
nwwarn = n_elements(wavsolwarn)-1
if nwwarn ge 1 then begin
   logprint, 'Wavelength solutions for the following spectra are questionable; check sky-line verification plots:'
   for f = 1, n_elements(wavsolwarn)-1 do begin
      h = headfits(wavsolwarn[f])
      logprint, clip(removepath(wavsolwarn[f]),32), rclip(sxpar(h,'ELAPTIME'),5)+'s', ' ', clip(sxpar(h,'GRNAME'),9), ' ', clip(sxpar(h,'TARGNAME'),16)
   endfor
endif
nfluxfail = n_elements(fluxcalfail)-1
if nfluxfail ge 1 then begin
   logprint, 'Unable to flux-calibrate the following spectra:'
   for f = 1, n_elements(fluxcalfail)-1 do begin
      h = headspec(fluxcalfail[f])
      logprint, clip(removepath(fluxcalfail[f]),32), rclip(sxpar(h,'ELAPTIME'),5)+'s', ' ', clip(sxpar(h,'GRNAME'),9), ' ', 'D'+clip(sxpar(h,'DICHNAME'),7), ' ', clip(sxpar(h,'FILTER'),5), ' ', clip(sxpar(h,'TARGNAME'),16)
   endfor
endif

nafail = n_elements(relastrofail)-1 + n_elements(fullastrofail)-1 + n_elements(catastrofail)-1
if nafail ge 1 then begin
  logprint
  if n_elements(catastrofail) gt 1 then logprint, 'Unable to produce astrometric catalogs for the following reference images:'
  for f = 1, n_elements(catastrofail)-1 do begin
    h = headfits(catastrofail[f])
    logprint, clip(removepath(catastrofail[f]),25), string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ', clip(sxpar(h,'FILTER'),4), ' ', clip(sxpar(h,'TARGNAME'),16), ' ', clip(sxpar(h,'COUNTS'))
  endfor
  if n_elements(relastrofail) gt 1 then logprint, 'Relative astrometry failed for the following images, but absolute was successful:'
  for f = 1, n_elements(relastrofail)-1 do begin
    h = headfits(relastrofail[f])
    logprint, clip(removepath(relastrofail[f]),25), string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ', clip(sxpar(h,'FILTER'),4), ' ', clip(sxpar(h,'TARGNAME'),16), ' ', clip(sxpar(h,'COUNTS'))
  endfor
  if n_elements(fullastrofail) gt 1 then logprint, 'All astrometry failed for the following images (not stacked):'
  for f = 1, n_elements(fullastrofail)-1 do begin
    h = headfits(fullastrofail[f])
    logprint, clip(removepath(fullastrofail[f]),25), string(sxpar(h,'ELAPTIME'),format='(I5)'), ' ', clip(sxpar(h,'FILTER'),4), ' ', clip(sxpar(h,'TARGNAME'),16), ' ', clip(sxpar(h,'COUNTS'))
  endfor
endif; else begin
  ;print, 'Finished with no astrometric problems.'
;endelse

if keyword_set(continuous) then begin
   logprint, 'Pipeline iteration complete.'
   logprint, 'Continuous mode is active.'
   logprint, 'Pausing for 30 seconds, then will loop from the beginning.  Press Ctrl+C to stop.'
   for s = 30, 0, -1 do begin 
      wait, 1.0
   endfor
endif

npipeiter += 1
endwhile


print, 'Processing complete.'
;while !window ge 0 do wdelete, !window
!quiet = 0

; cleanup- should ultimately put these things in imredux

if file_test('temp*.*') gt 0 then spawn, 'rm -f temp*.*'
if file_test('det.*') gt 0 then spawn, 'rm -f det.*'
if file_test('cat.*') gt 0 then spawn, 'rm -f cat.*'
if file_test('match*') gt 0 then spawn, 'rm -f match*'
if file_test('sex*') gt 0 then spawn, 'rm -f sex*' ; this expunges user preferences, of course...
if file_test('swarp.xml') gt 0 then spawn, 'rm -f swarp.xml'
if file_test('default.swarp') gt 0 then spawn, 'rm -f default.swarp'

!p.multi = 0
!p.position = 0

logprint, ' --- LPIPE version '+lpversion+' completed UT '+tstamp(/utc)+' ---', /logonly
logprint, /close

end


; add some tools
; pro setobjectpos, name, y, width, bakmin, bakmax, tracefile
; pro setbadfile, cam, number

 
